;**********************************************************
; This is a sample program to query a video adapters capabilities
; conforming to VESA SVGA BIOS Extension.
; It prints to a standard output device VESA BIOS's version
; and name, amount of video memory installed, and characteristics
; of all video modes, avilable with the card.
; The program is compiled under MASM 6.0
; (c) This programm is written by Kiselev Y.
;**********************************************************

dosseg
.model small
.stack 100h
include vesainfo.inc			; VESA structures & constants

.data
VESASign	db 'VESA'               ; VESA signature
SVGAInfo	SVGAInfoBlock <>
ModeInfo	ModeInfoBlock <>

					; Messages
VESABIOSNotInst	db 'VESA BIOS extensions not installed.',10,13,'$'
RetError	db 'Error retrieving SVGA information : ','$'
SSign		db 'Unknown signature - ','$'
VersionStr	db 'VESA BIOS extension version ','$'
Inst		db ' is installed.',10,13,'$'
NoOEMStr	db 'No OEM string is present.',10,13,'$'
OEMStr		db 'OEM string is : ','$'
DevComp		db 'Device capabilities are ','$'
CallnotAvail	db 'Mode information not available.',10,13,'$'
RetError1	db 'Error accessing mode information : ','$'
MemBlkInst	db 'Installed video memory size (in Kb) : ','$'

					; Header of table
Header1		db 'Mode Attr WindowA WindowB Window Mem contr BPL   Format   Char    BPP   MM    Pg','$'
Header2		db '          att seg att seg gr siz    call                  Cell NBP  Bnks   SoB',10,13,'$'

; Variable StringBuffer is filled with ASCII characters and
; then printed via procedure DispString. I found it an easy way
; to provide formatted output.
;
StringBuffer	db 80 dup (' '),'$'             ; String buffer
hex		db '0123456789ABCDEF'		; table for hex convertion
Ver		dw ?				; internal variable


.code
.startup				; initialize segment registers
.286
					; Main program
		mov ax,@data
		mov es,ax
		call VESAPresent	; Inquire about VESA information
		jc vesa_nopresent       ; Jump if VESA not installed
		call PutVESAInfo        ; Display VESA information

		lea dx,Header1		; Show table header
		mov ah,9
		int 21h
		lea dx,Header2
		int 21h

		call PutModesInfo	; Inquire about mode information
vesa_nopresent:
		mov ax,4c00h		; Exit program
		int 21h

					; Subprcedures

					; Inquire about VESA information

VESAPresent	proc near
		mov ah,VESACall         ; AH - function number
		mov al,GetVESAInfo      ; AL - subfunction number
		lea di,SVGAInfo         ; ES:DI - where to store info
		int 10h			; Make VESA call
		cmp al,4fh		; AX = 004fh if success
		je vesa_inst
		lea dx,VESABIOSNotInst	; "VESA not installed" message
		call DispString
		stc
		ret
vesa_inst:
		or ah,ah
		jz no_error

		lea dx,RetError		; Error massage
		call DispString

		mov al,ah		; Output error number
		lea bx,StringBuffer     ; Convert number to ASCII
		xor di,di
		call PutHexByte
		mov byte ptr [bx+di],'$'
		mov dx,bx
		call DispString		; Output ASCII string

		call LFCR		; Terminate string
		stc
		ret
no_error:
		lea si,VESASign		; Check VESA signature
		mov cx,4
		repe cmpsb
		je ok_sign
					; Not 'VESA' signature
		lea dx,SSign
		call DispString         ; Report error

		mov cx,4		; Output actual signature
		mov ah,2
next_char:
		mov dl,[di]
		int 21h
		loop next_char

		call LFCR
		stc
		ret
ok_sign:
		ret
VESAPresent	endp

					; Display VESA information

PutVESAInfo	proc near
		lea dx,VersionStr	; Put VESA version number
		call DispString
		mov dx,SVGAInfo.VESAVersion
		mov Ver,dx
		or dx,3030h		; Convert to decimal
		xchg dh,dl
		mov ah,2
		int 21h
		mov dl,'.'		; Version number looks like  X.X
		int 21h
		mov dl,dh
		int 21h
		lea dx,Inst
		call DispString

		cmp word ptr SVGAInfo.OEMStringPtr[2],0	;NULL check
		jne show_oemstring
		lea dx,NoOEMStr		; OEM String is absent
		call DispString
		jmp short next_inf
show_oemstring:
		lea dx,OEMStr
		call DispString
		les di,SVGAInfo.OEMStringPtr[0]
		mov ah,2
next_char1:                             ; Put OEM string
		mov dl,es:[di]
		or dl,dl      		; ASCIIZ string
		jz end_str
		int 21h
		inc di
		jmp short next_char1
end_str:
		call LFCR
next_inf:
		lea dx,DevComp		; Output capabilities
		call DispString
		mov ax,word ptr SVGAInfo.Capabilities[0]
		lea bx,StringBuffer
		xor di,di
		call PutHexWord
		mov ax,word ptr SVGAInfo.Capabilities[2]
		call PutHexWord
		mov byte ptr [bx+di],'$'
		mov dx,bx
		call DispString
		call LFCR

		cmp Ver,0100h
		je leave_pr
					; Output video memory installed
		lea dx,MemBlkInst       ; (Version 1.1)
		call DispString
		lea bx,StringBuffer
		xor di,di
		mov ax,SVGAInfo.MemBlockInst	; in 64K blocks
		shl ax,6			; *64 (size in Kb)
		call PutDecWord
		mov byte ptr [bx+di],'$'
		mov dx,bx
		call DispString
		call LFCR
leave_pr:
		les si,SVGAInfo.VideoModePtr	; ES:[SI] -> modes list
		ret
PutVESAInfo	endp

;Inquire about mode information

PutModesInfo	proc near
next_mode:
		mov cx,es:[si]		; CX = video mode
		or cx,cx		; -1 = end of list
		js the_end
		mov ax,ds
		push es
		mov es,ax
		lea di,ModeInfo		; ES:DI - where to store info
		mov ah,VESACall		; AH - function number
		mov al,GetModeInfo	; AL - subfunction number
		int 10h  		; Make VESA call
		pop es
		cmp al,4fh		; AX = 004fh if success
		je call_ok

		lea dx,CallnotAvail	;call not available
		call DispString
		add si,2
		jmp short next_mode
call_ok:
		cmp ah,0
		je no_err
		lea dx,RetError1	; Error message
		call DispString
		mov al,ah
		lea bx,StringBuffer	; Output error number
		xor di,di
		call PutHexByte
		mov byte ptr [bx+di],'$'
		mov dx,bx
		call DispString
		call LFCR
		add si,2
		jmp short next_mode
no_err:
		call PutModeInfo	; Display mode information
		add si,2
		jmp short next_mode
the_end:
		ret
PutModesInfo	endp

					; Display mode information
PutModeInfo	proc near
		lea bx,StringBuffer
		mov dx,es
		mov ax,ds
		mov es,ax
		mov ax,2020h		; Fill StringBuffer with spaces
		push cx
		mov cx,40
		mov di,bx
		rep stosw
		mov es,dx
		pop cx
		mov ax,cx
		xor di,di
		call PutHexWord
		mov ax,ModeInfo.ModeAttributes
		test al,HardSupport		; Hardware configuration support
		jz not_hsupp
		mov byte ptr [bx+5],'H'
not_hsupp:
		test al,BIOSSupport		; BIOS function support
		jz not_biossupp
		mov byte ptr [bx+6],'B'
not_biossupp:
		mov byte ptr [bx+7],'M'
		test al,Color			; Mode is color or mono
		jz mono
		mov byte ptr [bx+7],'C'
mono:
		mov byte ptr [bx+8],'T'
		test al,GraphicsMode		; Mode is graphics or text
		jz textm
		mov byte ptr [bx+8],'G'
textm:
		mov al,ModeInfo.WinAAttributes
		test al,WinExists
		jz no_winxa
		test al,WinRead			;window A readable
		jz no_reada
		mov byte ptr [bx+10],'R'
no_reada:
		test al,WinWrite		;window A writable
		jz no_writea
		mov byte ptr [bx+11],'W'
no_writea:
		mov di,13
		mov ax,ModeInfo.WinASegment
		call PutHexWord
no_winxa:
		mov al,ModeInfo.WinBAttributes
		test al,WinExists
		jz no_winxb
		test al,WinRead                 ;window B readable
		jz no_readb
		mov byte ptr [bx+18],'R'
no_readb:
		test al,WinWrite
		jz no_writeb
		mov byte ptr [bx+19],'W'	;window B writable
no_writeb:
		mov di,21
		mov ax,ModeInfo.WinBSegment
		call PutHexWord
no_winxb:
		mov di,26
		mov ax,ModeInfo.WinGranularity
		call PutDecWord
		mov di,29
		mov ax,ModeInfo.WinSize
		call PutDecWord
		mov di,33
		mov ax,word ptr ModeInfo.WinFuncPtr[2]
		call PutHexWord
		mov byte ptr [bx+di],':'	; Far pointer - XXXX:XXXX
		inc di
		mov ax,word ptr ModeInfo.WinFuncPtr[0]
		call PutHexWord
		mov di,43
		mov ax,ModeInfo.BytesPerLine
		call PutDecWord
		mov di,48
		mov ax,ModeInfo.XResolution
		call PutDecWord
		mov byte ptr [bx+di],'x'	; Resolution like 1024x768
		inc di
		mov ax,ModeInfo.YResolution
		call PutDecWord
		mov di,58
		mov ah,0
		mov al,ModeInfo.XCharCellSize
		call PutDecWord
		mov byte ptr [bx+di],'x'	; Char cell like 8x16
		inc di
		mov ah,0
		mov al,ModeInfo.YCharCellSize
		call PutDecWord
		mov di,64
		mov ah,0
		mov al,ModeInfo.NumberOfPlanes
		call PutDecWord
		mov di,66
		mov ah,0
		mov al,ModeInfo.BitsPerPixel
		call PutDecWord
		mov di,69
		mov ah,0
		mov al,ModeInfo.NumberOfBanks
		call PutDecWord
		mov di,72
		mov ah,0
		mov al,ModeInfo.MemoryModel
		call PutHexByte
		mov di,75
		mov ah,0
		mov al,ModeInfo.BankSize
		call PutDecWord
		mov di,78
		mov ah,0
		mov al,ModeInfo.NumberofPages
		call PutDecWord
		mov dx,bx
		call DispString		; Output mode information
		ret
PutModeInfo	endp

;output text string
DispString	proc near
		mov ah,9
		int 21h
		ret
DispString	endp

;perform line feed and carriage return
LFCR		proc near
		mov dl,10
		mov ah,2
		int 21h
		mov dl,13
		int 21h
		ret
LFCR		endp

;convert byte bin2hex
;AL = 8 bit value to convert
PutHexByte	proc near
		push bx
		add di,bx
		lea bx,hex
		mov dh,al
		shr al,4
		xlat		;convert high half byte
		mov [di],al
		mov al,dh
		and al,0fh
		xlat		;convert low half byte
		mov [di+1],al
		add di,2
		pop bx
		sub di,bx
		ret
PutHexByte	endp

;covert word bin2hex
;AX = 16 bit value to convert
PutHexWord	proc near
		mov cl,al
		mov al,ah
		call PutHexByte
		mov al,cl
		call PutHexByte
		ret
PutHexWord	endp

;Binary to ASCII conversion routine
;limit value - 65535
;AX = 16 bit value to convert
PutDecWord	proc near
		push bx
		add di,bx
		mov bx,10
		xor cx,cx
next_digit:
		xor dx,dx
		div bx		;convert to decimal digits
		push dx
		inc cx
		or ax,ax
		jnz next_digit
putnext_digit:
		pop ax
		add al,'0'	;to get ASCII code
		mov [di],al
		inc di
		loop putnext_digit
		pop bx
		sub di,bx
		ret
PutDecWord	endp
end