#include <mem.h>
#include "common.h"
#include "graph.h"
#include "grfile.h"
#include "palette.h"
#include "mapper2.h"
#include "quant.h"
#include "formconv.h"

// auxiliar class that hides details of color quantization

const rbits = 3;
const gbits = 3;
const bbits = 2;
// init simple standard palette for 256 colors mode
void stand_palette(VGApalette &pal) {
	int i = 0,r,g,b;
	for ( r = 0; r < (1<<rbits); r++)
		for ( g = 0; g < (1<<gbits); g++)
			for ( b = 0; b < (1<<bbits); b++){
				pal[i].r = r << (6 - rbits);
				pal[i].g = g << (6 - gbits);
				pal[i++].b = b << (6 - bbits);
				}
      setDACblock(0,256,pal);
}

void put_TC_row(BYTE *rp, BYTE *gp, BYTE *bp,
		int x,int y,int xcount) {
	int i,r,g,b,c;
	BYTE *line,*l;

	line = new BYTE[xcount];
	l = line;
	for ( i = 0; i < xcount; i++) {
		c =  (*(rp++) >> (8-rbits)) << (gbits+bbits);
		c += (*(gp++) >> (8-gbits)) << bbits;
		c += *(bp++) >> (8-bbits);
		*(l++) = c;
	}
	putrow(rowptr(line),x,y,xcount);
	delete line;
}


color_reductor::color_reductor(graph_file_reader *input_file_,
				BOOL view_pass1_, BOOL view_pass2_,
				int colors) {

	ready_for_pass2 = initialized = FALSE;
	selector = NULL; keeper = NULL; mapper = NULL; tmp_line = NULL;
	input_file = input_file_;
	view_pass1 = view_pass1_;
	view_pass2 = view_pass2_;
	if ((input_file -> source_type != TRUECOLORIMG) &&
		(input_file -> source_type != COLOR256IMG))
		return;
	selector = new color_selector(min(colors,255),input_file -> width);
	if ((selector == NULL) || (! selector -> initialized)) return;

	keeper = new image_keeper(FALSE, // force_file,
			input_file -> source_type,
			input_file -> width,
			input_file -> height,
			input_file -> palette,
			input_file ->top_to_bottom);
	if ((keeper == NULL) || (! keeper -> file_ok)) return;
	if (input_file -> source_type == COLOR256IMG)
	   selector->prepare_scan_palette (input_file -> palette);

	initialized = TRUE;
}
color_reductor:: ~color_reductor() {
	if (selector != NULL) delete selector;
	if (keeper != NULL) delete keeper;
	if (mapper != NULL) delete mapper;
	if (tmp_line != NULL) delete tmp_line;
}

int color_reductor::pass1() {
	BOOL ok;
	VGApalette pal;
	int cols,y, inc_y,i = 0;

	if (!initialized) return(-1);

	if (view_pass1) stand_palette(pal);
	input_file->seek_start();
	if  (input_file->top_to_bottom) {
		y = 0;
		inc_y = 1;
		}
	else  {
		y = getmaxy();
		inc_y = -1;
		}

	do {
		ok =  (input_file -> get_stand_line() == 0);
		if (input_file -> source_type == COLOR256IMG)
		     selector->prescan_line_palette(input_file -> standart_line);
		else selector->prescan_line(input_file ->standart_line,
					&input_file ->standart_line[input_file->width],
					&input_file ->standart_line[2*input_file->width]);
		keeper-> store_line(input_file ->standart_line);
		if (view_pass1) put_TC_row(input_file ->standart_line,&input_file ->standart_line[input_file->width],&input_file ->standart_line[2*input_file->width], 0,y,input_file->width);
		else refresh((++i*30L)/input_file->height,FALSE);
		y += inc_y;
	} while(ok);

	selector->select_colors();
	cols = selector->actual_number_of_colors;
	memmove(my_colormap,selector->my_colormap,cols*sizeof(colortype));
	delete selector; selector = NULL;
	mapper = new color_mapper_palette_dith(cols,input_file->width,my_colormap);
	ready_for_pass2 = mapper -> initialized;

	input_file->seek_start();
	if (view_pass2) {
		for ( i = 0; i < mapper->actual_number_of_colors; i++) {
			pal[i].r = mapper->my_colormap[i].r / 4;
			pal[i].g = mapper->my_colormap[i].g / 4;
			pal[i].b = mapper->my_colormap[i].b / 4;
			}
		setDACblock(0,256,pal);
		}

	if  (input_file->top_to_bottom)  y = 0;	else  y = getmaxy();
	if (!ready_for_pass2) return(-1);
	else {
		if (input_file -> source_type == COLOR256IMG)
		   tmp_line = new BYTE[3*input_file ->width+1];
		return(1);
	     }
}

void color_reductor::process_line(int line_n, BYTE *out_line) {
	BYTE *curline;
	if (!ready_for_pass2) return;
	curline = keeper -> give_line(line_n);

	if (input_file -> source_type == COLOR256IMG) {
	     byte2RGBlines(curline, tmp_line,input_file -> width,input_file -> palette);
	     mapper->process_line(tmp_line,
			&tmp_line[input_file->width],
			&tmp_line[2*input_file->width],
			out_line);
	     }
	else mapper->process_line(curline,
			&curline[input_file->width],
			&curline[2*input_file->width],
			out_line);
	if (view_pass2)	putrow((rowptr)out_line,0,line_n,input_file -> width);
}

void refresh(int i,BOOL pass2) {
if (pass2) i = 30 + (i*7)/10;
printf("%3d%%",i);
printf("\b\b\b\b");
}
