/* Format conversion test --------------------------
 * manage to convert BMP (DIB),GIF,TGA and PCX files one into another
 * USAGE:
 *  convert sourcefile destfile [color option]
 *  color option defines desirable type of output image
 * 	(normally output image has same type as input)
 *  possiible values of option:
 *  M - mono image
 *  G - gray image(256 shadows)
 *  16 - 16 colors image
 *  256 - 256 colors image
 *  H - High color image(32000 colors)
 *  T - True color image(24 bits)
 * for example :	 convert world.tga world.gif 256
*/

#pragma -ml

#include <stdlib.h>
#include <string.h>
#include <stdlib.h>
#include <dir.h>
#include <mem.h>
#include <alloc.h>
#include <dos.h>

#include "formats.h"
#include "formconv.h"
#include "graph.h"
#include "quant.h"

extern unsigned _stklen = 15000;
graph_file_reader *grf_r;
graph_file_saver *grf_s;
BYTE *dest,*dest2,*tmp;
BOOL change_order,pass2_flag = FALSE;
image_keeper *keeper;
BGRpalette BGRpal;

void convert_2lowcolors(int colors) {
	int i;
	color_reductor *reductor = NULL;

	// if  (grf_s->dest_type == COLOR16IMG)  colors = 16; else colors = 255;

	reductor = new 	color_reductor (grf_r, FALSE,  colors);
	if (!reductor->initialized) {
		printf("Not enough memory!");
		exit(0);
		}

	if (reductor-> pass1() < 0 ) return;

	grf_s->update_palette(&reductor->my_colormap);


	for( i = 0; i < grf_s->height; i++) {
		reductor->process_line(grf_s->line_n,dest);
		grf_s->put_stand_line(dest);
		refresh((i*100L)/grf_s->height,TRUE);
		}
	delete reductor;
}

BYTE * get_next_line() {
	 if (!change_order) {
		grf_r->get_stand_line();
		return(grf_r->standart_line);
		}
	  else return(keeper -> give_line(grf_s -> line_n));
}

void justcycle() {
int i;
	for(i = 0; i < grf_r->height; i++) {
		tmp = get_next_line();
		grf_s->put_stand_line(tmp);
		refresh(((long)i*100)/grf_r->height,pass2_flag);
	}
}

void main(int argc, char *argv[])
{
int j,i;
BYTE v,white;
BOOL need_quant;

BGRpalette pal;
BGRpalette *palptr;
color_mapper_gray  *gray_mapper;
char s[80];

image_type dest_type;
image_extention_type dest_ext_type;
int  dest_colors = 0;


printf("Image file format conversion. By ^Z 1994.\n");
printf("Supported formats : TGA,PCX,GIF,Windows BMP(DIB)\n");

if (coreleft() < 64000) {
	printf("Not enough memory!");
	exit(0);
	}

if(argc < 3) {
printf("usage : convert sourcefile destfile [color option].\n");
printf("color option :\n");
printf("               M - mono image\n");
printf("               G - gray image(256 shadows)\n");
printf("               16 - 16 colors image\n");
printf("               256 - 256 colors image\n");
printf("               H - High color image(32000 colors)\n");
printf("               T - True color image(24 bits)\n");
printf("               Pxxx - xxx colors in 256 color palette\n");
printf("               	       where 8 <= xxx <= 256 \n");

printf("for example :\n");
printf("              convert world.tga world.gif 256\n");
exit(0);
}
grf_r = open_image_file(argv[1]);
if (grf_r==NULL) {printf("%s\n",format_file_error_report()); exit(0);}


dest_type = grf_r->source_type;
dest_ext_type = grf_r->source_ext_type;
if (argc >= 4) {
   strcpy(s,argv[3]);
   strlwr(s);
   dest_ext_type = NO_EXTENTION;
   if  (!memcmp(s,"m",2))  dest_type = MONOIMG;
   else if  (!memcmp(s,"g",2)) {dest_type = GRAY256IMG; dest_ext_type = GRAY_PALETTE;}
   else if  (!memcmp(s,"16",3)) dest_type = COLOR16IMG;
   else if  (!memcmp(s,"256",4)) dest_type = COLOR256IMG;
   else if  (!memcmp(s,"h",2)) { dest_type = TRUECOLORIMG; dest_ext_type = RGB15;}
   else if  (!memcmp(s,"t",2)) { dest_type = TRUECOLORIMG; dest_ext_type = BGR24;}
   else if  (!memcmp(s,"p",1)) { dest_type = COLOR256IMG; dest_colors = atoi(s+1);
				 dest_colors = min(256,max(dest_colors,8));
				}
}
if (dest_type == GRAY256IMG) {
	for(i=0;i<256;i++) {
		pal[i].r = i;
		pal[i].g = i;
		pal[i].b = i;
	};
	palptr = &pal;
}
else palptr = &grf_r->palette;


grf_s = open_saver_file(argv[2],
		dest_type, dest_ext_type,
		grf_r->width,grf_r->height,
		palptr,
		grf_r->top_to_bottom,
		grf_r->left_to_right);
if (grf_s==NULL) {printf("%s\n",format_file_error_report()); exit(0);}

minheap = 110000L;

printf("progress : ");
if (!EXM_initialized)  manager_initialize();

if (dest_colors == 0)
	switch (grf_s->dest_type) {
	case COLOR16IMG: dest_colors = 16;
		    break;
	case COLOR256IMG: dest_colors = 256;
		    break;
	}

need_quant =  ((grf_r->source_type == COLOR256IMG) && (grf_s->dest_type == COLOR16IMG) ||
	(grf_r->source_type == COLOR256IMG) && (grf_s->dest_type == COLOR256IMG) && (dest_colors < 256) ||
	(grf_r->source_type == TRUECOLORIMG) && (grf_s->dest_type == COLOR16IMG) ||
	(grf_r->source_type == TRUECOLORIMG) && (grf_s->dest_type == COLOR256IMG));

change_order = (grf_s->top_to_bottom ^ grf_r->top_to_bottom);
if (change_order && !need_quant) {
	keeper = new image_keeper(UNKNOWN,
			grf_r -> source_type,
			grf_r -> width,
			grf_r -> height,
			grf_r -> palette,
			grf_r ->top_to_bottom);

	for(i = 0; i <  grf_r->height; i++) {
		if (grf_r->get_stand_line() < 0) return;
		keeper->store_line(grf_r->standart_line);
		refresh((i*30L)/grf_r->height,FALSE);
	}
	pass2_flag = TRUE;
}

if ((!need_quant) && (grf_r->source_type == grf_s->dest_type)) {
	justcycle();
	goto finish; // ----------------> all done
	}

dest = new BYTE [3*(grf_r->width+1)]; // we'll need it

if (need_quant) {
	convert_2lowcolors(dest_colors);
	goto finish; // ----------------> all done
	}

if (grf_s->dest_type == MONOIMG ) {
	dest2 = new BYTE[grf_r->width+1];
	gray_mapper = new color_mapper_gray(2, grf_r->width,255);
	if ((gray_mapper == NULL) || (!gray_mapper->initialized)) goto finish;

	switch(grf_r->source_type) {
	case COLOR16IMG:
	case COLOR256IMG:  // byte per pixel to byte per pixel
		palette2gray(grf_r->palette, pal);
		grf_s->update_palette(&pal);
	}

	for(i = 0; i < grf_r->height;  i++) {
		switch(grf_r->source_type) {
		case COLOR16IMG:
		case COLOR256IMG:  // byte per pixel to byte per pixel
			tmp = get_next_line();
			paletted2gray(tmp, dest, grf_r->width, pal);
			break;

		case GRAY256IMG:
			dest = get_next_line();
			break;
		case  TRUECOLORIMG:
			tmp = get_next_line();
			RGBlines2gray(tmp, dest,grf_r->width);
		}

		gray_mapper->one_row(dest, dest2);
		monoGIF2mono(dest2, dest,grf_r->width);
		grf_s->put_stand_line(dest);
		refresh((i*100L)/grf_r->height,pass2_flag);
	}

	delete gray_mapper;
	delete dest2;
	goto finish; // ----------------> all done
}

if ((grf_s->dest_type == COLOR16IMG ) &&
    (grf_r->source_type ==  GRAY256IMG)) {
	dest2 = new BYTE[grf_r->width+1];
	gray_mapper = new color_mapper_gray(16, grf_r->width,255);
	if ((gray_mapper == NULL) || (!gray_mapper->initialized)) goto finish;

	for(i = 0; i < 16;  i++) {
		BGRpal[i].r =  BGRpal[i].g = BGRpal[i].b = i*16;
	}

	grf_s->update_palette(&BGRpal);

	for(i = 0; i < grf_r->height;  i++) {
		tmp = get_next_line();
		gray_mapper->one_row(tmp, dest2);
		grf_s->put_stand_line(dest2);
		refresh((i*100L)/grf_r->height,pass2_flag);
	}

	delete gray_mapper;
	delete dest2;
	goto finish; // ----------------> all done
}


 // and now, remaining cases:

	switch(grf_r->source_type) {
	case MONOIMG:  	// black & white
		if (grf_s->dest_type == GRAY256IMG) white = 255;
		else white = 1;
		for(i = 0; i < grf_r->height; i++) {
			tmp = get_next_line();
			if (grf_s->dest_type == TRUECOLORIMG)
				mono2RGB24(tmp,dest,grf_r->width);
			else 	mono2row(tmp,dest,white,grf_r->width);
			grf_s->put_stand_line(dest);
			refresh((i*100L)/grf_r->height,pass2_flag);
		       }
		break;

	case GRAY256IMG:  if  (grf_s->dest_type == COLOR16IMG) break;

	case COLOR16IMG:
	case COLOR256IMG:  // byte per pixel to byte per pixel
		switch(grf_s->dest_type) {

		// case COLOR16IMG:  gray to 16 - above
				// 256 to 16 see where  convert_2lowcolors is called

		case GRAY256IMG: // colors  to gray
			palette2gray(grf_r->palette, pal);
			grf_s->update_palette(&pal);

			if (grf_s->dest_ext_type != GRAY_PALETTE) {
			// convert to true value of gray, not just palette entry #
				for(i = 0; i < grf_r->height; i++) {
					tmp = get_next_line();
					paletted2gray(tmp, dest, grf_r->width, pal);
					grf_s->put_stand_line(dest);
					refresh((i*100)/grf_r->height,pass2_flag);
				}
				break;  // here all done!
			 }
				  //else  pass to next case entry {


		case COLOR256IMG:  // 16 to 256 ,  and see above
			justcycle();
			break;
		case TRUECOLORIMG: for(i = 0; i < grf_r->height; i++) {
					tmp = get_next_line();
					byte2RGBlines(tmp, dest,grf_r->width,grf_r->palette);
					grf_s->put_stand_line(dest);
					refresh((i*100L)/grf_r->height,pass2_flag);
				}
		}
		break;

	case TRUECOLORIMG:
		switch(grf_s->dest_type) {
		// to 256 or 16 see where  convert_2lowcolors is called
		case GRAY256IMG:
			grf_s->update_palette(&pal);
			for(i = 0; i < grf_r->height; i++) {
				tmp = get_next_line();
				RGBlines2gray(tmp, dest,grf_r->width);
				grf_s->put_stand_line(dest);
				refresh((i*100L)/grf_r->height,pass2_flag);
			}
			break;
		}
		break;
	}
finish:
	if (EXM_initialized)  manager_finish(TRUE);
	if (dest != NULL) delete dest;
	delete grf_s;
	delete grf_r;
if (change_order) delete keeper;
printf("OK.\n");
}
