; This program is the driver for SVGA 256 colors modes
; Compiled with TASM 3.1
; Written by Yuri Kiselev, 1994.

.model large,c

viewporttype struc
left	dw 0
top	dw 0
right	dw 0
bottom	dw 0
ends

public bank
assume ds : graphdata
graphdata segment public
	bank db 0
graphdata ends

extrn viewport : viewporttype
extrn videoselector : word
extrn linepattern : word
extrn fillpatternptr : dword
extrn patterntype : byte
extrn trflag : byte
extrn drawcolor : byte
extrn fillcolor : byte
extrn backcolor : byte
extrn bytesperline : word
extrn writemode : byte
extrn ClipRectArea
extrn VESAcallswitchbank

public proctables256

.data
proctables256	dd far ptr putpixel
		dd far ptr drawline
		dd far ptr bar
		dd far ptr putimage
		dd far ptr getimage
		dd far ptr imagesize
		dd far ptr putrow
		dd far ptr drawpolygon
		dd far ptr changeput
		dd far ptr getpixel
.code
.386

countcase	db 2
counterchanges	db 10
NormalPutCode	db 88h
XorPutCode	db 30h
AndPutCode	db 20h
OrPutCode	db 8
ChangeOffset	dw offset ChangePut1 + 1
		dw offset ChangePut2 + 1
		dw offset ChangePut3 + 1
		dw offset ChangePut4 + 1
		dw offset ChangePut5 + 1
		dw offset ChangePut6 + 1
		dw offset ChangePut7 + 1
		dw offset ChangePut8 + 1
		dw offset ChangePut9 + 1
		dw offset ChangePut12 + 1
		db 2
		db 89h
		db 31h
		db 21h
		db 9
		dw offset ChangePut10 + 1
		dw offset ChangePut11 + 1

changeput	proc far
		push si
		push di
		mov al,ah
		mov ah,0
		mov ch,countcase
		mov dx,cs
		mov es,dx
		lea si,counterchanges
		mov bx,5
		sub bx,ax
next_case:
		mov cl,es:[si]
		add si,ax
		mov dl,es:[si+1]
		add si,bx
next_ofs:
		mov di,word ptr es:[si]
		mov es:[di],dl
		add si,2
		dec cl
		jnz next_ofs

		dec ch
		jnz next_case

		pop di
		pop si
		ret
changeput	endp

wrpixx          EQU word ptr [bp+6]
wrpixy          EQU word ptr [bp+8]
Color		EQU byte ptr [bp+10]

PutPixel	proc far
		push bp
		mov bp,sp
		push ds
		push di
		mov ax,SEG graphdata
		mov ds,ax
		mov ax,wrpixy
		mov bx,wrpixx
		call PixelAddr
		mov al,Color
		mov es:[di],al
		pop di
		pop ds
		pop bp
		ret
PutPixel	endp

;
; procedure GetPixel(x,y : integer;RGB : pointer);
;
;

x		EQU word ptr [bp+6]
y		EQU word ptr [bp+8]

GetPixel	proc far
		push bp
		mov bp,sp
		push ds
		push di

		mov ax,SEG graphdata
		mov ds,ax

		mov ax,y
		mov bx,x
		call PixelAddr	; es:di -> Point address

		mov al,es:[di]	 ; reading color

		pop di
		pop ds
		pop bp
		ret
GetPixel	endp

;
;
;
;

LX1		EQU word ptr [bp+6]
LY1		EQU word ptr [bp+8]
LX2		EQU word ptr [bp+10]
LY2		EQU word ptr [bp+12]

Inc1		EQU word ptr [bp-2]
Inc2		EQU word ptr [bp-4]
Routine		EQU word ptr [bp-6]

DrawLine	proc  far
		push bp
		mov bp,sp
		sub sp,6
		push ds
		push di
		push si

		mov ax,SEG graphdata
		mov ds,ax

		push linepattern
		cld

; check for vertical line

		mov si,bytesperline ; initial y-increment

		mov cx,LX2
		sub cx,LX1
		jz VerticalLine
		jns X1toX2		; jump if x2 > x1
		mov ax,LY2		; exchange y1 and y2
		xchg ax,LY1
		mov LY2,ax

		mov ax,LX2		; exchange x1 and x2
		xchg ax,LX1
		mov LX2,ax

		neg cx
X1toX2:
		mov ax,LY2
		sub ax,LY1
		jz HorizontalLine	; jump if horizontal line

		jns positslope		; jump if slope is positive

		neg ax
		neg si			; negate increment for buffer interleave

			; select appropriate routine for slope of line
positslope:
		push si
		mov Routine,offset LoSlopeLine
		cmp ax,cx
		jle L04

		mov Routine,offset HiSlopeLine
		xchg ax,cx

			; calculate initial decision variable and increments
L04:
		shl ax,1
		mov Inc1,ax		; Inc1 = 2 * dy
		sub ax,cx
		mov si,ax
		sub ax,cx
		mov Inc2,ax		; Inc2 = 2 * (dy - dx)

		mov ax,LY1
		mov bx,LX1
		call PixelAddr

		inc cx			; CX = # of pixels to draw

		pop bx			; BX = y-increment
		jmp Routine		; jump to appropriate routine for slope

				; routine for vertical lines
VerticalLine:
		mov ax,LY1
		mov cx,LY2
		sub cx,ax	; CX = dy
		jge Y1toY2	; jump if dy >= 0

		neg cx
		mov ax,LY2
Y1toY2:
		inc cx		; CX = # of pixels to draw
		mov bx,LX1

		call PixelAddr
		mov al,drawcolor
		mov bx,linepattern
		cmp bx,0ffffh
		jne vertpatln
		; Normal put
ChangePut1:
		mov es:[di],al   ; updating byte
		add di,bytesperline   ; increment to next line
		jnc short no_inc_bank
		inc bank
		call far ptr VESAcallswitchbank
no_inc_bank:
		loop ChangePut1

		jmp linedone
vertpatln:
		ror bx,1
		jnc Skippix
ChangePut2:
		mov es:[di],al		; updating byte
Skippix:
		add di,bytesperline	; increment y
		jnc short no_inc_bankpat
		inc bank
		call far ptr VESAcallswitchbank
no_inc_bankpat:
		loop vertpatln

		jmp linedone
					; routine for horizontal lines
HorizontalLine:
		mov ax,LY1
		mov bx,LX1
		call PixelAddr

		inc cx
		mov al,drawcolor
		mov ah,al
		mov bx,linepattern
		cmp bx,0ffffh
		jne XOR_put1

		cmp writemode,0
		jne XOR_put1

		add di,cx
		jc short div_Line
		sub di,cx
		shr cx,1
		rep stosw	;  updating
		adc cl,0
		rep stosb
		jmp linedone
div_Line:
		sub di,cx
		mov bx,cx
		xor cx,cx
		sub cx,di
		sub bx,cx
		shr cx,1
		rep stosw
		rcl cl,1
		rep stosb
		inc bank
		call far ptr VESAcallswitchbank
		mov cx,bx
		shr cx,1
		rep stosw
		rcl cl,1
		rep stosb

		jmp linedone
XOR_put1:
		mov si,cx
		add di,cx
		jc short div_Linex
		sub di,cx
next_pixl:
		ror bx,1
		jnc skippix1
ChangePut3:
		mov es:[di],al
skippix1:
		inc di
		loop next_pixl
skip_byte4x:
		jmp linedone
div_Linex:
		sub di,cx
		mov dx,cx
		xor cx,cx
		sub cx,di
		sub dx,cx
		push dx
next_pixl1:
		ror bx,1
		jnc skippix2
ChangePut4:
		mov es:[di],al
skippix2:
		inc di
		loop next_pixl1
skip_byte5x:
		inc bank
		call far ptr VESAcallswitchbank
		pop cx
next_pixl2:
		ror bx,1
		jnc skippix3
ChangePut5:
		mov es:[di],al
skippix3:
		inc di
		loop next_pixl2
skip_byte6x:
		jmp linedone
; routine for dy <= dx (slope <= 1)     ; ES:DI -> video buffer
					; BX = y-increment
					; CX = counter pixels to draw
					; SI = decision variable
LoSlopeLine:
		mov al,drawcolor
		or bx,bx
		xchg bx,linepattern
		js short negIncY
L11:
		ror bx,1
		jnc Skippix4
ChangePut6:
		mov es:[di],al
Skippix4:
		add di,1
		jz short inc_bankline
		or si,si		; test sign of d
		jns short L12		; jump if d >= 0

		add si,Inc1		; d = d + Inc1
		loop L11
		jmp linedone
inc_bankline:
		inc bank
		call far ptr VESAcallswitchbank

		or si,si		; test sign of d
		jns short L12		; jump if d >= 0

		add si,Inc1		; d = d + Inc1
		loop L11

		jmp linedone
L12:
		add si,Inc2		; d = d + Inc2
		add di,linepattern	; increment y
		jnc short no_incb
		inc bank
		call far ptr VESAcallswitchbank
no_incb:
		loop L11
		jmp linedone
negIncY:
		ror bx,1
		jnc Skippix5
ChangePut7:
		mov es:[di],al
Skippix5:
		add di,1
		jz short inc_bankline1
		or si,si		; test sign of d
		jns short L122		; jump if d >= 0

		add si,Inc1		; d = d + Inc1
		loop negIncY
		jmp linedone
inc_bankline1:
		inc bank
		call far ptr VESAcallswitchbank

		or si,si		; test sign of d
		jns short L122		; jump if d >= 0

		add si,Inc1		; d = d + Inc1
		loop negIncY
		jmp linedone
L122:
		add si,Inc2		; d = d + Inc2
		add di,linepattern	; increment y
		jc short no_incbne
		dec bank
		call far ptr VESAcallswitchbank
no_incbne:
		loop negIncY

		jmp linedone

; routine for dy > dx (slope > 1)       ; ES:DI -> video buffer
					; BX = y-increment
					; CX = #pixels to draw
					; SI = decision variable
HiSlopeLine:
		mov al,drawcolor
		inc bx
		js short negIncY1
L21:
		ror linepattern,1
		jnc Skippix6
ChangePut8:
		mov es:[di],al
Skippix6:
		add di,bx		; increment y
		jnc short no_incb1
		inc bank
		call far ptr VESAcallswitchbank
no_incb1:
		or si,si		; test sign of d
		jns short L23		; jump if d >= 0

		add si,Inc1		; d = d + Inc1
		sub di,1
		jc short dec_bnk_obl
		loop L21
		jmp short linedone
dec_bnk_obl:
		dec bank
		call far ptr VESAcallswitchbank
		loop L21
		jmp short linedone
L23:
		add si,Inc2		; d = d + Inc2
		loop L21
		jmp short linedone
negIncY1:
		ror linepattern,1
		jnc Skippix7
ChangePut9:
		mov es:[di],al
Skippix7:
		add di,bx		; increment y
		jc short no_incb122
		dec bank
		call far ptr VESAcallswitchbank
no_incb122:
		or si,si		; test sign of d
		jns short L213		; jump if d >= 0

		add si,Inc1		; d = d + Inc1
		sub di,1
		jc short dec_bnk_obl1
		loop negIncY1
		jmp short linedone
dec_bnk_obl1:
		dec bank
		call far ptr VESAcallswitchbank
		loop negIncY1
		jmp short linedone
L213:
		add si,Inc2		; d = d + Inc2
		loop negIncY1
linedone:
		pop linepattern
		pop si
		pop di
		pop ds
		mov sp,bp
		pop bp
		ret
DrawLine 	endp

;
; procedure Bar(x,y,x1,y1 : integer);
;
;

XO		EQU word ptr [bp+6]
YO		EQU word ptr [bp+8]
BarX2		EQU word ptr [bp+10]
BarY2		EQU word ptr [bp+12]

XOtop		EQU word ptr [bp-2]
YOtop		EQU word ptr [bp-4]
begin_line	EQU word ptr [bp-6]
end_line	EQU word ptr [bp-8]
begin_bit	EQU word ptr [bp-10]
end_bit		EQU word ptr [bp-12]
cntrows		EQU word ptr [bp-18]
cntbyte		EQU word ptr [bp-20]
IncNextLine	EQU word ptr [bp-22]
FillLine	EQU word ptr [bp-24]
PatternOfs      EQU word ptr [bp-26]

Bar		proc far
		push bp
		mov bp,sp
		sub sp,34
		push ds
		push di
		push si

		mov ax,SEG graphdata
		mov ds,ax
		cld

		mov si,YO
		mov di,XO
		mov cx,BarX2
		sub cx,di
		inc cx
		mov dx,BarY2
		sub dx,si
		inc dx
		call far ptr ClipRectArea		; lines in window end_line-begin_line
		jns barvis
		jmp barinvis		; sign if bar is unvisible
barvis:
		mov ax,YOtop
		mov bx,XOtop
		call PixelAddr

		mov ax,end_line
		sub ax,begin_line
		inc ax
		mov cntrows,ax

		mov ax,end_bit
		sub ax,begin_bit
		inc ax
		mov cntbyte,ax
		mov bx,BytesPerLine
		sub bx,ax
		mov IncNextLine,bx

		cmp patterntype,1
		je solidfl
		mov ax,offset PatternLine
		cmp trflag,1
		jne no_tr_patt
		mov ax,offset GlassLine
no_tr_patt:
		mov FillLine,ax
		mov es,videoselector
		mov ah,backcolor
		mov al,fillcolor
		lgs si,fillpatternptr
		mov bx,si
		mov PatternOfs,bx

		mov cx,XOtop
		and cl,7
		jz no_shiftpat
		mov ch,8
next_patbyte:
		rol byte ptr gs:[si],cl
		inc si
		dec ch
		jnz next_patbyte
		sub si,8
no_shiftpat:
		mov cx,YOtop
		and cx,7
		add si,cx
		add bx,8
next_rowbarp:
		mov cx,cntbyte
		call FillLine
		inc si
		cmp si,bx
		jne no_endpat
		mov si,PatternOfs
no_endpat:
		add di,IncNextLine
		jnc short no_incbank
		inc Bank
		call far ptr VESAcallswitchbank
no_incbank:
		dec cntrows
		jnz next_rowbarp

		mov si,PatternOfs
		mov cx,XOtop
		and cl,7
		jz no_shiftpat1
		mov ch,8
next_patbyte1:
		ror byte ptr gs:[si],cl
		inc si
		dec ch
		jnz next_patbyte1
no_shiftpat1:
		jmp barinvis
solidfl:
		mov ax,offset FastFillLine
		mov FillLine,ax
		cmp writemode,0			; NormalPut
		je norm_put
		mov ax,offset NoNormFillLine
		mov FillLine,ax
norm_put:
		mov al,fillcolor
		mov ah,al
		mov bx,IncNextLine
		mov si,cntbyte
next_rowbar:
		mov cx,si
		call FillLine
		add di,bx
		jnc short no_incbank1
		inc Bank
		call far ptr VESAcallswitchbank
no_incbank1:
		dec cntrows
		jnz next_rowbar
barinvis:
		pop si
		pop di
		pop ds
		mov sp,bp
		pop bp
		ret
Bar	  	endp

; fill line solid color (fastest)
FastFillLine	proc near
		add di,cx
		jc short inc_bank
		sub di,cx
		shr cx,1
		rep stosw
		adc cl,0
		rep stosb
		ret
inc_bank:
		mov dx,cx
		sub di,cx
		xor cx,cx
		sub cx,di
		sub dx,cx
		shr cx,1
		rep stosw
		rcl cl,1
		rep stosb
		mov cx,dx
		inc bank
		call far ptr VESAcallswitchbank
		shr cx,1
		rep stosw
		adc cl,0
		rep stosb
		ret
FastFillLine	endp

	; fill line solid color with XOR,OR,AND write mode

NoNormFillLine	proc near
		add di,cx
		jc short inc_banknn
		sub di,cx
		jmp NoNormPutRow
inc_banknn:
		mov dx,cx
		sub di,cx
		xor cx,cx
		sub cx,di
		sub dx,cx
		call NoNormPutRow
		mov cx,dx
		inc Bank
		call far ptr VESAcallswitchbank
		jmp NoNormPutRow
NoNormFillLine	endp

; call only from NoNormFillLine

NoNormPutRow	proc near
		shr cx,1
		jcxz chk_byte
		jc odd_cntbyte
ChangePut10:
next_word:
		xor es:[di],ax
		add di,2
		loop next_word
		ret
odd_cntbyte:
ChangePut11:
		xor es:[di],ax
		add di,2
		loop odd_cntbyte
chk_byte:
ChangePut12:
		xor es:[di],al
		inc di
		ret
NoNormPutRow	endp

PatternLine	proc near
		mov dh,gs:[si]
next_point:
		rol dh,1
		jc fill_col
		mov es:[di],ah
		inc di
		jz inc_bankpl
		loop next_point
		ret
fill_col:
		mov es:[di],al
		inc di
		jz inc_bankpl
		loop next_point
		ret
inc_bankpl:
		inc bank
		call far ptr VESAcallswitchbank
		loop next_point
		ret
PatternLine	endp

GlassLine	proc near
		mov dh,gs:[si]
gnext_point:
		rol dh,1
		jc gfill_col
		inc di
		jz inc_bankplg
		loop next_point
		ret
gfill_col:
		mov es:[di],al
		inc di
		jz inc_bankplg
		loop gnext_point
		ret
inc_bankplg:
		inc bank
		call far ptr VESAcallswitchbank
		loop next_point
		ret
GlassLine	endp

;
;
; returned : longint
;

iX1		EQU word ptr[bp+6]
iY1		EQU word ptr[bp+8]
iX2		EQU word ptr[bp+10]
iY2		EQU word ptr[bp+12]

ImageSize	proc far
		push bp
		mov bp,sp

		mov ax,iX2
		sub ax,iX1
		inc ax
		mov dx,iY2
		sub dx,iY1
		inc dx
		mul dx
		add ax,10
		adc dx,0
		; DX,AX - required number of bytes

		pop bp
		ret
ImageSize	endp

; Name: GetImage
;
; Caller: GetImage(X1,Y1,X2,Y2:integer;buffer:pointer);

HeaderLength	EQU 10

IX1		EQU word ptr [bp+6]
IY1		EQU word ptr [bp+8]
IX2		EQU word ptr [bp+10]
IY2		EQU word ptr [bp+12]
buffer  	EQU dword ptr [bp+14]

real_len1	EQU word ptr [bp-4]
screen_len1 	EQU word ptr [bp-6]
cntline		EQU word ptr [bp-8]

GetImage   proc far
		push bp
		mov bp,sp
		sub sp,8
		push ds
		push di
		push si

		mov ax,SEG graphdata
		mov ds,ax

		mov ax,IY1
		mov bx,IX1
		call PixelAddr	; es:di -> Point address
		mov si,di

		les di,buffer
		mov word ptr es:[di+0],0
		mov word ptr es:[di+2],0
		mov word ptr es:[di+8],0

		mov ax,IY2
		sub ax,IY1
		mov es:[di+6],ax
		inc ax
		mov cntline,ax

		mov ax,IX2
		sub ax,IX1
		mov es:[di+4],ax
		inc ax
		mov real_len1,ax
		mov bx,bytesperline
		sub bx,ax
		add di,HeaderLength
		cld
		mov ds,videoselector
next_rowbargi:
		mov cx,real_len1
		add si,cx
		jc short inc_bankgi

		sub si,cx
		shr cx,1
		rep movsw
		adc cl,0
		rep movsb

		add si,bx
		jnc short no_incbankgi
		push ds
		mov ax,SEG graphdata
		mov ds,ax
		inc bank
		call far ptr VESAcallswitchbank
		pop ds
no_incbankgi:
		dec cntline
		jnz next_rowbargi
		jmp short endcopy
inc_bankgi:
		mov dx,real_len1
		sub si,cx
		xor cx,cx
		sub cx,si
		sub dx,cx
		shr cx,1
		rep movsw
		adc cl,0
		rep movsb
		mov cx,dx
		push ds
		mov ax,SEG graphdata
		mov ds,ax
		inc bank
		call far ptr VESAcallswitchbank
		pop ds
		shr cx,1
		rep movsw
		adc cl,0
		rep movsb
		add si,bx
		dec cntline
		jnz next_rowbargi
endcopy:
		pop si
		pop di
		pop ds
		mov sp,bp
		pop bp
		ret
GetImage   endp

; Name: PutImage
; Function: Show and clipping picture (from buffer)
;
; Caller: PutImage(XO,YO:integer;buffer:pointer);
;

DataBuffer 	EQU dword ptr [bp+10]
YO		EQU word ptr [bp+8]
XO		EQU word ptr [bp+6]

BPL		EQU word ptr [bp-24]
Width_picture	EQU word ptr [bp-26]

PutImage	proc far
		push bp
		mov bp,sp
		sub sp,26	; vars from bar
		push ds
		push di
		push si

		mov ax,SEG graphdata
		mov ds,ax

		mov ax,bytesperline
		mov BPL,ax

		les si,DataBuffer
		mov ax,es:[si+4]	; x2
		sub ax,es:[si]		; x1
		inc ax
		mov Width_picture,ax
		mov cx,ax
		mov dx,es:[si+6]	; y2
		sub dx,es:[si+2]	; y1
		inc dx
		mov si,YO
		mov di,XO

		call far ptr ClipRectArea
		js iminvis

		mov ax,YOtop
		mov bx,XOtop
		call PixelAddr

		mov ax,end_line
		sub ax,begin_line
		inc ax
		mov cntrows,ax

		lds si,DataBuffer
		mov ax,end_bit
		sub ax,begin_bit
		inc ax
		mov cntbyte,ax
		sub Width_picture,ax
		mov bx,BPL
		sub bx,ax
		cmp begin_line,0
		je imno_search_line
		mul begin_line
imno_search_line:
		add si,begin_bit
		add si,HeaderLength
next_rowbarpi:
		mov cx,cntbyte
		add di,cx
		jc short inc_bankpi
		sub di,cx
		shr cx,1
		rep movsw
		adc cl,0
		rep movsb
		add di,bx
		jnc short no_incbankpi
		push ds
		mov ax,SEG graphdata
		mov ds,ax
		inc bank
		call far ptr VESAcallswitchbank
		pop ds
no_incbankpi:
		add si,Width_picture
		dec cntrows
		jnz next_rowbarpi
		jmp short iminvis
inc_bankpi:
		mov dx,cx
		sub di,cx
		xor cx,cx
		sub cx,di
		sub dx,cx
		shr cx,1
		rep movsw
		adc cl,0
		rep movsb
		mov cx,dx
		push ds
		mov ax,SEG graphdata
		mov ds,ax
		inc bank
		call far ptr VESAcallswitchbank
		pop ds
		shr cx,1
		rep movsw
		adc cl,0
		rep movsb
		add di,bx
		add si,Width_picture
		dec cntrows
		jnz next_rowbarpi
iminvis:
		pop si
		pop di
		pop ds
		mov sp,bp
		pop bp
		ret
PutImage	endp

;
;
;
;
RowPtr		EQU dword ptr [bp+6]
XO		EQU word ptr [bp+10]
YO		EQU word ptr [bp+12]
Xcounter	EQU word ptr [bp+14]	; (X2 - X1 + 1)

PutRow		proc far
		push bp
		mov bp,sp
		sub sp,24
		push ds
		push si
		push di

		mov ax,SEG graphdata
		mov ds,ax

		mov si,YO
		mov di,XO
		mov cx,Xcounter
		mov dx,1
		call far ptr ClipRectArea
		js short InvisRow

		mov ax,YOtop
		mov bx,XOtop
		call PixelAddr

		mov ax,end_bit
		sub ax,begin_bit
		inc ax

		lds si,RowPtr
		add si,begin_bit
		mov cx,ax
		add di,cx
		jc inc_bankpr
		sub di,cx
		shr cx,1
		rep movsw
		adc cl,0
		rep movsb
		jmp short InvisRow
inc_bankpr:
		mov dx,ax
		sub di,cx
		xor cx,cx
		sub cx,di
		sub dx,cx
		shr cx,1
		rep movsw
		adc cl,0
		rep movsb
		mov cx,dx
		push ds
		mov ax,SEG graphdata
		mov ds,ax
		inc bank
		call far ptr VESAcallswitchbank
		pop ds
		shr cx,1
		rep movsw
		adc cl,0
		rep movsb
InvisRow:
		pop di
		pop si
		pop ds
		mov sp,bp
		pop bp
		ret
PutRow		endp

;
;
;
;

sizeonerow	EQU word ptr [bp+14]
Ymmin		EQU word ptr [bp+12]
YmmaxminusYmmin EQU word ptr [bp+10]
CurrOffset	EQU word ptr [bp+8]
SegXarray	EQU word ptr [bp+6]

XOtop		EQU word ptr [bp-2]
YOtop		EQU word ptr [bp-4]
CountX		EQU byte ptr [bp-6]
PolyFillLine	EQU word ptr [bp-8]
OfsBlmask	EQU word ptr [bp-10]
ControlMaskpl	EQU word ptr [bp-12]
Pattptr		EQU word ptr [bp-14]
PatternEndOfs	EQU word ptr [bp-16]
PatternOfs	EQU word ptr [bp-18]
PrevX		EQU word ptr [bp-20]


DrawPolygon	proc far
		push bp
		mov bp,sp
		sub sp,20
		push ds
		push di
		push si

		mov ax,SEG graphdata
		mov ds,ax

		xor bx,bx	; x = 0
		mov ax,Ymmin
		mov YOtop,ax
		call PixelAddr
		mov Ymmin,di

		cmp patterntype,1
		jne patt_fill
		jmp dpsolidfl
patt_fill:
		mov ax,offset PatternLine
		cmp trflag,1
		jne dpno_tr_patt
		mov ax,offset GlassLine
dpno_tr_patt:
		mov PolyFillLine,ax

		lgs ax,fillpatternptr
		mov PatternOfs,ax
		mov bx,YOtop
		and bx,7
		add bx,ax
		mov word ptr Pattptr,bx
		add ax,8
		mov PatternEndOfs,ax

		mov al,fillcolor
		mov ah,backcolor
		mov fs,SegXarray
nextfillLinep:
		xor di,di
		mov PrevX,di
		mov bx,CurrOffset
		mov cl,fs:[bx]
		shr cl,1	; div 2 pair X - coord
		mov countX,cl
		dec di
		or cl,cl
		jz no_putlinep
		mov di,Ymmin
next_lineinrowp:
		mov si,fs:[bx+1]	;X1
		or si,si
		js no_putpairp
		mov cx,fs:[bx+1]	; X1
		mov dx,si
		sub dx,PrevX
		jns short no_equpointsp
		xor dx,dx
		mov cx,fs:[bx+1]	; X1
		inc cx
		inc si
no_equpointsp:
		add di,dx
		jnc short no_incBankPolp
		inc bank
		call far ptr VESAcallswitchbank
no_incBankPolp:
		mov dx,fs:[bx+3]	; X2
		inc dx
		mov PrevX,dx
		sub dx,si
		and cl,7
		mov si,Pattptr
		rol byte ptr gs:[si],cl
		xchg cx,dx
		call PolyFillLine
		mov cl,dl
		ror byte ptr gs:[si],cl
no_putpairp:
		add bx,4
		dec CountX
		jnz next_lineinrowp
no_putlinep:
		mov cx,word ptr Pattptr
		inc cx
		cmp cx,PatternEndOfs
		jne no_endpatp
		mov cx,PatternOfs
no_endpatp:
		mov word ptr Pattptr,cx

		mov cx,bytesperline
		add Ymmin,cx
		jc incbf
		jnz no_incbf
incbf:
		or di,di
		jns no_incbf
		inc bank
		call far ptr VESAcallswitchbank
no_incbf:
		mov si,sizeonerow
		add CurrOffset,si
		dec YmmaxminusYmmin
		jnz nextfillLinep

		jmp end_drawpoly
					;--------------------------
dpsolidfl:
		mov ax,offset FastFillLine
		mov PolyFillLine,ax
		cmp writemode,0		;NormalPut
		je norm_putpl
		mov ax,offset NoNormFillLine
		mov PolyFillLine,ax
norm_putpl:
		mov al,fillcolor
		mov ah,al
		mov fs,SegXarray
nextfillLine:
		xor di,di
		mov PrevX,di
		mov bx,CurrOffset
		mov cl,fs:[bx]
		shr cl,1	; div 2 pair X - coord
		mov countX,cl
		dec di
		or cl,cl
		jz no_putline
		mov di,Ymmin
next_lineinrow:
		mov si,fs:[bx+1]	;X1
		or si,si
		js no_putpair
		mov dx,si
		sub dx,PrevX
		jns short no_equpoints
		xor dx,dx
		inc si
no_equpoints:
		add di,dx
		jnc short no_incBankPol
		inc bank
		call far ptr VESAcallswitchbank
no_incBankPol:
		mov cx,fs:[bx+3]	; X2
		inc cx
		mov PrevX,cx
		sub cx,si
		call PolyFillLine
no_putpair:
		add bx,4
		dec CountX
		jnz next_lineinrow
no_putline:
		mov cx,bytesperline
		add Ymmin,cx
		jc incbfp
		jnz no_incbfp
incbfp:
		or di,di
		jns no_incbfp
		inc bank
		call far ptr VESAcallswitchbank
no_incbfp:
		mov si,sizeonerow
		add CurrOffset,si
		dec YmmaxminusYmmin
		jnz nextfillLine
end_drawpoly:
		pop si
		pop di
		pop ds
		mov sp,bp
		pop bp
		ret
DrawPolygon	endp

; call : eax = Y
;	 ebx = X
;
; returned : es:di -> Point address
;		dx = 3ceh
;		di,bx = Xoffset
;

PixelAddr	proc near
		mov es,VideoSelector
		mul BytesPerLine
		add bx,ax	; offsetY+offsetX
		adc dl,0
		mov di,bx	; DI = offset pixel
		cmp dl,Bank
		je short no_swich_bank
		mov Bank,dl
		call far ptr VESAcallswitchbank
no_swich_bank:  ret
PixelAddr	endp
end