/* This program prints out the information about memory available
in the system, including conventional, upper, high, expanded, and
extended memory. It also reports about installed memory managers.

Written by Yuri Kiselev 1994
*/
#include<mem.h>
#include<dos.h>
#include<stdlib.h>
#include<stdio.h>

char DOSmajver,DOSminver;
char XMSmajver,XMSminver;
char XMSdrvmajver,XMSdrvminver;
char EMMver;
char DPMImajver,DPMIminver;
char VCPImajver,VCPIminver;
char DOSflags;
int DPMIflags;
unsigned int free_HMA_block,max_XMS,total_XMS,total_ext_mem;
unsigned int EMM_pages,EMM_free_pages;
unsigned int Conv_free,UMB_all,UMB_free,UMB_largest;
unsigned int max_conv_block;
void* xms_call;
char EMM_dev_name[8] = {"EMMXXXX0"};

void memwalk(void); // Searches the MCB chain

void main(void)
{
asm mov ax,3306h
asm int 21h		// Get DOS version
DOSmajver = _BL;
DOSminver = _BH;
DOSflags  = _DH;

printf("MS-DOS version %d.%d\n",DOSmajver,DOSminver);
asm mov ah,48h
asm mov bx,0ffffh
// Get the size of the largest available block in conventional memory
asm int 21h
max_conv_block = _BX;
memwalk();

printf("Free conventional memory: %ld bytes\n",(long)Conv_free << 4);
printf("Largest free block in conventional memory: %ld bytes\n",
	(long)max_conv_block << 4);

if (DOSmajver >= 5) {
	if (UMB_all != 0) {
		printf("Upper memory (UMB): %ld bytes\n",(long)UMB_all << 4);
		printf("Free UMB: %ld bytes\n",(long)UMB_free << 4);
		printf("Largest free UMB: %ld bytes\n",(long)UMB_largest << 4);
	}
	if (DOSflags & 0x10) {
		printf("DOS loaded in HMA\n");
		asm mov ax,4A01h
		asm int 2fh
		free_HMA_block = _BX;
		if (_BX != 0) printf("Free high memory: %d bytes\n",
			free_HMA_block);
	}
	asm mov ax,4300h
	asm int 2fh		// Check if XMS driver is installed
	if (_AL == 0x80) {
		asm mov ax,4310h
		asm int 2fh	// Get XMS driver entry point
		asm mov word ptr xms_call,bx
		asm mov word ptr xms_call+2,es
		asm mov ah,0
		asm call xms_call	// Get XMS version
		XMSmajver = _AH;
		XMSminver = _AL;
		XMSdrvmajver = _BH;
		XMSdrvminver = _BL;
		printf("XMS version %d.%d.",XMSmajver,XMSminver);
		printf("Driver version %d.%d\n",XMSdrvmajver,XMSdrvminver);
		asm mov ah,8
		asm call xms_call	// Get free XMS memory
		max_XMS = _AX;
		total_XMS = _DX;
		printf("XMS memory size %d Kb\n",total_XMS);
		printf("Largest XMS memory block is %d Kb\n",max_XMS);
	}
}
// Read extended memory size from CMOS
asm mov al,17h
asm out 70h,al
asm nop
asm in al,71h
asm mov byte ptr total_ext_mem,al
asm mov al,18h
asm out 70h,al
asm nop
asm in al,71h
asm mov byte ptr total_ext_mem+1,al
printf("Total extended memory %d Kb\n",total_ext_mem);

asm mov ah,88h
asm int 15h		// Get total extended memory size via Int 15h
printf("Int 15 total extended memory %d\n",_AX);

asm mov ah,35h
asm mov al,67h
asm int 21h		// Get Int 67h vector

// Check if EMM is installed
if (!memcmp(MK_FP(_ES,10),EMM_dev_name,8)) {
	asm mov ah,46h
	asm int 67h	// Get EMS version
	EMMver = _AL;
	printf("EMM version %d.%d\n",EMMver >> 4,EMMver & 0x0f);
	asm mov ah,42h
	asm int 67h	// Get page info
	EMM_pages = _DX;
	EMM_free_pages = _BX;
	printf("EMM total pages %d (%d Kb)\n",EMM_pages,EMM_pages << 4);
	printf("EMM unallocated pages %d (%d Kb)\n",EMM_free_pages,EMM_free_pages << 4);
	asm mov ax,0DE00h
	asm int 67h	// Check if VCPI is present
	if (_AH == 0) {
		VCPImajver = _BH;
		VCPIminver = _BL;
		printf("VCPI version %d.%d\n",VCPImajver,VCPIminver);
		asm mov ax,0DE03h
		asm int 67h	// Get free page count
		printf("VCPI number of free 4Kb pages is %d (%dKb)\n",_DX,_DX*4);
		}
}
asm mov ax,1687h
asm int 2fh	// Check if DPMI is installed
if (_AX == 0) {
	DPMImajver = _DH;
	DPMIminver = _DL;
	DPMIflags = _BX;
	printf("DPMI version %d.%d\n",DPMImajver,DPMIminver);
	if (DPMIflags & 1) printf("DPMI supported 32-bit\n");
	}
}

void memwalk() {
char link;
void *root_mcb;
	asm mov ax,5802h
	asm int 21h		// Get upper memory link
	asm mov link,al		// store upper-memory link
	asm mov bx,1
	asm mov ax,5803h
	asm int 21h		// set upper-memory link on

	asm mov ah,52h
	asm int 21h		// Get DOS List of Lists

	asm mov ax,es:[bx-2]	// AX = Root MCB segment
next_block:
	asm mov es,ax
	asm cmp ax,0A000h	// Check if a MCB is in UMB
	asm mov ax,es:[3]
	asm ja umb_block
	asm cmp word ptr es:[1],0	// Free MCB?
	asm jne end_block		// No
	asm add Conv_free,ax    	// Increase free conv. memory sum
	asm jmp end_block
umb_block:
	asm cmp word ptr es:[1],8	// System MCB?
	asm jne no_sysblock		// No
	asm cmp word ptr es:[8],4353h	// 'SC' - System code
	asm je end_block
no_sysblock:
	asm add UMB_all,ax		// Increase total UMBs size
	asm cmp word ptr es:[1],0	// Free MCB ?
	asm jne end_block		// No
	asm add UMB_free,ax		// Increase free upper memory size
	asm cmp ax,UMB_largest		// find largest free UMB
	asm jbe end_block
	asm mov UMB_largest,ax
end_block:
	asm mov bx,es
	asm add ax,bx
	asm inc ax			// AX = segment of the next block
	asm cmp byte ptr es:[0],'Z'	// last block ?
	asm jne next_block		// No

	asm xor bx,bx
	asm mov bl,link
	asm mov ax,5803h
	asm int 21h		// restore upper-memory link
}