/*
* This program is a simple expanded/extanded memory manger
* Written by Podvoysky E. & Kiselev J. CZ 1994.
*/

#include <mem.h>
#include <alloc.h>
#include "manager.h"
#define PAGESIZE 0x10000L
#define HANDLE  WORD
#define OWNER  WORD
#define SIZE_SHIFT 2
#define PAGESIZEDW (PAGESIZE >> SIZE_SHIFT)
#define PAGEALLOC (PAGESIZE >> 14)

typedef  struct {
	HANDLE handle;		// 0xFFFF - empty link
	int offset, size; 	// offset, size in DWORDs
	int next;   		// -1 means last link in the chain
	} block_link;

typedef  struct {
	OWNER owner;	// reserved for the resident version
	//long size,
	int first_link; // -1 - empty descriptor
	} userblock_descr;

typedef  struct {
	HANDLE handle;
	int offset; 	// in DWORDS
	} handle_descr;
// Free area is always at the top of a page
// offset is the start of the free area divided by (1<<SIZE_SHIFT)

/////////////////////////////////////////////////////////////////////
////////////////////////////////////// data structures
userblock_descr NEAR * user_blocks;
int  blocks_size = 32; // Actual array size (number of used blocks).
BYTE *blocks_states;	// For each 16 elements of user_blocks array
			// blocks_states holds the number of free elements
			// array size = blocks_size / 16
int first_possibly_free_block = 0;

block_link NEAR * links;  	// heap array
int links_size = 32,    // actual links' size
	last_link = 0;
BYTE *links_states;	// For each 16 elements of the array links
			// links_states holds the number of free elements
			// array size = links_size / 16

int first_possibly_free_link = 0;

#define STATES_SHIFT 4
// each "states" variable counts elements in 16-elements units
// make sure that array size increment is multiple of 16
#define SIZE_INC (1 << STATES_SHIFT)

handle_descr NEAR * handles;	// sorted in descending order of free size
			// or ascending order of offsets
			// completely full handles are not skipped!
int handles_size = 16, handles_num;

int EXM_error = 0;
BOOL EXM_initialized = FALSE, is_XMS;
int handles_remaind = 0;

/////////////////////////// middle - level methods /////////////////////
///////////////////////////////////////////////////////////////////////
HANDLE our_handle;	//for our internal use
USER_HANDLE our_block;

int get_block(); 	// find a free block, mark as filled, etc.;
			// -1 if error

int free_link(int index); // mark link as free, and free memory block;
	// returns -1 if error
int get_link(); 	// find a free link, mark as filled, etc.;
			// -1 if error
int continue_chain(int &first_link, int  &cur_link);
	// get a link, add it to the chain which starts with first_link
	// you must initially set first_link to -1 !
	// do not forget to fill links[cur_link] with neccesary stuff
int link_with_new_page(int &first_link, int  &cur_link, long size);
	// tries to allocate a new page and create a link which
	// reffers to it
	// returns 0 if Ok
	//	   -1 if error
	// 	   1  if a new page can not be allocated
int find_block_offset(int &link, SIZE_TYPE &offset);
	// input - the first link in the chain, offset in the block
	// output - the link to subblock, containing offset byte of block
	// and local offset in this block (in bytes, not DWORDS!)


int find_handle_place(WORD offset);
	// returns the first PAGE # with .offset greater than offset

int insert_handle(handle_descr handle); // insert according to the sorting
int insert_handle_by_no(handle_descr handle, int where);
	// checks for array size
void delete_handle(int where);
	// just delete from the array, do not free memory

int make_tail(int &first_link, int  &cur_link, long size );
// allocation strategy
// size already in DWORDs!
// finds where is it possible to settle a piece of memory
// (normally size < PAGESIZE)
// returns the first link number in the created chain, -1 if error
// doesn't free created subchain if an error has occured

int free_chain(int first_link);

int resize_array(BYTE * &ptr, WORD recsize, // gets old ptr, returns new
		int &size, WORD incsize,    // gets old size (in records),
						// returns new
		BYTE fill); 		    // fill a new part of array with ...
	// returns -1 if error

////////////////////////////////////////////////////////////////////////////
//  implementation of user-oriented routines
////////////////////////////////////////////////////////////////////////////

USER_HANDLE EXM_alloc(SIZE_TYPE size) {
	int first_link = -1, cur_link = -1;
	HANDLE  handle;
	int i;
	handle_descr h_descr;

	EXM_error = 0;
	if (!EXM_initialized) {
		EXM_error = 10;
		return(NULL_HANDLE);
	}

	if (size <= 0) {
			EXM_error = 5;
			return(NULL_HANDLE);
	}


	size = (size + ((1 << SIZE_SHIFT)-1)) >> SIZE_SHIFT;

	while (size >= PAGESIZEDW) { // allocate full pages
		i = link_with_new_page(first_link, cur_link, PAGESIZEDW);
		if (i < 0) goto l_error;
		if (i > 0) break; // no more free pages, change strategy
		size -= PAGESIZEDW;
	}

	if (make_tail(first_link,cur_link,size) < 0) goto l_error;

	i = get_block();

	if (i < 0) goto l_error;
	else {
		// user_blocks[i].owner = ...;
		user_blocks[i].first_link = first_link;
		return(i);
	}

l_error:
	free_chain(first_link);
	return(NULL_HANDLE);
}


int EXM_free(USER_HANDLE handle) {
	int i;
	EXM_error = 0;
	if (!EXM_initialized) {
		EXM_error = 10;
		return(NULL_HANDLE);
	}
	if ((handle >= blocks_size) || (user_blocks[handle].first_link == -1))
		{EXM_error = 3; return(-1);}
	//??? check OWNER
	i = free_chain(user_blocks[handle].first_link);
	user_blocks[handle].first_link = -1; // mark block as free

	blocks_states[ handle >> STATES_SHIFT]++;
	first_possibly_free_block = min( first_possibly_free_block, handle);
	return(i);
};


int EXM_move(void FAR *conv,  USER_HANDLE handle,
	     SIZE_TYPE offset, SIZE_TYPE size, BYTE direction) {
	     // direction 1 - to conv., 0 - to ex.

	int link_no,i;
	SIZE_TYPE sz,ofs,ofs1;
	block_link *link;
	BOOL error;
	BYTE tmp[2];

	EXM_error = 0;
	if (!EXM_initialized) {
		EXM_error = 10;
		return(NULL_HANDLE);
	}

	link_no = user_blocks[handle].first_link;
	if ((handle >= blocks_size)  || (link_no == -1))
		{EXM_error = 3; return(-1);}
	if (find_block_offset(link_no, offset) < 0) return(-1);

	while (link_no >= 0)  {
		link = links+link_no; // &links[link_no]
		sz =  ((long)link->size << SIZE_SHIFT) - offset;
		sz =  min(sz, size);
		ofs =  offset+ (link->offset << SIZE_SHIFT);
		if (is_XMS && (sz & 1)) { // bloody XMS demands even size!!!
		    if (sz == 1) {
			   // it may be last or first byte in page, so....
			    ofs1 = ofs & 0xFFFE;
			    // even offset, word does lie whithin a page
			    error =  move_ext_mem(link->handle,ofs1,
						2, tmp, 1) < 0;
			    // move ex. memory contents to tmp
			    if  (direction) ((BYTE *)conv)[0] = tmp[ofs & 1];
			    else {
				tmp[ofs & 1] = ((BYTE *)conv)[0];
				error =  error ||  move_ext_mem(link->handle,ofs1,
					2, tmp, 0) < 0;
			    }
		    }
		    else {
			    error =  move_ext_mem(link->handle,ofs,
				sz-1, conv, direction) < 0;
			    error = error || (move_ext_mem(link->handle,ofs + sz-2,
				2, (char far *)conv+sz-2, direction) < 0);
		    }

		} // special XMS
		else  error =  move_ext_mem(link->handle,ofs,
				sz, conv, direction) < 0;

		if (error) {
			EXM_error = 100;
			return(-1);
		}

		offset = 0;
		size -= sz;
		if (size <= 0) return(0);
		(char far *)conv += sz;
		link_no = link->next;
	}

	EXM_error = 4;
	return(-1);
}

int EXM_move_to_conv(void FAR *dest, USER_HANDLE source_handle,
			SIZE_TYPE offset, SIZE_TYPE size) {
	return EXM_move(dest, source_handle, offset, size, 1);
}

int EXM_move_to_ext(USER_HANDLE dest_handle, void FAR *source,
			SIZE_TYPE offset, SIZE_TYPE size) {
	return EXM_move(source, dest_handle, offset, size, 0);
}


int manager_initialize() {   // alloc arrays, init low level
	int state_size, rez, i;
	long l;
	WORD *pg;

	EXM_error = 0;
	if (EXM_initialized) return(0);

	rez = init_ext_mem();
	switch (rez)  {
	case -1:  return(-1);
	case 0:  is_XMS = TRUE;
		 break;
	case 1:  is_XMS = FALSE;
	}


	i = ((long)get_free_ext_mem() << 10) / PAGESIZE;

	pg = new WORD[i+1];
	for (handles_remaind = 0; handles_remaind < i; handles_remaind++) {
		pg[handles_remaind] = alloc_ext_mem(PAGEALLOC); // try to get new page
		if (pg[handles_remaind] == NULL_HANDLE) break; // no more pages!
	}
	for (i = handles_remaind-1; i >= 0; i--) free_ext_mem(pg[i]);
	delete pg;

	handles = new handle_descr[handles_size];

	state_size = blocks_size >> STATES_SHIFT;
	blocks_states = new BYTE[state_size];
	user_blocks = new userblock_descr[blocks_size];

	state_size = links_size >> STATES_SHIFT;
	links_states = new BYTE[state_size];
	links = new block_link[links_size];

	if (links == NULL) {
		EXM_error = 1;
		return (EXM_error);
	}

	memset(user_blocks,0xFF,sizeof(userblock_descr)*blocks_size );
	memset(blocks_states,(1 << STATES_SHIFT),state_size);
	memset(handles,0xFF,sizeof(handle_descr)*handles_size );
	memset(links_states,(1 << STATES_SHIFT),state_size);
	memset(links,0xFF,sizeof(block_link)*links_size);

	EXM_initialized = TRUE;

	our_block = EXM_alloc(PAGESIZE);
	our_handle = links[0].handle;
	if  (our_block ==  NULL_HANDLE) {
		EXM_initialized = FALSE;
		EXM_error = 2;
		return(-1);
		}
	return( 0);
};

void manager_finish(BOOL dispose_conv) {
	int i;

	if (!EXM_initialized) return;
	for (i = 0; i < handles_num; i++)
		if (free_ext_mem(handles[i].handle)) {
			EXM_error = 100;

		}
	if  (dispose_conv) {
		delete  user_blocks;
		delete  links;
		delete  blocks_states;
		delete  links_states;
		delete   handles;
	}
	EXM_initialized = FALSE;
}

long EXM_coreleft() {
	int i;
	long l = 0;

	if (!EXM_initialized) return (0);
	for (i = 0; i < handles_num; i++)
		l += PAGESIZEDW - handles[i].offset;
	l <<= SIZE_SHIFT;

	// l += (long)get_free_ext_mem() << 10; //in bytes
	l += (long)handles_remaind * PAGESIZE; //in bytes

	return (l);
}

int EXM_to_EXM(USER_HANDLE dest_handle, SIZE_TYPE dest_offset,
	       USER_HANDLE source_handle, SIZE_TYPE source_offset,
	       SIZE_TYPE size) {

	int link_no1,link_no2,i;
	block_link *link1, *link2;
	SIZE_TYPE ofs1, ofs2, sz1, sz2, sz, ofs_even;
	BOOL error;
	BYTE tmp1[2],tmp2[2];

	EXM_error = 0;
	if (!EXM_initialized) {
		EXM_error = 10;
		return(-1);
	}


	// special deal with overlap
	if ((dest_handle ==  source_handle) &&
	    (source_offset < dest_offset) &&
	    (source_offset + size > dest_offset))  { // dangerous overlap
		// We never promised to do THIS fast!
		ofs1 =  dest_offset+size;
		ofs2 =  source_offset+size;
		// we'll move 64K size blocks starting from the last
		while (size > 0) {
		       sz = min(size, PAGESIZE);
		       ofs1 -=  sz;
		       ofs2 -=  sz;
		       if (EXM_to_EXM(our_block, 0,
				      source_handle,ofs2, sz)) return(-1);
		       if (EXM_to_EXM(dest_handle,ofs1,
				      our_block, 0, sz)) return(-1);
		       size -= sz;
		}
		return(0);
	}

	link_no1 = user_blocks[dest_handle].first_link;
	if ((dest_handle >= blocks_size)  || (link_no1 == -1))
		{EXM_error = 3; return(-1);}
	if (find_block_offset(link_no1, dest_offset) < 0) return(-1);

	link_no2 = user_blocks[source_handle].first_link;
	if ((source_handle >= blocks_size)  || (link_no2 == -1))
		{EXM_error = 3; return(-1);}
	if (find_block_offset(link_no2, source_offset) < 0) return(-1);

	link1 = links+link_no1; // &links[link_no]
	ofs1 = dest_offset + (link1->offset << SIZE_SHIFT);
	sz1 =  ((long)link1->size << SIZE_SHIFT) - dest_offset;

	link2 = links+link_no2;
	ofs2 = source_offset + (link2->offset << SIZE_SHIFT);
	sz2 =  ((long)link2->size << SIZE_SHIFT) - source_offset;

	while (TRUE) {
		sz =  min3(sz1, sz2, size);
		// overlap test!
		if ((link1->handle == link2->handle) &&
		    (ofs1 + sz > ofs2) && (ofs2 + sz > ofs1)) {
			if (is_XMS && (sz & 1)) {
			 // here size > 1!
				error = move_ext2ext(our_handle,0, link2->handle,
						ofs2, sz-1) < 0;
				error = error || (move_ext2ext(link1->handle,ofs1,
						      our_handle,0, sz-1) < 0);
				error = error || (move_ext2ext(our_handle,0,
						    link2->handle,ofs2+sz-2,2) < 0);
				error = error || (move_ext2ext(link1->handle,ofs1+sz-2,
						   our_handle,0, 2) < 0);
			}
			else {
				error = move_ext2ext(our_handle,0, link2->handle,
						ofs2, sz) < 0;
				error = error || (move_ext2ext(link1->handle,ofs1,
						   our_handle,0, sz) < 0);
		       }
		// if overlap, use temporary buffer - our_handle page
		}
	       /* XMS specification says nothing about moving with overlap.
		* But I'am sure no reasonable implemantation of moving can
		* cause an error due to overlap when moving from
		* higher adresses to lower, (we've checked for
		* another case of overelap at the beginning of the routine!)
		* so you can bravely delete "if" and leave
		* only the "else" part
		*/
		else {	if (is_XMS && (sz & 1)) {
			  if (size == 1) {
			   // it may be last or first byte in page, so....
				ofs_even = ofs1 & 0xFFFE;
				error =  move_ext_mem(link1->handle,ofs_even,
					2, tmp1, 1) < 0;

				ofs_even = ofs2 & 0xFFFE;
				error = error ||  (move_ext_mem(link2->handle,ofs_even,
							2, tmp2, 1) < 0);
				tmp1[ofs1 & 1] = tmp2[ofs2 & 1];
				// other byte in tmp1 states unchanged!

				ofs_even = ofs1 & 0xFFFE;
				error = error || (  move_ext_mem(link1->handle,ofs_even,
					2, tmp1, 0) < 0);
			  }
			  else {
				error = move_ext2ext(link1->handle,ofs1,
				     link2->handle,ofs2, sz-1) < 0;
				error = error || (move_ext2ext(link1->handle,ofs1+sz-2,
				     link2->handle,ofs2+sz-2, 2) < 0);
			  }
			}
			else // at last - normal situation!
				error = move_ext2ext(link1->handle,ofs1,
				     link2->handle,ofs2, sz) < 0;
		}

		if (error) {
			EXM_error = 100;
			return(-1);
		}

		size -= sz;
		if (size <= 0) return(0);
		ofs1 += sz;
		ofs2 += sz;

		sz1 -= sz;
		if (sz1 <= 0) {
			link_no1 = link1->next;
			if (link_no1 < 0) {
				EXM_error = 4;
				return(-1);
			}
		link1 = links+link_no1;
		ofs1 = link1->offset << SIZE_SHIFT;
		sz1 =  (long)link1->size << SIZE_SHIFT;
		}

		sz2 -= sz;
		if (sz2 <= 0) {
			link_no2 = link2->next;
			if (link_no2 < 0) {
				EXM_error = 4;
				return(-1);
			}
		link2 = links+link_no2;
		ofs2 = link2->offset << SIZE_SHIFT;
		sz2 =  (long)link2->size << SIZE_SHIFT;
		}

	}


}


/////////////////////////// middle - level methods, implementation ////
///////////////////////////////////////////////////////////////////////


int free_piece(block_link &bl) {  // free memory block, pack page; -1 if error
	block_link *blp;
	int i, handleno;
	long sz,ofs;
	HANDLE handle;
	handle_descr h_descr;

	// find handle index in handle array
	for (handleno = 0; handleno < handles_num; handleno++)
		if (handles[handleno].handle == bl.handle) break;
	// debug
	if  (handleno >=  handles_num) {
		EXM_error = 101;
		return(-1);
	}

	h_descr = handles[handleno];

	if (bl.size > h_descr.offset) {EXM_error = 103; return(-1);}

	if ((bl.size == h_descr.offset)
		#ifdef  EMM
		&& (is_XMS)
		#endif
	   ){ 	// free HANDLE only
		if (free_ext_mem(bl.handle)) {
		// return the HANDLE to the system
			EXM_error = 100;
			return(-1);
		}
		handles_remaind++;
		delete_handle(handleno);
		return(0);
	}

	if (h_descr.offset != bl.offset + bl.size) {
		// we'll move filled tail of the page to lower addresses
		// to avoid fragmentation

		blp = links;

		for (i = last_link; i >= 0; i--,blp++)
		    if (( (*blp).handle == bl.handle) &&
			    ( (*blp).offset > bl.offset))
			(*blp).offset -= bl.size; // block is moved up


		ofs =  (long)(bl.offset+bl.size) << SIZE_SHIFT;
		sz =  ((long) h_descr.offset << SIZE_SHIFT) -  ofs;

		// actually :
		// move sz byte (from bl.handle, ofs)
		// to  (bl.handle, bl.offset<< SIZE_SHIFT)

		if ( ( move_ext2ext(our_handle,0, bl.handle, ofs, sz) < 0) ||
		     ( move_ext2ext(bl.handle,(long)bl.offset<< SIZE_SHIFT,
				our_handle,0, sz) < 0) ) {
			EXM_error = 100;
			return(-1);
			}

	       /* XMS specification says nothing about moving with overlap.
		* But I'am sure no decent implemantation of moving can cause
		* error due to overlap in such a situation (moving from
		* higher adresses to lower)
		* so you may bravely substitute the two movings above
		* with this one:
		*/
	       /*
		if (move_ext2ext(bl.handle, (long)bl.offset<< SIZE_SHIFT,
			    bl.handle, ofs, sz ) < 0 ) {
			EXM_error = 100;
			return(-1);
			}
	       */

	} // defragmenting was needed

	h_descr.offset -= bl.size;
	delete_handle(handleno);
	insert_handle(h_descr);
	return(0);
}


int free_link(int index) {
	block_link tmplink;
	int i;

	tmplink = links[index];
	links[index].handle = NULL_HANDLE;
	// mark as empty before defragmenting the heap


	links_states[index >> STATES_SHIFT ]++;
	first_possibly_free_link = min( first_possibly_free_link, index);
	if (index == last_link) last_link--;

	// free memory block
	return (free_piece(tmplink));
}

int free_chain(int first_link) {
	int cur_link,next;
	cur_link = first_link;
	while (cur_link >= 0) {
		next = links[cur_link].next;
		if (free_link(cur_link) < 0) return(-1);
		cur_link = next;
	}
	return(0);
}


int get_block() { 	// find, mark as filled, etc.; -1 if error
	int i,j;

	for (i = first_possibly_free_block >> STATES_SHIFT;
		i < (blocks_size >> STATES_SHIFT); i++)
		if (blocks_states[i]) break;
	i <<= STATES_SHIFT;
	i = max(i,first_possibly_free_block);

	for (; i < blocks_size; i++) // find free user block descriptor
			if (user_blocks[i].first_link < 0) break;



	if (i >= blocks_size) {
		j = blocks_size >> STATES_SHIFT; //  states size
		if (resize_array((BYTE *) blocks_states, 1,
				j, SIZE_INC >> STATES_SHIFT,(1 << STATES_SHIFT))
		   ||  resize_array((BYTE *)user_blocks , sizeof(userblock_descr),
				blocks_size, SIZE_INC ,0xFF)) {
			EXM_error = 1;
			return(-1);
		}
		// else i already is old blocks_size
	}

	blocks_states[i >> STATES_SHIFT]--;
	first_possibly_free_block = i+1;
	return(i);
}

int get_link() {
	int i,j;

	for (i = first_possibly_free_link >> STATES_SHIFT;
		i < (links_size >> STATES_SHIFT); i++)
		if (links_states[i]) break;

	i <<= STATES_SHIFT;
	i = max(i,first_possibly_free_link);

	for (; i < links_size; i++) // find free user link descriptor
			if (links[i].handle == NULL_HANDLE) break;



	if (i >= links_size) {
		j = links_size >> STATES_SHIFT; //  states size
		if (resize_array((BYTE *) links_states, 1,
				j, SIZE_INC >> STATES_SHIFT,(1 << STATES_SHIFT))
		   ||  resize_array((BYTE *)links , sizeof(block_link),
				links_size, SIZE_INC ,0xFF)) {
			EXM_error = 1;
			return(-1);
		}
		// else i already is old links_size
	}

	links_states[i >> STATES_SHIFT]--;
	first_possibly_free_link = i+1;

	last_link = max(last_link,i);
	return(i);
}

int continue_chain(int &first_link, int  &cur_link)  {
	int i_link;

	i_link = get_link();
	if (i_link < 0) { return(-1);}
	if (first_link < 0) first_link = i_link;
	else links[cur_link].next = i_link;
	cur_link = i_link;
	return(i_link);
}

int find_block_offset(int &link, SIZE_TYPE &offset) {
	// input - the first link in the chain, offset in the block
	// output - a link to subblock, containing offset of the block,
	// and local offset in this block

	while (link >= 0)  {
		offset -= (long)links[link].size << SIZE_SHIFT;
		if (offset < 0) goto l_Ok;
		link = links[link].next;
	}
	EXM_error = 4;
	return(-1);
l_Ok:  	offset += (long)links[link].size << SIZE_SHIFT;
	return(0);
}

int find_handle_place(WORD offset) {
	int left, right, j;
	if (handles_num == 0) return(0);
	if (offset >= handles[handles_num-1].offset) return(handles_num);
	if (offset <= handles[0].offset) return(0);
	left = 0;
	right = handles_num-1;
	while (right - left > 1)  {
	  j =  (right + left + 1) >> 1; // (right + left)/2.0
	  if  (handles[j].offset > offset) right = j;
	  else left = j;
	  }
	return right;
}

int insert_handle(handle_descr handle) {
	int place;
	place = find_handle_place(handle.offset);
	return insert_handle_by_no(handle, place);
}

int insert_handle_by_no(handle_descr handle, int where) {
	int newsize;

	if (handles_num >= handles_size)  // resize array
		if (resize_array((BYTE *) handles, sizeof(handle_descr),
			handles_size, 8 ,0xFF)) {
			EXM_error = 1;
			return(-1);
		}


	if (where < handles_num)
		memmove(handles+where+1,handles+where,
			sizeof(handle_descr) * (handles_num - where));
	handles[where] = handle;
	handles_num++;
	return(0);
};


void delete_handle(int where) {
	if (where < (--handles_num))
		memmove(handles+where,handles+where+1,
			sizeof(handle_descr) * (handles_num - where));
}

int link_with_new_page(int &first_link, int  &cur_link, long size) {
	HANDLE handle;
	handle_descr h_descr;

	handle = alloc_ext_mem(PAGEALLOC); // try to get new page

	if (handle != NULL_HANDLE) {
		handles_remaind--;
		h_descr.handle = handle;
		// we can get here only if size is less than PAGESIZEDW
		h_descr.offset = size;
		if (insert_handle(h_descr) < 0) return(-1);

		// get free link
		if (continue_chain(first_link, cur_link) < 0) return(-1);
		// fill link
		links[cur_link].handle = handle;
		links[cur_link].offset = 0;
		links[cur_link].size = size;
		links[cur_link].next = -1;
		return(0);
		}
	else return(1); // new page can not be allocated
}

// -------------- allocation strategy ------------------

// this routine is called if the manager is sure of success
// (no checkings)
int make_tail_lo(int &first_link, int  &cur_link, long size) {
// size already in DWORDs!
	int  handleno;
	int piece;
	handle_descr h_descr;

	while (size > 0) {
		handleno = find_handle_place(max(0,PAGESIZEDW-size)) - 1;
		// find_handle_place returns first PAGE with free space
		// less then we need now, so "-1"
		if (handleno < 0) handleno = 0;
		piece = min(PAGESIZEDW-handles[handleno].offset,size);
		h_descr = handles[handleno];

		// get free link
		if (continue_chain(first_link, cur_link) < 0) return(-1);

		// fill link
		links[cur_link].handle = h_descr.handle;
		links[cur_link].offset = h_descr.offset;
		links[cur_link].size = piece;
		links[cur_link].next = -1;
		// process handle descriptor
		h_descr.offset += piece;
		delete_handle(handleno);
		insert_handle(h_descr);
		size -= piece;
	}
	return(0);
}

int make_tail(int &first_link, int  &cur_link, long size) {
	long accum_size = 0;
	int i;
	HANDLE handle;
	handle_descr h_descr;

	for (i = 0; (i < 2) && ( i < handles_num); i++)
		accum_size += PAGESIZEDW - handles[i].offset;

	if  (accum_size >= size) return(make_tail_lo(first_link,cur_link,size));
	else {
		i = link_with_new_page(first_link,cur_link, size);
		if (i < 0) return (-1);
		if (i > 0) {   // try to use all we have
			accum_size = 0;
			for (i = 0; i < handles_num; i++)
				accum_size += PAGESIZEDW - handles[i].offset;
			if (accum_size < size ) {EXM_error = 2; return(-1);}
			else return(make_tail_lo(first_link,cur_link,size));
		}
		else return(0);
	}
}

int resize_array(BYTE * &ptr, WORD recsize,
		int &size, WORD incsize,
		BYTE fill) {
	BYTE *tmp;
	int newsize;

	newsize = size + incsize;
	tmp = new BYTE[newsize*recsize];
	if (tmp == NULL)  return(-1);
	memset(tmp, fill,recsize*newsize);
	memcpy(tmp, ptr, recsize*size);
	delete ptr;
	ptr = tmp;
	size = newsize;
	return(0);
}
