/* This is a sample program which demonstrates using the memory manager.
/* It allocates memory, horses around, and terminates.
/* A more useful example is in the files GRFILE.H and GRFILE.CPP,
/* where image_keeper uses memory above 1Mb to keep huge images.
/* example of using extended/expanded memory manager

Written by E. Podvoysky, ^Z, 1994
*/

#include <mem.h>
#include <alloc.h>
#include "manager.h"
#include "stdlib.h"
#include "stdio.h"

typedef  struct {
	USER_HANDLE handle;
	long size;
	BYTE code;
	} myblock_descr;

#define BUFSIZE 0xC000
BYTE *p; // BUFSIZE

int init_block(myblock_descr &descr) {
	int i;
	long size,ofs,sz;

	size =  descr.size;
	descr.handle = EXM_alloc(descr.size);
	if  (descr.handle == 0xFFFF) return(-1);
	i = descr.code;
	ofs = 0;

	while (size > 0) {
		sz = min(size,BUFSIZE);
		memset(p,i++,BUFSIZE>>1);
		memset(p+(BUFSIZE>>1),i++,BUFSIZE>>1);
		if (EXM_move_to_ext(descr.handle, p, ofs, sz)) return(-1);
		size -= sz;
		ofs += sz;
       }
       return(0);
}

void main() {
	myblock_descr block1, block2, block3;
	long ofs,sz, ofs2;
	BYTE val;

	if (manager_initialize()) {
		printf("Manager can not be initialized, err = %d \n", EXM_error);
		exit(0);
	}
	p = new BYTE[BUFSIZE];
	printf("%ld bytes available \n ", EXM_coreleft());
	sz = min(EXM_coreleft(), 2000000);

	block1.size = sz / 2;
	block1.code = 1;
	if (init_block(block1)) goto l_error;

	block2.size = sz / 3;
	block2.code = 100;
	if (init_block(block2)) goto l_error;

	if (EXM_free(block1.handle))  goto l_error;

	sz = min(EXM_coreleft(),1000000);
	block3.size = sz / 3;
	block3.code = 20;
	if (init_block(block3)) goto l_error;

	ofs =  block3.size/3;
	if ( EXM_move_to_conv(p,  block3.handle,
			 ofs , min(block3.size/2,BUFSIZE))) goto l_error;

	val = block3.code + ofs /  (BUFSIZE >> 1);
	if (p[0] != val) {
		printf("Read error! \n");
		goto l_finish;
	}
	ofs2 = block2.size/2;
	if ( EXM_to_EXM(block3.handle, ofs, // destination
		       block2.handle, ofs2, // source
		       min(block3.size/2,BUFSIZE))) goto l_error;
	if ( EXM_move_to_conv(p,  block3.handle,
			 ofs , min(block3.size/2,100))) goto l_error;
	val = block2.code + ofs2 /  (BUFSIZE >> 1);
	if (p[0] != val) {
		printf("Read error! \n");
		goto l_finish;
	}


	printf("Ok...\n");
	goto l_finish;

l_error:
	printf("EXM_error = %d \n", EXM_error);
l_finish:
	manager_finish(TRUE);
	sz = get_free_ext_mem();
	printf("Free ex. memory after finish: %ld bytes \n ", sz*0x400L);
}