; Low level routines for expanded/extended memory manager
; Written by Kiselev J. CZ 1994.

include emsapi.inc
include xmsapi.inc
.model large
public _init_ext_mem,_get_free_ext_mem,
public _alloc_ext_mem,_free_ext_mem,_move_ext_mem
public _move_ext2ext
.data
	EMM_dev_name db 'EMMXXXX0'
	EMB_move ext_mem_move <>
	EMM_move exp_mem_move <>
.code
.286
		XMS_or_EMS db -1	; 0 - XMS, 1 - EMS, -1 - nothing
		xms_call dd 0

; return: AX = -1 if error
_init_ext_mem	proc far
		push ds
		push si
		push di

		EMM_install_check
		jnz no_EMM
		Get_EMM_version
		jnz no_EMM
		cmp al,4 		; LIM 4.0
		jb no_EMM
		mov ax,1
		mov XMS_or_EMS,al
		jmp short ok_
no_EMM:
		XMS_install_check
		mov ax,-1
		jnz no_XMS_drv
		Get_XMS_drv_address
					; ES:BX points to driver entry point
		mov word ptr xms_call,bx
		mov word ptr xms_call+2,es
		xor ax,ax
		mov XMS_or_EMS,al
ok_:
no_XMS_drv:
		pop di
		pop si
		pop ds
		ret
_init_ext_mem	endp

; AX = size in KB
_get_free_ext_mem proc far
		cmp XMS_or_EMS,-1
		je get_free_error

		cmp XMS_or_EMS,1
		je get_free_EMM
		XMS_get_free_ext_mem
		ret
get_free_EMM:
		EMM_get_free_exp_mem
		jnz get_free_error
		mov ax,bx
		ret
get_free_error:
		xor ax,ax		; if error SIZE =  0
		ret
_get_free_ext_mem endp


block_size	EQU word ptr ss:[bx+4]	; in 64KB pages
; return: AX = -1 if error
_alloc_ext_mem	proc far
		cmp XMS_or_EMS,-1
		je alloc_error

		mov bx,sp
		mov ax,block_size
		cmp XMS_or_EMS,1
		je alloc_EMM
		shl ax,4		; *16
		XMS_alloc_ext_mem ax
		jnz alloc_error
		mov ax,dx
		ret
alloc_EMM:
		EMM_alloc_exp_mem ax
		jnz alloc_error
		mov ax,dx
		ret
alloc_error:
		mov ax,-1
		ret
_alloc_ext_mem	endp

handletofree	EQU word ptr ss:[bx+4]	; in pages (16KB)
; return: AX = -1 if error
_free_ext_mem	proc far
		cmp XMS_or_EMS,-1
		je free_error

		mov bx,sp
		mov ax,handletofree

		cmp XMS_or_EMS,1
		je free_EMM
		XMS_free_ext_mem ax
		jnz free_error
		xor ax,ax
		ret
free_EMM:
		EMM_free_exp_mem ax
		jnz free_error
		xor ax,ax
		ret
free_error:
		mov ax,-1
		ret
_free_ext_mem	endp

; move_ext_mem(handle,offset,size,real_address,flag_direction)
; handle,offset,size - block address in extra memory
; real_address - block address in conventional memory
; flag_direction - 0 from conv to ext

handle		EQU word ptr [bp+6]
block_offset	EQU word ptr [bp+8]
block_size	EQU dword ptr [bp+10]
real_address	EQU dword ptr [bp+14]
flag_direction	EQU byte ptr [bp+18]

_move_ext_mem	proc far
		push bp
		mov bp,sp
		push ds
		push di
		push si
		mov ax,SEG EMB_move
		mov ds,ax

		cmp XMS_or_EMS,-1
		jne no_error
		jmp move_error
no_error:
		les di,real_address
		mov cx,block_offset
		mov dx,handle

		cmp XMS_or_EMS,1
		je move_EMM

		test al,1
		;jnz move_error  

		mov si,offset EMB_move
		mov ax,word ptr block_size
		mov word ptr [si].xms_counter,ax
		mov ax,word ptr block_size+2
		mov word ptr [si].xms_counter+2,ax

		test flag_direction,1
		jz from_conv2ext
					; move from ext to conv
		mov [si].xms_source_handle,dx
		mov word ptr [si].offset_source,cx
		mov word ptr [si].offset_source+2,0
		mov [si].xms_dest_handle,0
		mov word ptr [si].offset_dest,di
		mov word ptr [si].offset_dest+2,es
		jmp short do_move_ext
from_conv2ext:
					; move from conv to ext
		mov [si].xms_source_handle,0
		mov word ptr [si].offset_source,di
		mov word ptr [si].offset_source+2,es
		mov [si].xms_dest_handle,dx
		mov word ptr [si].offset_dest,cx
		mov word ptr [si].offset_dest+2,0
do_move_ext:
		mov ah,0bh
		call xms_call
		dec ax
		jnz move_error
		jmp ok_move		; AX = 0 if OK
move_EMM:
		mov si,offset EMM_move
		mov ax,word ptr block_size
		mov word ptr [si].ems_counter,ax
		mov ax,word ptr block_size+2
		mov word ptr [si].ems_counter+2,ax
		mov ax,cx
		and ah,03fh		; rest
		shr cx,14		; page
		test flag_direction,1
		jz from_conv2exp
					; move from exp to conv
		mov [si].source_mem_type,1
		mov [si].source_handle,dx
		mov [si].source_init_ofs,ax
		mov [si].source_init_seg_page,cx

		mov [si].dest_mem_type,0
		mov [si].dest_handle,0
		mov [si].dest_init_ofs,di
		mov [si].dest_init_seg_page,es
		jmp short do_move_exp
from_conv2exp:
					; move from conv to exp
		mov [si].source_mem_type,0
		mov [si].source_handle,0
		mov [si].source_init_ofs,di
		mov [si].source_init_seg_page,es

		mov [si].dest_mem_type,1
		mov [si].dest_handle,dx
		mov [si].dest_init_ofs,ax
		mov [si].dest_init_seg_page,cx
do_move_exp:
		mov ax,5700h
		int 67h
		or ah,ah
		jnz move_error
		xor ax,ax
		jmp short ok_move		; AX = 0 if OK
move_error:
		mov ax,-1
ok_move:
		pop si
		pop di
		pop ds
		pop bp
		ret
_move_ext_mem	endp

; move_ext2ext(dest_handle,dest_offset,source_handle,source_offset,size)
; dest_handle,dest_offset - destination address
; source_handle,source_offset - source address
; size

d_handle	EQU word ptr [bp+6]
d_offset	EQU word ptr [bp+8]
s_handle	EQU word ptr [bp+10]
s_offset	EQU word ptr [bp+12]
ext_size	EQU dword ptr [bp+14]

_move_ext2ext	proc far
		push bp
		mov bp,sp
		push ds
		push di
		push si
		mov ax,SEG EMB_move
		mov ds,ax

		cmp XMS_or_EMS,-1
		jne xmsems_pres
		jmp ext_move_error
xmsems_pres:
		mov cx,s_offset
		mov dx,s_handle
		mov bx,d_offset
		mov di,d_handle
		mov ax,word ptr ext_size
		mov bp,word ptr ext_size+2

		cmp XMS_or_EMS,1
		je ext_move_EMM

		test al,1
		;jnz not_even  ???

		mov si,offset EMB_move
		mov word ptr [si].xms_counter,ax
		mov word ptr [si].xms_counter+2,bp

		mov [si].xms_source_handle,dx
		mov word ptr [si].offset_source,cx
		mov word ptr [si].offset_source+2,0
		mov [si].xms_dest_handle,di
		mov word ptr [si].offset_dest,bx
		mov word ptr [si].offset_dest+2,0

		mov ah,0bh
		call xms_call
		dec ax
		jnz ext_move_error
		jmp short ok_ext_move
ext_move_EMM:
		mov si,offset EMM_move
		mov word ptr [si].ems_counter,ax
		mov word ptr [si].ems_counter+2,bp

		mov ax,cx
		and ax,03fffh		; rest
		shr cx,14		; page

		mov [si].source_mem_type,1
		mov [si].source_handle,dx
		mov [si].source_init_ofs,ax
		mov [si].source_init_seg_page,cx

		mov ax,bx
		and ax,03fffh		; rest
		shr bx,14		; page

		mov [si].dest_mem_type,1
		mov [si].dest_handle,di
		mov [si].dest_init_ofs,ax
		mov [si].dest_init_seg_page,bx

		mov ax,5700h		; move
		int 67h
		cmp ah,92h		; overlap
		je overlap
		or ah,ah
		jnz ext_move_error
overlap:
		xor ax,ax
		jmp short ok_ext_move
ext_move_error:
		mov ax,-1
ok_ext_move:
		pop si
		pop di
		pop ds
		pop bp
		ret
_move_ext2ext	endp
end
