; The protected mode kernel
; (Chapter 16 Assembly Language Master Class)
; Written by Y. Kiselev

; General-purpose procedures used throughout the kernel

include hexconv.inc

server_data32 segment public use32
str_buff db 80 dup (0)
reg_name db 'axcxdxbxspbpsidi'
x db 10
y db 5
count db 0
server_data32 ends

server_code32 segment public use32

getch		proc near
		xor ax,ax
		int 16h
		ret
getch		endp

; put debug messages (all registers and last four dword of stack)
; call 32-bit codes

assume ds:server_data32

; This procedure prints the CPU state. Is called in critical situations
; for debugging purposes only.
print_debug_info proc near ; far!
		pushad
		push ds
		mov ebp,[esp+10h]
		pushfd

		mov eax,data32_sel
		mov ds,eax
		mov count,8
		mov x,10
		mov y,3
		mov str_buff,'e'
		mov str_buff[3],'='
		mov edx,offset reg_name
next_reg:
		push edx
		sub ebp,4
		mov esi,offset str_buff
		mov ax,[edx]
		mov [esi+1],ax
		mov eax,[ebp]
		mov ebx,4
		call PutHexDWord
		mov byte ptr [ebx+esi],0
		movzx eax,y
		movzx ebx,x
		mov cl,9
		mov esi,offset str_buff
		call disp_string
		inc y
		pop edx
		add edx,2
		dec count
		jnz short next_reg

		; print eflags
		pop eax
		mov esi,offset str_buff
		mov dword ptr [esi],616c6665h
		mov dword ptr [esi+4],20207367h
		mov ebx,7
		call PutHexDWord
		mov byte ptr [ebx+esi],0
		movzx eax,y
		movzx ebx,x
		mov cl,9
		mov esi,offset str_buff
		call disp_string

		mov x,26
		mov y,2
		mov word ptr str_buff[1],'='*256+'s'
		mov eax,cs
		mov byte ptr str_buff,'c'
		call put_seg_reg

		mov eax,ss
		mov byte ptr str_buff,'s'
		call put_seg_reg

		mov eax,[esp]
		mov byte ptr str_buff,'d'
		call put_seg_reg

		mov eax,es
		mov byte ptr str_buff,'e'
		call put_seg_reg

		mov eax,fs
		mov byte ptr str_buff,'f'
		call put_seg_reg

		mov eax,gs
		mov byte ptr str_buff,'g'
		call put_seg_reg

		; print stack
		mov ebp,[esp+10h]
		add y,3

		mov eax,[ebp+14h]
		mov esi,offset str_buff
		xor ebx,ebx
		call PutHexDWord
		mov byte ptr [ebx+esi],0
		movzx eax,y
		movzx ebx,x
		mov cl,9
		mov esi,offset str_buff
		call disp_string

		inc y
		mov eax,[ebp+10h]
		mov esi,offset str_buff
		xor ebx,ebx
		call PutHexDWord
		mov byte ptr [ebx+esi],0
		movzx eax,y
		movzx ebx,x
		mov cl,9
		mov esi,offset str_buff
		call disp_string

		inc y
		mov eax,[ebp+0ch]
		mov esi,offset str_buff
		xor ebx,ebx
		call PutHexDWord
		mov byte ptr [ebx+esi],0
		movzx eax,y
		movzx ebx,x
		mov cl,9
		mov esi,offset str_buff
		call disp_string

		inc y
		mov eax,[ebp+08h]
		mov esi,offset str_buff
		xor ebx,ebx
		call PutHexDWord
		mov byte ptr [ebx+esi],0
		movzx eax,y
		movzx ebx,x
		mov cl,9
		mov esi,offset str_buff
		call disp_string

		inc y
		mov eax,[ebp+4]
		mov esi,offset str_buff
		xor ebx,ebx
		call PutHexDWord
		mov byte ptr [ebx+esi],0
		movzx eax,y
		movzx ebx,x
		mov cl,9
		mov esi,offset str_buff
		call disp_string

		pop ds
		popad
		retf
print_debug_info endp

put_seg_reg	proc near
		inc y
		mov esi,offset str_buff
		mov ebx,3
		call PutHexWord
		mov byte ptr [ebx+esi],0
		movzx eax,y
		movzx ebx,x
		mov cl,9
		mov esi,offset str_buff
		call disp_string
		ret
put_seg_reg	endp

; eax = y
; ebx = x
; cl = attr
; ds:esi -> ASCIIZ string
;

disp_string	proc near
		push es
		mov dx,80
		mul dx
		add eax,ebx
		shl eax,1
		mov edi,eax
		mov ax,vtext_sel
		mov es,ax
		mov ah,cl
next_char:
		lodsb
		or al,al
		jz short end_char
		stosw
		jmp short next_char
end_char:
		pop es
		ret
disp_string	endp

; eax = y
; ebx = x
; cl = attr
; ch = char counter
; ds:esi -> ASCII string
;

disp_stringcx	proc near
		push es

		mov dx,80
		mul dx
		add eax,ebx
		shl eax,1
		mov edi,eax
		mov ax,vtext_sel
		mov es,ax
		mov ah,cl
next_char1:
		lodsb
		stosw
		dec ch
		jnz next_char1

		pop es
		ret
disp_stringcx	endp

; es:ebx points to descriptor
; edx = base (32 bits)
; ecx = limit (20 bits)
; al  = size
; ah  = Access Right byte
fill_descriptor32 proc near
		and bl,0f8h
		mov es:[ebx].desc.Limit_16,cx
		mov es:[ebx].desc.Base_lo,dx
		shr edx,16
		mov es:[ebx].desc.Base_hi,dl
		mov es:[ebx].desc.Acc_r,ah
		shr ecx,16
		or cl,al
		mov es:[ebx].desc.Limit_20,cl
		mov es:[ebx].desc.Base_32,dh
		ret
fill_descriptor32 endp
server_code32 ends

server_code16 segment public use16

; es:ebx points to descriptor
; edx = base (32 bits)
; ecx = limit (20 bits)
; al  = granularity,code size,AVL
; ah  = Access Right byte
fill_descriptor	proc near
		and bl,0f8h
		mov es:[bx].desc.Limit_16,cx
		mov es:[bx].desc.Base_lo,dx
		shr edx,16
		mov es:[bx].desc.Base_hi,dl
		mov es:[bx].desc.Acc_r,ah
		shr ecx,16
		or cl,al
		mov es:[bx].desc.Limit_20,cl
		mov es:[bx].desc.Base_32,dh
		ret
fill_descriptor	endp

make_callgate	proc near
		and bl,0f8h
		mov es:[bx],ax
		mov es:[bx+2],cx
		shr eax,16
		mov es:[bx+6],ax
		mov al,dl
		and al,0f0h
		shl ax,8
		or ah,CALL_GATE
		or al,dl
		and al,0fh
		mov es:[bx+4],ax
		ret
make_callgate	endp

SetPIC		proc	near
		cli

		in al,0a1h
		mov ah,al
		in al,21h
		push ax

		mov al,11h	;Ŀ
		out 0a0h,al     ;    Set ICW1 00010001b
		jmp short $+2   ;   
		jmp short $+2   ;   
		out 20h,al      ;
		jmp short $+2
		jmp short $+2
		mov al,bl       ;Ŀ
		out 0a1h,al     ;             01110xxx  (BIOS 70h - 77h)
		jmp short $+2   ;    Set ICW2
		jmp short $+2   ;   
		mov al,bh       ;             00001xxx  (BIOS 8h - 15h)
		out 21h,al      ;
		jmp short $+2
		jmp short $+2
		mov al,2        ;Ŀ
		out 0a1h,al     ;   
		jmp short $+2   ;    Set ICW3
		jmp short $+2   ;   
		mov al,4        ;   
		out 21h,al      ;
		jmp short $+2
		jmp short $+2
		mov al,1        ;Ŀ
		out 0a1h,al     ;   
		jmp short $+2	;    Set ICW4
		jmp short $+2   ;   
		out 21h,al      ;
		jmp short $+2
		jmp short $+2
		mov al,0	;Ŀ
		out 0a1h,al     ;    Set OCW1 (IMR(Mask))
		jmp short $+2   ;   
		jmp short $+2   ;   
		out 21h,al      ;
		pop ax
		jmp short $+2
		jmp short $+2
		out 21h,al
		jmp short $+2
		jmp short $+2
		mov ah,al
		out 0a1h,al
		sti
		ret
SetPIC		endp

; Return AL coprocessor type
; 0 - FPU not present
; 1 - 87
; 2 - 287
; 3 - 387
; ....
cpu_type	EQU byte ptr[bp+6]

fpu_stat	EQU word ptr[bp-2]
CheckCoprocessor proc far
		push bp
		mov bp,sp
		sub sp,2

		mov bl,al
		fninit
		mov fpu_stat,5a5ah
		fnstsw fpu_stat
		mov ax,fpu_stat
		or al,al
		mov al,0
		jnz end_fpu_chk
		fnstcw fpu_stat
		mov ax,fpu_stat
		and ax,103fh
		cmp ax,3fh
		mov al,0
		jne end_fpu_chk
		inc al
		cmp cpu_type,2
		jb end_fpu_chk
		inc al
		je end_fpu_chk
		inc al
		cmp cpu_type,3
		jne end_fpu_chk
		fld1
		fldz
		fdiv
		fld st
		fchs
		fcompp
		fstsw fpu_stat
		test fpu_stat,4000h
		jnz end_fpu_chk
		dec al
end_fpu_chk:
		mov sp,bp
		pop bp
		ret
CheckCoprocessor endp

; Return in AL processor type
; AL = 0..N (0 - 86,1 - 186,2 - 286,3 - 386,4 - 486,5 - Pentium,...)
CheckProcessor	proc far
		mov ax,0ffffh
		mov cl,42h
		shr ax,cl
		test ax,3fffh
		jnz chk186
		mov al,0
		ret
chk186:
		push sp
		pop bx
		cmp bx,sp
		je chk286
		inc al
		ret
chk286:
		mov bx,7000h
		push bx
		popf
		pushf
		pop bx
		and bh,70h
		jnz chk386
		inc al
		ret
chk386:
		mov bx,sp		;Intel original text
		and sp,not 3
		pushfd
		pop eax
		mov ecx,eax
		xor eax,40000h
		push eax
		popfd
		pushfd
		pop eax
		mov sp,bx
		xor eax,ecx
		jnz chk486
		mov al,3	;386
		ret
chk486:
		mov bx,sp
		and sp,not 3
		pushfd
		pop eax
		mov ecx,eax
		xor eax,200000h
		push eax
		popfd
		pushfd
		pop eax
		push ecx
		popfd
		mov sp,bx
		xor eax,ecx
		jnz chkP5
		mov al,4
		ret
chkP5:
		mov eax,1
		db 0fh,0a2h	; cpuid see Pentium chapter
		and ax,0f00h
		shr ax,8
		ret
CheckProcessor	endp

; returned in AX of memory installed on matherbord (in Kb)
get_system_memory proc near
		cli
		mov al,15h
		out 70h,al	; get base memory lo
		jmp short $+2
		jmp short $+2
		in al,71h
		mov ah,al
		mov al,16h
		out 70h,al	; get base memory hi
		jmp short $+2
		jmp short $+2
		in al,71h
		xchg ah,al
		mov dx,ax
		mov al,17h
		out 70h,al	; get base memory lo
		jmp short $+2
		jmp short $+2
		in al,71h
		mov ah,al
		mov al,18h
		out 70h,al	; get base memory hi
		jmp short $+2
		jmp short $+2
		in al,71h
		xchg ah,al
		add ax,dx
		sti
		ret
get_system_memory endp
server_code16	ends

; print AL directly on a screen (for exceptions debugging)
print_al	macro
local _0_9_,__0_9_
		mov ah,9
		mov dl,al
		shr al,4
		add al,'0'	;convert high half byte
		cmp al,'9'
		jbe _0_9_
		add al,'A'-'9'-1
_0_9_:
		mov es:[edi],ax
		mov al,dl
		and al,0fh
		add al,'0'
		cmp al,'9'
		jbe __0_9_	;convert low half byte
		add al,'A'-'9'-1
__0_9_:
		mov es:[edi+2],ax
		add edi,4
endm

deb macro
local wait1
		push ds
		pushad
		push cs
		call print_debug_info

		mov ecx,02ffffffh
wait1:
		dec ecx
		jnz wait1

		popad
		pop ds
endm
