; The protected mode kernel (Chapter 16 Assembly Language Master Class)
; Written by Y. Kiselev

; common EQUs and structures
Desc 		struc
	Limit_16 dw 0
	Base_lo dw 0
	Base_hi	db 0
	Acc_r db 0
	Limit_20 db 0
	Base_32	db 0
Desc		ends

idt_entry	struc			;structure of an IDT entry
	dest_ofs_lo dw 0		;destination offset (0-15)
	dest_selector dw 0			;destination selector
	copy_cnt db 0			;word count to copy (call gate)
	gate_typ db 0			;access type for gate
	dest_ofs_hi dw 0		;destination offset (16-31)
idt_entry	ends

; Access rights (bit PRESENT = 1 !)
RO_DATA		EQU	10010000b	; read only data
RW_DATA		EQU	10010010b	; read/write data
RW_STACK	EQU	10010110b	; read/write stack
EO_CODE		EQU	10011000b	; exec only code
ER_CODE		EQU	10011010b	; read/exec code
CN_CODE		EQU	10011100b	; exec only conforming code
CR_CODE		EQU	10011110b	; read/exec conforming code

;segment access byte
PRESENT		EQU	10000000b
ACCESSED	EQU	00000001b
CONFORM		EQU	00000100b
BUSY_TASK	EQU	00000010b

BIG_SEG		EQU	01000000b
BIG_GRANUL	EQU	10000000b
ALLOC_SEG	EQU	00010000b

DPL3		EQU	01100000b
DPL2		EQU	01000000b
DPL1		EQU	00100000b

RPL3		EQU	00000011b
RPL2		EQU	00000010b
RPL1		EQU	00000001b

LDT		EQU	00000010b OR PRESENT
TSS		EQU	00001001b OR PRESENT
CALL_GATE	EQU	00001100b OR PRESENT
TASK_GATE	EQU	00000101b OR PRESENT
INT_GATE	EQU	00001110b OR PRESENT
TRAP_GATE	EQU	00001111b OR PRESENT

; Attributes of Page Table Entry
PAGE_PRESENT	EQU	1
PAGE_WRITE	EQU	2
PAGE_USER	EQU	4
PAGE_ACCESSED	EQU	20h
PAGE_DIRTY	EQU	40h
PAGE_OCCUPIED	EQU	200h
PAGE_ONDISK	EQU	400h
PAGE_NDISCARD	EQU	800h

; flags of Eflags
TI		EQU	00000100b	; Table Index
IFFlag		EQU	200h
VMFlag		EQU	20000h
NTFlag		EQU	04000h
RFFlag		EQU	10000h
IOPL3		EQU	03000h
IOPL2		EQU	02000h
IOPL1		EQU	01000h

MASTER8259A	EQU	20h
SLAVE8259A	EQU	0a0h
EOI		EQU	20h

FALSE		EQU	0
TRUE		EQU	-1
