.model small
.stack 200h
.data
DPMI_not_inst_mess db 'DMPI host not installed',10,13,'$'
error_mess db 'Error.',10,13,'$'
ok_mess db 'DPMI works OK.',10,13,'$'

PM_door	dd 0		; entry point to the entry procedure
block_handle dw 0,0	; Parameters of a block which will be
linear_addr dw 0,0	; allocated
block_selector dw 0
.code
.286
Start:
		mov ax,@data
		mov ds,ax		; Load data segment

		mov bx,ss
		sub bx,ax
		shl bx,4
		cli
		mov ss,ax		; Load stack pointer
		add sp,bx
		sti

		mov bx,sp
		add bx,15		; Round up to next paragraph
		shr bx,4
		add ax,bx		; AX = SS + SP / 16 = segment address
					; of the end of the program space
		mov bx,es
		sub ax,bx		; AX = required ammount of paragraphs
		mov bx,ax
		mov ah,4ah
		int 21h 		; Resize block
		jnc Resize_ok
		jmp error
Resize_ok:
		mov ax,1687h
		int 2fh		; get address switch entry point
		or ax,ax	; error if AX != 0
		jz DPMIpresent
		jmp DPMI_not_inst
DPMIpresent:
				; Int 2fh/1686h returns:
				; BX, bit 0: clear = 16-bit host, set = 32-bit
				; CL - processor type
				;	2 - 286
				; 	3 - 386
				; 	4 - 486
				; 	5 - P5 ...
				; DH - DPMI major version
				; DL - DPMI minor version
				; SI - number of paragraphs required for
				;	DPMI host private data
				; ES:DI - points to protected mode switch
				; 	entry point

		mov word ptr PM_door,di		; Fill Up PM_door
		mov word ptr PM_door+2,es

		or si,si
		jz no_priv_data		; No private DPMI data

		mov bx,si
		mov ah,48h
		int 21h		; allocate some paragraphs for DPMI
				; host private data
		jnc alloc_ok
		jmp error
alloc_ok:
		mov es,ax

no_priv_data:

		mov ah,0	; AX: bit 0 =	0 for 16-bit application
				;      		1 for 32-bit application
				; ES host private data segment

		call PM_door	; switch to  protected mode

; ---- we are in protcted mode with:

			; CS = selector with base real mode CS and 64K limit
			; SS = selector with base real mode SS and 64K limit
			; DS = selector with base real mode DS and 64K limit
			; ES = PSP selector with 256 byte limit
			; FS = GS = 0

		mov ax,0501h	; DPMI function Allocate Memory Block
		xor bx,bx	; BX:CX = Byte count
		mov cx,40000
		int 31h		; allocate 40000 bytes in extended memory
		jc error

		mov block_handle,di
		mov block_handle+2,si
		mov linear_addr,bx
		mov linear_addr+2,cx

		xor ax,ax	; DPMI function 0 Allocate an LDT descriptor
		mov cx,1
		int 31h		; allocate 1 LDT descriptor, return selector
				; in AX
		jc error
		mov block_selector,ax

		mov ax,7	; DPMI function 7, Set segment base address
		mov dx,linear_addr+2
		mov cx,linear_addr
		mov bx,block_selector
		int 31h		; set segment base address
		jc error

		mov ax,8	; DPMI function 8, Set segment limit
		xor cx,cx
		mov dx,40000-1	; limit field = 39999
		mov bx,block_selector
		int 31h		; set segment limit
		jc error

		mov ax,9	; DPMI function 9, Set access rights
		mov bx,block_selector
		mov cx,cs
		and cx,3	; get CPL
		shl cl,5
		or cl,92h	; Present,Data,Expand-up,Read/Write segment
		int 31h		; set descriptor access rigth
		jc error

		; play with memory block
		; . . .
		; I fill the allocated block with 12h

		mov es,block_selector
		xor di,di
		mov al,12h
		mov cx,40000
		rep stosb

		mov ax,1	; DPMI function 1, Free descriptor in LDT
		mov bx,block_selector
		int 31h		; free descriptor
		jc error

		mov ax,0502h	; DPMI function 0502h,
				; Free extended memory block
		mov di,block_handle
		mov si,block_handle+2
		int 31h		; free extended memory block
		jc error
		
		mov dx,offset ok_mess	; Report success
exit:
		mov ah,9		; Output string
		int 21h

		mov ax,4c00h		; Terminate program
		int 21h
error:
		mov dx,offset error_mess
		jmp short exit

DPMI_not_inst:

		mov dx,offset DPMI_not_inst_mess
		jmp short exit
end Start
