; The protected mode kernel
; (Chapter 16 Assembly Language Master Class)
; Written by Y. Kiselev

; General protection fault handler


server_data32 segment para public use32
SGPF_mess db 'Server GPF.',0
UGPF_mess db 'User GPF.',0
server_data32 ends

; Stack frame
Old_gs		EQU word ptr [ebp+40]
Old_fs		EQU word ptr [ebp+36]
Old_ds		EQU word ptr [ebp+32]
Old_es		EQU word ptr [ebp+28]
Old_ss		EQU word ptr [ebp+24]
Old_esp		EQU dword ptr [ebp+20]

Old_eflags	EQU dword ptr [ebp+16]
Old_cs		EQU word ptr [ebp+12]
Old_eip		EQU dword ptr [ebp+8]
Error_code	EQU dword ptr [ebp+4]

server_code32 segment para public use32
exception_13	LABEL WORD
		test dword ptr [esp+12],VMflag	; interrupt in V86 mode ?
		jz from_PM		; NO
					; YES

		push ebp
		mov ebp,esp		; point to stack frame
		pushad			; save registers
		push ds

		mov eax,flat_sel
		mov ds,eax		; ds = 0 base and unlimited

		movzx esi,Old_cs	; get old cs
		shl esi,4
		add esi,Old_eip		; get old eip
					; DS:[ESI] points to instruction
					; which cause the exception
		xor ecx,ecx
		mov ah,-1
next_opcode:
		inc ah
		inc ecx			; ECX = size counter
		mov al,[esi]
		inc esi			; next opcode byte
		cmp al,67h 		; Was it prefix opcode?
		jbe next_opcode		; Yes
					; No
					; AH prefix present flag

		cmp al,9ch		; PUSHF(D)
		jz short emul_pushf

		cmp al,9dh		; POPF(D)
		jz short emul_popf

		cmp al,0cdh		; INT
		jz emul_int

		cmp al,0cfh		; IRET
		jz emul_iret

		cmp al,0f0h		; LOCK
		jz emul_lock

		cmp al,0fah		; CLI
		jz emul_cli

		cmp al,0fbh		; STI
		jz emul_sti

		mov esi,offset SGPF_mess
		jmp something_wrong		; Panic! Exit DEX immediately.

; -----------------POPF(D) emulator ---------------------------------
emul_popf:
		movzx ebx,Old_ss	; get old ss
		shl ebx,4
		add ebx,Old_esp		; get old esp
					; ebx poinst to top of V86 stack
		mov edx,4
		or ah,ah
		jnz _popfd
		movzx eax,word ptr [ebx]; get flags (word) from V86 stack
		shr edx,1
		jmp _popf
_popfd:
		mov eax,[ebx]		; get flags (dword) from V86 stack
_popf:
		mov ebx,Old_eflags	; get old flags
		and bx,7000h		; preserve NT & IOPL flags
		and eax,8fffh		; clear NT,IOPL flags
		or eax,ebx
		mov Old_eflags,eax	; save flags

		add Old_esp,edx		; adjust real mode stack
		jmp short back_to_virtual

; -------------- PUSHF(D) emulator -------------------------------
emul_pushf:
		movzx ebx,Old_ss	; get old ss
		shl ebx,4
		add ebx,Old_esp		; get old esp
					; ebx poinst to the top of V86 stack
		or ah,ah
		mov eax,Old_eflags
		jnz _pushfd

		mov [ebx-2],ax		; "push"
		sub Old_esp,2		; adjust V86 stack
		jmp short back_to_virtual
_pushfd:
		mov [ebx-4],eax		; "push"
		sub Old_esp,4		; adjust V86 stack

		jmp short back_to_virtual

; ---------------------- CLI emulator --------------------------
emul_cli:
		and Old_eflags,0FFFFFDFFh	; clear IF flag
						; Note:DEX is not multitasking
		jmp short back_to_virtual

; ----------------------- STI emulator -----------------------------
emul_sti:
		or Old_eflags,not 0FFFFFDFFh	; set IF flag
						; Note:DEX is not multitasking

; --------------------------------------------------------------------
; It's a shame but I don't know how LOCK emulator should work
emul_lock:
back_to_virtual:
		add Old_eip,ecx
back_to_v86:
		and Old_eflags,0FFFEFFFFh  ; clear Resume (RF) flags
		pop ds
		popad
		pop ebp
		add esp,4		; remove error code
		iretd

; --------- software interrupt, int nn emulator --------------------
emul_int:
		movzx edx,byte ptr [esi]; get interrupt number
		inc ecx			; inc command size

		movzx ebx,Old_ss	; get old ss
		shl ebx,4
		add ebx,Old_esp		; get old esp
					; ebx poinst to top of V86 stack
		or dl,dl		; Was it Int 0 (DEX reserved)?
		jz myint_emul
no_no_no:
					; dummy back link on V86 stack
		mov eax,Old_eflags
		and eax,06ccffh
		mov [ebx-2],ax		; save flags
		mov ax,Old_cs
		mov [ebx-4],ax		; save cs
		mov eax,Old_eip
		add eax,ecx
		mov [ebx-6],ax		; save ip
		sub Old_esp,6		; adjust V86 stack

		; EDX = interrupt index
		; Now get real mode interrupt vector
		movzx eax,word ptr [edx*4]
		mov Old_eip,eax
		mov ax,[edx*4+2]
		mov Old_cs,ax		; change back link (where to return
					; on exit from the GPF handler).

		jmp short back_to_v86

; Emulate some specific interrupts
myint_emul:
		mov ax,Old_cs
		cmp ax,server_code16	; Was it Int 0 in DEX's code?
		jne no_no_no

		mov eax,[esp+32]	; get Old_ax
		cmp al,15h		; Was it Int 15h?
		jne something_wrong

; \\\\\\\\\\\\\\\\\\\\\\\  int 15h handler \\\\\\\\\\\\\\\\\\\\\\
int_15_emul:
		mov ax,[ebx]		; get ax from V86 stack
		cmp ah,89
		jne no_entPM

		xor eax,eax
		add Old_esp,8		; adjust V86 stack (flags,cs,ip,ax)

		mov ax,[ebx+2]
		mov Old_eip,eax
		mov ax,[ebx+4]
		mov Old_cs,ax
		mov dx,[ebx+6]

		and edx,0fffh
		or dl,1			; set CF flag (error)
		mov eax,Old_eflags
		and ax,7000h
		or eax,edx
		mov Old_eflags,eax	; save flags

		jmp back_to_v86
no_entPM:
		cmp ah,87h
		je move_extmem

		mov word ptr [esp+32],ax
					; transfer control to
					; real mode Int 15h handler
		add Old_esp,2		; adjust V86 stack (AX reg)

		mov eax,data32_sel
		mov ds,eax

		movzx eax,word ptr old_int15
		mov Old_eip,eax			; take old address int15
		mov ax,old_int15+2
		mov Old_cs,ax		; change back link

		jmp back_to_v86

; The only Int 15h function supported is Move Block
move_extmem:
		movzx ecx,word ptr [esp+28]	; get Old_cx (word counter)
		movzx eax,Old_es
		shl eax,4
		movzx edx,word ptr [esp+8]	; Old_si (es:si points to desc)
		add eax,edx
		mov esi,[eax+12h]
		and esi,0ffffffh	; get 24-bit source address
		mov edi,[eax+1ah]
		and edi,0ffffffh	; get 24-bit destination address
					; ??? check block address' in
					; mem_blocks list ???
		mov eax,ds
		mov es,eax
		shr ecx,1
		rep movsd
		adc cl,0
		rep movsw

		xor eax,eax
		add Old_esp,8		; adjust V86 stack (flags,cs,ip,ax)

		mov ax,[ebx+2]
		mov Old_eip,eax
		mov ax,[ebx+4]
		mov Old_cs,ax
		mov dx,[ebx+6]

		and edx,0fffh
		and dl,not 1			; clear CF flag (no error)
		mov eax,Old_eflags
		and ax,7000h
		or eax,edx
		mov Old_eflags,eax	; save flags
		mov byte ptr [esp+33],0	; AH = 0 on return

		jmp back_to_v86

; ------------------------ IRET emulator --------------------------

emul_iret:
		movzx ebx,Old_ss	; get old ss
		shl ebx,4
		add ebx,Old_esp		; get old esp
					; ebx poinst to top of V86 stack
		mov dx,word ptr [ebx+4]	; get return flags from V86 stack

		mov eax,data32_sel
		mov ds,eax

		cmp esi,my_iret_linear_addr	; handler used RET 2,RET ...
		jne v86_iret			; NO (may be normal IRET)
						; YES
		mov edx,Old_Eflags
iret_to_PM:
		mov eax,[ebp+80]	; get PM task's Eflags
		and edx,0fffh		; clear NT,IOPL flags
		and eax,0fffe7200h	; clear RF and prevent NT,IOPL & IF flags
		or eax,edx
		mov [ebp+80],eax	; save new flags

		pop ds
		popad
		pop ebp
		add esp,40		; remove from PL0 V86 backlink (with error code)

		push eax
		push ebx
		push ds

		mov eax,data32_sel
		mov ds,eax

		cli
		add word ptr real_stack,6	; adjust V86 stack

		str ax                  ; current TSS selector
		sub ax,8                ; alias TSS selector
		mov ds,eax
		xor ebx,ebx

		mov eax,[esp+32+12]	; selector where to return
					; from the interrupt
		and al,3		; Will it be an interlevel IRET?
		jnz sub_20_bytes
		add [ebx].tss_str.esp_0,12+16+12; adjust PL0 stack within TSS
						; 12 bytes iret frame for level 0
						; 16 bytes saved segment registers
						; 12 bytes soft_int_emul structure
		jmp short sub_12_bytes
sub_20_bytes:
		add [ebx].tss_str.esp_0,20+16+12; adjust PL0 stack within TSS
						; 20 bytes iret frame for non 0 level
						; 16 bytes saved segment registers
						; 12 bytes soft_int_emul structure
sub_12_bytes:
		pop ds
		pop ebx
		pop eax
		jmp to_returner			; (See module SOFT_INT.INC)
v86_iret:
		push flat_sel
		pop ds
		mov eax,[ebx]			; CS:IP on the V86 stack
		push data32_sel			; Does it point to my IRET?
		pop ds
		cmp eax,dword ptr my_iret_addr
		je iret_to_PM			; YES (DEX returns to PM task)
						; NO (internal V86 IRET)

		push flat_sel
		pop ds

		add Old_esp,6		; adjust V86 stack

		movzx edx,ax
		mov Old_eip,edx
		shr eax,16
		mov Old_cs,ax
		mov dx,[ebx+4]

		and edx,0fffh
		mov eax,Old_eflags
		and ax,7000h
		or eax,edx
		mov Old_eflags,eax	; save flags

		jmp back_to_v86

; *** the end of emulation of V86 IOPL-sensitive instructions ***

from_PM:
		push eax
		mov eax,[esp+12]		; get Old_cs
		and al,RPL3
		pop eax
		jnz user_exception

		mov esi,offset SGPF_mess
		jmp something_wrong		; DEX error. Exit program.
user_exception:
		push ebp
		mov ebp,esp		; point to stack frame
		pushad
		push ds

		mov ds,Old_cs
		mov esi,Old_eip		; DS:[ESI] points to instruction
					; which caused the exception
		xor ecx,ecx
		mov ah,-1
next_opcode1:
		inc ah
		inc ecx			; ECX = size counter
		mov al,[esi]
		inc esi			; next opcode byte
		cmp al,67h 		; Prefix opcode?
		jbe next_opcode1	; Yes
					; No
					; AH prefix present flag

		cmp al,9ch		; PUSHF(D)
		jz short emul_user_pushf

		cmp al,9dh		; POPF(D)
		jz short emul_user_popf

		cmp al,0cdh		; INT
		jz emul_user_int

		cmp al,0cfh		; IRET
		jz emul_user_iret

		cmp al,0f0h		; LOCK
		jz emul_user_lock

		cmp al,0fah		; CLI
		jz emul_user_cli

		cmp al,0fbh		; STI
		jz emul_user_sti

		mov esi,offset UGPF_mess
		jmp something_wrong	; Exit program. (Error at user level)

; -------------- POPF(D) emulator for 32-bit PM code -------------
emul_user_popf:
		mov ds,Old_ss		; get old ss
		mov ebx,Old_esp		; get old esp
					; ebx poinst to top of user stack
		mov edx,4
		or ah,ah
		jz _popfd1
		movzx eax,word ptr [ebx]; get user's flags
		shr edx,1
		jmp _popf1
_popfd1:
		mov eax,[ebx]		; get user's eflags
_popf1:
		mov ebx,Old_eflags	; get old flags
		and bx,7000h		; preserve NT & IOPL flags
					; ??? check IF flags
					; (for multitasking version) ???
		and eax,8fffh		; clear NT,IOPL flags
		or eax,ebx
		mov Old_eflags,eax	; save flags

		add Old_esp,edx		; adjust real stack
		jmp short back_to_task

; --------------PUSHF(D) emulator for PM -------------------------
emul_user_pushf:
		mov ds,Old_ss		; get old ss
		mov ebx,Old_esp		; get old esp
					; ebx poinst to top of user stack
		or ah,ah
		mov eax,Old_eflags
		jz _pushfd1

		mov [ebx-2],ax		; "push"
		sub Old_esp,2		; adjust stack
		jmp short back_to_task
_pushfd1:
		mov [ebx-4],eax		; "push"
		sub Old_esp,4		; adjust stack

		jmp short back_to_task

; ----------------- CLI emulator ---------------------------
emul_user_cli:
		and Old_eflags,0FFFFFDFFh	; clear IF flag
		; clear VIF flag in TSS
						; ??? (check IF flag ???)
		jmp short back_to_task

; ------------------ STI emulator ---------------------------
emul_user_sti:
		or Old_eflags,not 0FFFFFDFFh	; set IF flag
		; set VIF flag in TSS
emul_user_iret:
emul_user_lock:
back_to_task:
		add Old_eip,ecx
		and Old_eflags,0FFFEFFFFh  ; clear RF flags
		pop ds
		popad
		pop ebp
		add esp,4		; remove error code
		iretd			; return to interrupted user task
emul_user_int:
		inc ecx
		add Old_eip,ecx
		pop ds
		popad
		pop ebp
		add esp,4		; remove error code
		jmp emul_interrupt	; go to interrupt emulating routine

; ------------------------ interrupt reflection to V86 mode ---------------

Add_const	EQU 16

Old_eflags_p	EQU dword ptr [esp+Add_const+60+12]	; +12 soft_int_emul structure
Old_cs_p	EQU word ptr [esp+Add_const+56+12]
Old_eip_p	EQU dword ptr [esp+Add_const+52+12]

New_gs		EQU word ptr [esp+Add_const+32]
New_fs		EQU word ptr [esp+Add_const+28]
New_ds		EQU word ptr [esp+Add_const+24]
New_es		EQU word ptr [esp+Add_const+20]
New_ss		EQU word ptr [esp+Add_const+16]
New_esp		EQU dword ptr [esp+Add_const+12]

New_eflags	EQU dword ptr [esp+Add_const+8]
New_cs		EQU word ptr [esp+Add_const+4]
New_eip		EQU dword ptr [esp+Add_const]

ref_to_v86:
		push gs			; Save PM segment registers
		push fs
		push ds
		push es

		sub esp,9*4		; frame to switch into V86
		push eax
		push ebx
		push edx

		push ds

		str ax                  ; current TSS selector
		sub ax,8                ; alias TSS selector
		mov ds,eax
		xor ebx,ebx

		mov eax,esp
		add eax,9*4+16
					; adjust PL0 stack in the current TSS
		mov [ebx].tss_str.esp_0,eax

		cmp byte ptr [esp+Add_const+62],TRUE	; hard_int flag
		je was_hard_int			; It was a hardware int

		mov eax,data32_sel
		mov ds,eax
		mov ax,real_mode_buffer

		mov New_gs,ax		; All V86 data segment registers
		mov New_fs,ax		; point to the transfer buffer
		mov New_ds,ax
		mov New_es,ax
was_hard_int:
		movzx edx,byte ptr [esp+Add_const+60]	; int_number

		mov eax,flat_sel
		mov ds,eax

		movzx eax,word ptr [edx*4]	; Get real mode vector
		mov New_eip,eax
		mov ax,[edx*4+2]
		mov New_cs,ax

		mov eax,data32_sel
		mov ds,eax

		movzx ebx,word ptr real_stack+2	; get old ss
		shl ebx,4
		movzx eax,word ptr real_stack	; get old esp
		add ebx,eax			; ebx poinst to top of V86 stack

		sub word ptr real_stack,6	; adjust V86 stack

		mov ax,word ptr real_stack+2
		mov New_ss,ax
		movzx eax,word ptr real_stack
		mov New_esp,eax

		mov eax,Old_eflags_p
		or eax,VMFlag		; EFlags VM = 1
		and eax,06ccffh		; IOPL = TF = IF = 0
		mov New_eflags,eax
					; dummy back link on V86 stack
		push flat_sel
		pop ds

		mov [ebx-2],ax		; save flags
		mov ax,seg back_from_v86int
		mov [ebx-4],ax		; save cs
		mov dx,offset back_from_v86int
		mov [ebx-6],dx		; save ip

		; enter to V86 mode
		pop ds
		pop edx
		pop ebx
		pop eax
		iretd		; fire! Call V86 handler.

; >>>>>>>>>>>>>>>>>>>>>>>>> hardware interrupt <<<<<<<<<<<<<<<<<<<<<<<<<<<<

Old_ss1		EQU word ptr [esp+28+12]	; +12 soft_int_emul structure
Old_esp1	EQU dword ptr [esp+24+12]
Old_eflags1	EQU dword ptr [esp+20+12]
Old_cs1		EQU word ptr [esp+16+12]
Old_eip1	EQU dword ptr [esp+12+12]

hardware_int:
		test dword ptr [esp+8+12],VMflag; hard int occur in V86 mode ?
		jz ref_to_v86	; No, reflect to V86
				; Yes
		push eax
		push ebx
		push ds

		mov eax,flat_sel
		mov ds,eax

		movzx ebx,Old_ss1	; get old ss
		shl ebx,4
		add ebx,Old_esp1	; get old esp
					; ebx poinst to top of V86 stack

					; dummy back link on V86 stack
		mov eax,Old_eflags1	; get Old_eflags1
		and ax,0fcffh		; clear IF & TF flags
		mov [ebx-2],ax		; save flags
		mov ax,Old_cs1		; get Old_cs1
		mov [ebx-4],ax		; save cs
		mov eax,Old_eip1	; get Old_ip1
		mov [ebx-6],ax		; save ip

		sub Old_esp1,6		; adjust V86 stack

		movzx ebx,byte ptr [esp+20]	; int number

		mov eax,flat_sel
		mov ds,eax
		movzx eax,word ptr [ebx*4]
		mov Old_eip1,eax	; change ip
		mov ax,[ebx*4+2]
		mov Old_cs1,ax		; change cs (back link)

		pop ds
		pop ebx
		pop eax
		add esp,12              ; remove soft_int_emul data from stack
		iretd			; back to V86 mode
server_code32 ends

