; The protected mode kernel
; (Chapter 16 Assembly Language Master Class)
; Written by Y. Kiselev

; This module will contains server's IDT

include except.inc
include exc_13.inc
include hard_int.inc
include soft_int.inc

idt_size = 256*8-1

set_handler	macro vector,handler
		mov eax,handler
		mov bl,vector
		call make_IDT_descr
endm

server_code32 segment para public use32
fill_IDT	proc near
		mov ax,idt_sel
		mov es,ax

		mov cx,(PRESENT shr 8) or INT_GATE
		mov eax,offset exceptions
		mov dx,code32_sel
		xor ebx,ebx
next_exc:
		call make_IDT_descr
		inc ebx
		cmp bl,12h
		jne next_exc

		set_handler 0,offset exception_00
		set_handler 4,offset exception_04
		set_handler 5,offset exception_05
		set_handler 6,offset exception_06
		set_handler 8,offset exception_08
		set_handler 10,offset exception_10
		set_handler 12,offset exception_12
		set_handler 13,offset exception_13
		set_handler 14,offset page_fault_handler
		set_handler 16,offset exception_16
		set_handler 17,offset exception_17

;//////////////////////////////////////////
		mov eax,offset softint1
		mov bl,12h
next_softint1:
		call make_IDT_descr
		inc ebx
		cmp bl,50h
		jne next_softint1

;///////     first PIC    //////////////////

		set_handler 50h,offset IRQ8
		set_handler 51h,offset IRQ9
		set_handler 52h,offset IRQA
		set_handler 53h,offset IRQB
		set_handler 54h,offset IRQC
		set_handler 55h,offset IRQD
		set_handler 56h,offset IRQE
		set_handler 57h,offset IRQF

;//////////////////////////////////////////
		mov eax,offset softint1
		mov bl,58h
next_softint2:
		call make_IDT_descr
		inc ebx
		cmp bl,70h
		jnz next_softint2

;////////////  second PIC ////////////////////

		set_handler 70h,offset IRQ70
		set_handler 71h,offset IRQ71
		set_handler 72h,offset IRQ72
		set_handler 73h,offset IRQ73
		set_handler 74h,offset IRQ74
		set_handler 75h,offset IRQ75
		set_handler 76h,offset IRQ76
		set_handler 77h,offset IRQ77

;//////////////////////////////////////////
		mov eax,offset softint1
		mov bl,78h
next_softint3:
		call make_IDT_descr
		inc ebx
		cmp ebx,100h
		jne next_softint3
		ret
fill_IDT	endp

; es:0 points to IDT
; bl = interrupt number
; dx:eax pointer to interrupt handler (only dx for task gate)
; cl = interrupt or trap gate (INT_GATE or TRAP_GATE or TASK_GATE)
; ch = Access Right byte (Pres & DPL bits)
make_IDT_descr	proc near
		push eax
		push ebx
		movzx ebx,bl
		shl ebx,3	; *8
		mov es:[ebx].idt_entry.dest_ofs_lo,ax	;handler offset (0-15)
		shr eax,16
		mov es:[ebx].idt_entry.dest_ofs_hi,ax	;handler offset (16-31)
		mov es:[ebx].idt_entry.dest_selector,dx	;handler selector
		mov al,cl
		or al,ch
		mov es:[ebx].idt_entry.gate_typ,al		;access type for gate
		pop ebx
		pop eax
		ret
make_IDT_descr	endp
server_code32 ends
