; The protected mode kernel
; (Chapter 16 Assembly Language Master Class)
; Written by Y. Kiselev

; Global heap manager (page-aligned segment allocation)

server_data32 segment para public use32
refresh_stop_flag db FALSE
memerror_code	dd 0
current_time 	dd 0
server_data32 ends

assume ds:server_data32
server_code32 segment para public use32

; prepare tables
Init_mem_manager proc near
		mov eax,data32_sel
		mov ds,eax
		cld
		mov edi,page_use_stat_address
		xor ecx,ecx
		mov cx,pages_in_system

		mov eax,flat_sel
		mov es,eax
		xor eax,eax
		mov current_time,eax

		rep stosd		; zero page usage statistic

		mov edi,index_table_address
		mov cx,DISK_pages
		dec eax
		rep stosd		; fill index table with -1
					; (pages are nowhere in the file)
		ret
Init_mem_manager endp

; WORD (PASCAL) allocate_segment (WORD pages,BYTE Discardable_flag);

req_pages	EQU word ptr [ebp+16]
discard_flag	EQU byte ptr [ebp+12]

allocate_segment proc near		; The procedure is actually far!
					; NEAR is needed to make
					; the Assembler happy.
		push ebp
		mov ebp,esp

		push ds
		push es

		mov eax,data32_sel
		mov ds,eax

		mov eax,flat_sel
		mov es,eax

		mov memerror_code,0

		movzx eax,req_pages
		cmp ax,Ext_free_mem	; Are there enough pages available?
		ja nomorefreepages
		or eax,eax		; Asked for 0 pages?
		jz zero_pages_req	; Crap! Exit.

		movzx ebx,first_free_page
		mov esi,ebx

		add ebx,page_tables_address	; EBX points to the first
						; free page entry

		xor edx,edx
		mov ecx,[ebp+8]		; check caller's RPL
		and cl,3
		cmp cl,3
		jb superpage		; Supervisor page
		or dl,PAGE_USER		; User page
superpage:
		cmp discard_flag,TRUE
		jne nondis
		or dx,PAGE_NDISCARD
nondis:
		or dx,PAGE_OCCUPIED OR PAGE_WRITE
alloc_nextpag:
		test word ptr es:[ebx],PAGE_OCCUPIED	; Is the page occupied?
		jnz bad_page_array		; jump if fatal error occur
		and word ptr es:[ebx],0f199h
		or es:[ebx],edx			; set page attributes
		add ebx,4
		dec eax
		jnz alloc_nextpag

		mov first_free_page,bx		; Adjust heap top pointer

		xor ebx,ebx
		mov ecx,items_in_LDT-1

		sldt dx			; get current LDT selector
		sub dx,8		; DX - alias LDT selector

		test byte ptr [ebp+8],3 ; check caller's RPL
		jnz no_zero_ring
		mov edx,GDT_sel		; Allocate descriptor in GDT
		mov bl,8		; skip null descriptor
		mov ecx,items_in_GDT-1
no_zero_ring:
		mov es,edx
next_descr:				; Find free descriptor
		test byte ptr es:[ebx+6],ALLOC_SEG
		jz found
skip_gates:
		add ebx,8
		dec ecx
		jnz next_descr
		jmp short nomorefreedescr
found:
		cmp byte ptr es:[ebx+5],0ddh
		jne no_last_descr
		mov word ptr es:[ebx+8+5],00ddh	; next descriptor will be
						; the last descriptor now
		jmp short filldes
no_last_descr:
		test byte ptr es:[ebx+5],4
		jnz skip_gates
filldes:
		push ebx

		mov edx,esi		; ESI = previous first_free_page value
		shl edx,10		; page_entry_offset/4*12 = page_number
		movzx ecx,req_pages
		push ecx

		dec ecx
		mov ah,RW_DATA
		mov al,[ebp+8]
		and al,3
		shl al,5
		or ah,al	; DPL = caller's CPL
		mov al,BIG_SEG OR ALLOC_SEG OR BIG_GRANUL
		call fill_descriptor32  ; fill up descriptor
		pop eax

		sub Ext_free_mem,ax	; adjust free extended memory variable
		pop eax
		and eax,0ffffh

		mov cl,[ebp+8]
		and cl,3
		or al,cl		; RPL = caller's CPL
		test byte ptr [ebp+8],3 ; check caller's RPL
		jz desc_in_gdt
		or al,TI
desc_in_gdt:
		clc			; AX = selector
		pop es
		pop ds
		pop ebp
		retf 8
; Errors
nomorefreedescr:
		inc memerror_code	; 4
nomorefreepages:
		inc memerror_code	; 3
zero_pages_req:
		inc memerror_code	; 2
bad_page_array:
		inc memerror_code	; 1

		stc
		pop es
		pop ds
		pop ebp
		retf 8
allocate_segment endp

; WORD free_segment (WORD selector);

free_selector	EQU word ptr [ebp+12]
segment_size	EQU dword ptr [ebp-4]
segment_base	EQU dword ptr [ebp-8]
free_pages_ofs	EQU dword ptr [ebp-12]

free_segment	proc near  	; The procedure is actually far!
				; NEAR is needed to make
				; the Assembler happy.
		push ebp
		mov ebp,esp
		sub esp,12

		push ds
		push es
		cld

		mov eax,data32_sel
		mov ds,eax

		mov memerror_code,0

		sldt dx			; get current LDT selector
		sub dx,8		; DX - alias LDT selector

		test byte ptr [ebp+8],3 ; check caller's RPL
		jnz no_zero_ring1
		mov edx,GDT_sel		; The descriptor is in GDT
no_zero_ring1:
		movzx ebx,free_selector
		and bl,0f8h
		mov es,edx
		test byte ptr es:[ebx+6],ALLOC_SEG
		jz dealloc_error	; Error! The segment wasn't allocated
		and byte ptr es:[ebx+6],not ALLOC_SEG	; mark descriptor as free

		movzx edx,byte ptr es:[ebx+6]
		and dl,0fh
		shl edx,16
		mov dx,es:[ebx]		; get segment size = Limit+1 (in pages)
		inc edx
		mov segment_size,edx

		movzx eax,byte ptr es:[ebx+7]
		shl eax,24
		mov edi,es:[ebx+2]	; get segment base address
		and edi,0ffffffh
		or edi,eax
		mov segment_base,edi

		shr edi,10
		add edi,page_tables_address
		mov free_pages_ofs,edi

		mov eax,flat_sel
		mov es,eax

;-----------------  mark pages as free (set default attributes) --------------
make_free_page:
		test word ptr es:[edi],PAGE_OCCUPIED	; Is the page occupied?
		jz free_free_page
		and word ptr es:[edi],0f199h	; clear D,A,U/S,R/W,
						; Discardable,
						; On Disk and Occupied bits
		add edi,4
		dec edx
		jnz make_free_page

;--------------------- save page entries to be freed --------------------
		push edi		; EDI points to first alloc page
		mov ecx,segment_size
		mov edi,swap_buff_address
		mov esi,free_pages_ofs

		rep movs dword ptr es:[edi],es:[esi]	; save pages to be freed
							; in swap_buffer

;------------------------ shrink array of entries ----------------------
		pop esi		; ESI points to first alloc page
		mov edi,free_pages_ofs

		movzx ecx,first_free_page

		mov eax,segment_size
		shl eax,2
		sub first_free_page,ax		; decrease heap top

		add ecx,page_tables_address
		sub ecx,esi
		jz free_last_block

		shr ecx,2		; /4 dword counter (page entries to move)
		push edi
		push esi
		push ecx

		cli			; no interrupts !!!
		rep movs dword ptr es:[edi],es:[esi]

;-------- move entries to be freed immediately above the heap top ----------
		mov esi,swap_buff_address
		mov ecx,segment_size

		rep movs dword ptr es:[edi],es:[esi]

; -------------- shrink statistics table, too -------------------------

		mov eax,page_tables_address
		mov ebx,page_use_stat_address

		pop ecx
		pop esi
		pop edi
		sub esi,eax
		sub edi,eax

		add esi,ebx
		add edi,ebx

		rep movs dword ptr es:[edi],es:[esi]

;----------------- adjust base addresses of descriptors ------------
free_last_block:
		mov edi,segment_base	; EDI = base address of descr
					; to be deallocated
		mov esi,segment_size
		shl esi,12		; ESI = base decrement

		mov eax,GDT_sel		; begin with GDT
		mov ds,eax

		mov ebx,8		; skip 0 descr in GDT
chk_ndesc:
		test byte ptr [ebx+6],ALLOC_SEG
		jz free_seg

		mov al,[ebx+5]
		test al,10h		; System bit
		jnz no_sys_descr	; DATA or CODE

		and al,0fh
		cmp al,2
		je LDT_descr

		and al,00001101b
		cmp al,00001001b	; TSS descriptor ?
		jne descr_wo_base
no_sys_descr:
		call change_base	; change base addr of segment
		jmp short free_seg
LDT_descr:
;----------------------- LDT ----------------------------------------
		call change_base	; change base addr of LDT
		push ebx
		push ds

		sub ebx,8
		mov ds,ebx		; load DS with LDT alias selector

		xor ebx,ebx
chk_ndesc1:
		test byte ptr [ebx+6],ALLOC_SEG
		jz free_seg1
		call change_base	; change base addr of segment
free_seg1:
		add ebx,8
		cmp byte ptr [ebx+5],0ddh	; if type of descriptor = 0ddh then
						; it's the last in table
		jnz chk_ndesc1
		pop ds
		pop ebx
;-----------------------------------------------------------------------
free_seg:
descr_wo_base:
		add ebx,8
		cmp byte ptr [ebx+5],0ddh	; if type of descriptor = 0ddh then
						; it's last in table
		jnz chk_ndesc

		mov eax,cr3
		mov cr3,eax			; refresh TLB !!!

		sti
		mov eax,data32_sel
		mov ds,eax

		mov eax,segment_size
		add Ext_free_mem,ax	; adjust free extended memory variable

		clc
		pop es
		pop ds
		mov esp,ebp
		pop ebp
		retf 4

free_free_page:
		inc memerror_code
dealloc_error:
		inc memerror_code

		stc
		pop es
		pop ds
		mov esp,ebp
		pop ebp
		retf 4
free_segment	endp

change_base     proc near
		movzx eax,byte ptr [ebx+7]
		shl eax,24
		mov edx,[ebx+2]		; get segment base address
		and edx,0ffffffh
		or edx,eax
		cmp edx,edi
		jb no_adj
		jne adj
		; descriptor with same base addr must be killed, too.
		and byte ptr [ebx+6],not ALLOC_SEG	; mark descriptor as free

		ret
adj:
		sub edx,esi
		mov [ebx+2],dx		; load address back to descriptor
		shr edx,16
		mov [ebx+4],dl
		mov [ebx+7],dh
no_adj:
		ret
change_base	endp

; refresh_statistic call from timer handler each half second
refresh_statistic proc near
		; During page swapping DEX sets the refresh_stop_flag
		; thus disabling collecting statistics.

		push ds
		push data32_sel
		pop ds

		cmp refresh_stop_flag,TRUE
		je no_refr

		push eax
		push ecx
		push edi
		push esi

		inc current_time

		; DEX checks every page and if it is Accessed, puts
		; the current time into the corresponding entry of
		; the statistic table.

		mov edi,page_use_stat_address
		mov esi,page_tables_address
		add edi,400h
		add esi,400h			; 1Mb
		movzx ecx,pages_in_system
		sub ecx,256			; pages in first Mb
		mov eax,current_time

		push flat_sel
		pop ds
chk_nxt_page:
		test byte ptr [esi],PAGE_ACCESSED
		jz no_used
		and byte ptr [esi],not PAGE_ACCESSED
		mov [edi],eax
no_used:
		add esi,4
		add edi,4
		dec ecx
		jnz chk_nxt_page

		pop esi
		pop edi
		pop ecx
		pop eax
no_refr:
		pop ds
		ret
refresh_statistic endp

; Page fault (Int 14) handler

min_time	EQU dword ptr [ebp-4]
page_for_save	EQU dword ptr [ebp-8]
page_phys_addr	EQU dword ptr [ebp-12]
linear_page_addr EQU dword ptr [ebp-16]
index_saved_page EQU dword ptr [ebp-20]

page_fault_handler:
		test byte ptr [esp],1	; check error code
					; (page protection violation or
					; page not present)
		jnz viol_page_prot

		push ebp
		mov ebp,esp
		sub esp,20

		pushad
		push ds
		push es

		mov eax,data32_sel
		mov ds,eax

		mov refresh_stop_flag,TRUE	; Suspend statistics
		sti

;----------------- Find least recently used page ----------------------

		mov eax,flat_sel
		mov es,eax
		cld

		mov edi,page_use_stat_address
		mov esi,page_tables_address
		add edi,400h
		add esi,400h                    ; 1Mb
		movzx ecx,pages_in_system
		sub ecx,256			; pages in first Mb

		mov min_time,-1
check_next_t:
		test word ptr es:[esi],PAGE_NDISCARD
		jnz page_nondiscard
		test byte ptr es:[esi],PAGE_PRESENT
		jz page_nondiscard

		mov eax,es:[edi]
		cmp eax,min_time
		ja newer
		mov min_time,eax
		mov page_for_save,esi	; page entry address of page
					; to be rolled out
page_nondiscard:
newer:
		add esi,4
		add edi,4
		dec ecx
		jnz check_next_t

		cmp min_time,-1
		je no_place_for_loaded_page	; Weird: no page to roll out

		mov ebx,page_for_save
		mov esi,ebx

		mov eax,es:[esi]
		and ax,0f000h		; take page phys address
		mov page_phys_addr,eax
		sub esi,page_tables_address
		shl esi,10		; *1024
		mov linear_page_addr,esi

;----------------- check swap conditions --------------------

		test word ptr es:[ebx],PAGE_ONDISK
		jz not_OnDisk

; 1. Is page not dirty and on disk ?
		test word ptr es:[ebx],PAGE_DIRTY	; if page is not Dirty and
						; it's copy is on disk
						; then there's no need to write it
		jnz page_is_dirty
						; page doesn't go to disk

		mov edi,index_table_address
		mov eax,page_phys_addr
		movzx ecx,DISK_pages
		repne scasd		; Find where the page
					; being rolled out lies
					; in the swap file
		jne virtual_mem_is_corrupt	; Fatal error

		sub edi,4
		mov dword ptr es:[edi],-2	; Mark this page in the
						; index table as absent in RAM
		sub edi,index_table_address
		shl edi,10		; find page offset within swap file
		mov index_saved_page,edi

		jmp load_page

page_is_dirty:
; 2. Page is dirty and on disk
				; in this case write page to it's place
				; in the swap file

		mov edi,index_table_address
		mov eax,page_phys_addr
		movzx ecx,DISK_pages
		repne scasd
		jne virtual_mem_is_corrupt

		sub edi,4
		mov dword ptr es:[edi],-2	; mark page as absent in RAM

		call save_page

		jmp load_page

;3. Page is not on disk
not_OnDisk:
; a. looking for -1 in index table (unoccupied record in the swap file)
		xor eax,eax
		mov edi,index_table_address
		dec eax
		movzx ecx,DISK_pages
		repne scasd
		jne no_free_index
					; write page to it
		sub edi,4
		mov dword ptr es:[edi],-2	; mark page as absent in RAM

		call save_page

		jmp load_page
no_free_index:
	; b. looking for Present, OnDisk, Dirty page
		mov edi,page_tables_address
		add edi,400h
		movzx ecx,pages_in_system
		sub ecx,256
chk_next_pe:
		test word ptr es:[edi],PAGE_PRESENT OR PAGE_ONDISK OR PAGE_DIRTY
		jnz such_page_is
		add edi,4
		dec ecx
		jnz chk_next_pe
		jmp short no_such_pages
such_page_is:
		and word ptr es:[edi],not PAGE_ONDISK ; mark as not On Disk
		mov eax,es:[edi]
		and ax,0f000h		; EAX = page physical address

		mov edi,index_table_address
		movzx ecx,DISK_pages
		repne scasd		; Find it's physical address in the
					; index table
		jne virtual_mem_is_corrupt

		sub edi,4
		mov dword ptr es:[edi],-2	; mark page as absent in RAM

		call save_page		; write the roll out page to
					; the found page's place
					; in the swap file

		jmp load_page
no_such_pages:
	; c. take a random entry from index table
	; and mark the corresponding page as not onDisk
		mov edi,index_table_address
		movzx ecx,DISK_pages
		mov eax,es:[46ch]	; get tick counter
		div ecx
		shl edx,2		; *4 (random value)
		add edi,edx
next_index:
		mov eax,es:[edi]	; get physical address
		cmp eax,-2		; if -2 then the found page
					; has no copy in RAM
		je find_nearest		; If so, find the nearest which
					; does have a copy in RAM
		jmp short find_it_in_table
find_nearest:
		sub edi,4
		cmp edi,index_table_address
		jnb next_index
		movzx edi,DISK_pages
		dec edi
		shl edi,2		; *4
		add edi,index_table_address
		jmp short next_index

find_it_in_table:
		mov esi,page_tables_address
		add esi,400h
		movzx ecx,pages_in_system
		sub ecx,256
cmp_next_addr:
		mov ebx,es:[esi]
		and bx,0f000h
		cmp eax,ebx	; Find the found page's entry
				; in the page tables :*)
		je find_ok
		add esi,4
		dec ecx
		jnz cmp_next_addr
		jmp virtual_mem_is_corrupt	; System out of order
find_ok:
		and word ptr es:[esi],not PAGE_ONDISK	; mark the found page
							; as not On Disk

		mov dword ptr es:[edi],-2	; mark the index table entry
					; for the found page as absent in RAM

		call save_page

;----------------------- load the requested page ----------------------
load_page:
		mov eax,cr2
		shr eax,10
		add eax,page_tables_address	; EAX - points to entry of
						; faulted page
		mov edi,eax

		mov eax,es:[edi]
		mov edx,eax
		and edx,0fffh
		or edx,page_phys_addr
		or dx,PAGE_PRESENT
		and dl,not PAGE_DIRTY		; clear Dirty bit
		mov es:[edi],edx

		and ax,0f000h
		cmp eax,0fffff000h
		je used_first_time

		or word ptr es:[edi],PAGE_ONDISK

		mov dx,ax
		shr eax,16
		mov cx,ax
		mov ax,4200h		; at beginning of the file
		mov bx,Swap_file_handl
		int 21h			; file seek
		jc swap_file_is_corrupt

		push ds
		mov bx,Swap_file_handl
		mov edx,linear_page_addr
		mov eax,es
		mov ds,eax
		mov ah,3fh
		mov ecx,1000h		; read one page
		int 21h
		pop ds
		jc swap_file_is_corrupt

used_first_time:

		mov edi,page_for_save
		mov eax,es:[edi]
		and eax,0fffh
		or eax,index_saved_page
		and al,not PAGE_PRESENT
		mov es:[edi],eax	; load in table entry discardable
					; page index in swap file

		mov refresh_stop_flag,FALSE	; Enable refresh_statistics

		pop es
		pop ds
		popad
		mov esp,ebp
		pop ebp
		add esp,4		; remove error code
		iretd

save_page	proc near
		sub edi,index_table_address
		shl edi,10		; find page index within swap file
		mov index_saved_page,edi

		mov ax,4200h		; at beginning of the file
		mov bx,Swap_file_handl
		mov dx,di
		shr edi,16
		mov cx,di
		int 21h			; file seek
		jc swap_file_is_corrupt

		push ds
		mov bx,Swap_file_handl
		mov edx,linear_page_addr
		mov eax,es
		mov ds,eax
		mov ah,40h
		mov ecx,1000h		; write one page
		int 21h
		pop ds
		jc swap_file_is_corrupt
		ret
save_page	endp

viol_page_prot:
		jmp something_wrong

no_place_for_loaded_page:
		mov eax,-1
virtual_mem_is_corrupt:
		mov eax,-2
swap_file_is_corrupt:
		mov eax,-3
		jmp something_wrong
server_code32 ends
