; The protected mode kernel
; (Chapter 16 Assembly Language Master Class)
; Written by Y. Kiselev

; The main module which does the following:

.386p
.seq

include define.inc
include common.inc
include gdt.inc
include idt.inc

include server32.inc
include user.inc

stack16_size	EQU 500h	; 500h bytes for 16-bit server's stack
server_stack16 segment para public use16
	db stack16_size dup (0)
server_stack16 ends
SWAP_FILE_SIZE	EQU 1024	; in Kb ! (align 4Kb)

server_data32 segment para public use32
gdt_reg	df 0			; Value for GDTR
idt_reg	df 0			; Value for IDTR
old_idt df 0

real_stack dd ?
exit_code db 0
Proc_type db 0
real_mode_buffer dw 0		; Segment address of the real mode transfer
				; buffer

my_iret_addr dw 0,0	; logical address of an IRET instruction
			; for V86 interrupt handling
my_iret_linear_addr dd 0; linear address of the IRET instruction

PSP_seg	dw 0		; PSP segment
CtrlC	db 0

XMS_present	EQU 1
VCPI_present	EQU 2
EMS_present	EQU 4
HMA_exist	EQU 8
DPMI_present	EQU 10h

System_flags	dw 0
XMS_farcall	dd 0	; The address of the XMM service routine
EMM_dev_name	db 'EMMXXXX0'
DOS_free_mem	dw 0		; in Kb
Ext_free_mem	dw 0		; in 4Kb

Swap_file_handl	dw 0	; For virtual memory
RAM_pages	dw 0	; The amount of physical RAM pages
DISK_pages	dw 0	; The amount of pages in the swap file
pages_in_system dw 0	; Total amount of pages in virtual memory
page_tables	dw 0	; The amount of second level page tables
first_free_page	dw 0	; Heap top pointer (offset in page tables array)

page_use_stat_address	dd 0	; Linear address of the page statistic table
index_table_address	dd 0	; for the swap file
page_tables_address	dd 0
page_dir_address	dd 0
swap_buff_address	dd 0	; of the buffer for swapping pages

gdt_address	dd 0
idt_address	dd 0
; DEX can use up to 4 free extended memory blocks (that is, what's left
; after other software, e.g. SmartDrive or RAMDrive, is installed).
; This limitation is not obligatory, but DEX is not supposed to work in
; a hard-fragmented environment.
mem_blocks      db 0		; Free memory blocks info
		dq 2 dup (0)	; 2 block (1 DOS & 1 XMS)

move_convtoext	dq 6 dup (0)	; reserved for building GDT (for moving a block
				; from conventional to extended memory
				; via int 15h)

tmp_path	db '.',0,14 dup (?)
mess_v86 	db 'Cannot work in V86 mode. Remove memory managers from CONFIG.SYS',10,13,'$'
mess_no386 	db '386+ not available.',10,13,'$'
mess_nomem 	db 'Not enough memory below 1Mb',10,13,'$'
mess_noextmem 	db 'Not enough memory above 1Mb',10,13,'$'
mess_A20_err	db 'Cannot enable A20 line.',10,13,'$'
mess_error_swp	db 'Cannot create swap file.',10,13,'$'
mess_err_table	db 'Cannot create system tables.',10,13,'$'

ok_mess 	db 'Hallo! Here I am still alive!','$'

old_int15	dw 0,0

data32_size	= ($-server_data32)-1
server_data32 ends

server_code16 segment para public use16
assume cs:server_code16, ds:server_data32
start:
		mov ax,server_stack16
		mov ss,ax
		mov sp,stack16_size

		mov ax,server_data32
		mov ds,ax
		mov ax,es
		mov PSP_seg,ax

		mov bx,seg zero_seg
		sub bx,ax
		inc bx  		; BX = program size
		mov ah,4ah
		int 21h			; resize program memory block

		mov ah,48h
		mov bx,4096
		int 21h                 ; allocate buffer
					; for preparing system tables
		jnc memok		; and data transfer between real
					; and protected modes
		mov dx,offset mess_nomem
		jmp error_exit
memok:
		mov real_mode_buffer,ax
		push ax
		mov ah,48h
		mov bx,0ffffh
		int 21h                 ; get max free DOS block
		shr bx,6		; in Kb
		mov DOS_free_mem,bx
		mov mem_blocks,1        ; The 1st mem_block is free DOS memory
		mov word ptr mem_blocks+1,bx
		mov word ptr mem_blocks+3,-1
		pop ax
		and eax,0ffffh
		shl eax,4
		mov dword ptr mem_blocks+5,eax

		call CheckProcessor	; check for 386+
		cmp al,3
		jnb _386h		; OK, the processor is 386+
		mov dx,offset mess_no386
		jmp error_exit		; Can't work with 286-
_386h:
		mov Proc_type,al

		smsw ax
		test ax,1		; Are we in V86?
		jz real_mode		; No
		mov dx,offset mess_V86
		jmp error_exit		; We're already in V86. Can't go on.
real_mode:
		call Check_environment	; Check for presence of memory managers
		call Allocate_memory	; Allocate all free system memory
		mov dx,offset mess_noextmem
		cmp mem_blocks,1
		je error_exit

		cmp word ptr mem_blocks+9,256	; Extended memory
						; block must be >= 256Kb
		jb error_exit			; Not enough
		call Create_Tables
		mov dx,offset mess_err_table
		jc error_exit
		call Create_Swap_File
		mov dx,offset mess_error_swp
		jc error_exit
		call Int_handling		; Intercept interrupts

		mov eax,seg back_from_v86int	; Calculate the linear
						; address of my magic IRET
						; (see Section 'Interrupts')
		mov my_iret_addr+2,ax
		shl eax,4
		mov ebx,offset back_from_v86int
		mov my_iret_addr,bx
		add eax,ebx
		inc eax			; points to next instruction
		mov my_iret_linear_addr,eax

		mov word ptr real_stack+2,ss
		mov word ptr real_stack,sp

		sidt old_idt		; Save old IDTR
		mov eax,gdt_address
		mov dword ptr gdt_reg+2,eax
		mov word ptr gdt_reg,gdt_size

		mov eax,idt_address
		mov dword ptr idt_reg+2,eax
		mov word ptr idt_reg,idt_size

		call get_A20_state
					; AL = 1 if A20 enable
		shl ax,15
		or System_flags,ax	; Save A20 state
		test ah,80h
		jnz A20_active

		call Enable_A20

		call get_A20_state
		mov dx,offset mess_A20_err
		test al,1
		jz error_exit		; I can't enable A20
A20_active:
		mov bx,5070h
		call SetPIC		; Redirect IRQ0-IRQ7 to int50-int57

		cli

		lgdt gdt_reg	; Load GDTR
		lidt idt_reg	; Load IDTR

		mov eax,cr3	; Get CR3
		and eax,0fffh
		or eax,page_dir_address
		mov cr3,eax	; Set CR3 (PDBR)

		mov eax,cr0
		or eax,80000001h
		mov cr0,eax             ; TURN paging and protection ON

		db 66h
		db 0eah
		dd offset entry32
		dw code32_sel		; Transfer control to 32-bit Kernel

; The entry point to real mode (when DEX is finished working)
back_to_real_world:

		cli
		mov ax,gdt_sel
		mov es,ax
		mov di,code16_sel
		mov word ptr es:[di],0ffffh	; code16 limit must be 0ffffh

		db 0eah      	; Refresh the prefetcher's queue
		dw offset $+4
		dw code16_sel

		; Load segment registers for real mode segments
		mov ax,real_mode_buffer_sel
		mov ds,ax
		mov es,ax
		mov gs,ax
		mov fs,ax
		mov ss,ax

		mov eax,cr0
		and eax,07ffffff2h	; Full off
		mov cr0,eax

		xor eax,eax
		mov cr3,eax

		db 0eah			; Jump to the next command
		dw offset $+4
		dw server_code16

		mov ax,server_data32	; Restore DS
		mov ds,ax
		lss sp,real_stack	; Restore stack pointer

		lidt old_idt		; Restore IDTR

		mov bx,0870h
		call SetPIC		; Restore IRQs
		sti

		test System_flags,8000h
		jnz ndA20

		call Disable_A20
ndA20:
		call Int_restoring
		call free_memory
		call delete_swap_file

		mov ah,9
		mov dx,offset ok_mess
		int 21h
exit_prot:
		mov al,exit_code
		mov ah,4ch
		int 21h			; Gameover !
error_exit:
		mov ah,9
		int 21h
		jmp exit_prot

;------------------------------------------------
; Magic IRET
back_from_v86int:
		iret	; catch return from a V86 interrupt
			; if not via IRET (e.g., RET 2 or something)
;------------------------------------------------
; Release XMS memory
free_memory	proc near
		test System_flags,XMS_present
		jz no_alloc
		mov si,offset mem_blocks
		add si,9
		mov dx,[si+2]
		mov ah,0dh
		call XMS_farcall
		dec ax
		mov ah,0ah
		call XMS_farcall
		dec ax
no_alloc:
		ret
free_memory	endp

Check_environment proc near
		mov ax,4300h
		int 2fh			; XMS installation check
		cmp al,80h
		jne no_XMS
		or System_flags,XMS_present
		mov ax,4310h
		int 2fh			; Get XMS call address
		mov word ptr XMS_farcall,bx
		mov word ptr XMS_farcall+2,es
		mov ah,0
		call XMS_farcall	; Get XMS version
		test dl,1
		jz no_hma
		or System_flags,HMA_exist
no_hma:
no_XMS:
		mov ah,35h	; EMM installation check
		mov al,67h
		int 21h
		mov di,10
		mov si,offset EMM_dev_name
		mov cx,2
		repe cmpsd
		jnz no_EMM
		or System_flags,EMS_present

		mov ax,0de00h	; VCPI installation check
		int 67h
		or ah,ah
		jnz no_VCPI
		or System_flags,VCPI_present
no_VCPI:
no_EMM:
		ret
Check_environment endp

; Occupy XMS memory
Allocate_memory	proc near
		mov di,offset mem_blocks + 9	; +9 (first DOS free)
		cmp System_flags,0
		jne alloc_mem
		mov ah,88h              ; No memory managers present
		int 15h			; Get free extended memory
					; via BIOS 'top-down' protocol
		mov [di],ax
		mov word ptr [di+2],-1
		mov dword ptr [di+4],100000h

		shr ax,2
		mov Ext_free_mem,ax	; in 4Kb pages
		inc mem_blocks
		ret
alloc_mem:
		test System_flags,XMS_present
		jz err_mem
					; get the largest
					; extended memory block
		mov ah,8
		call XMS_farcall	; Get free extended memory
		shr dx,2
		mov Ext_free_mem,dx	; in 4Kb pages
		shl dx,2
		mov [di],dx
		mov ah,9
		call XMS_farcall	; allocate ext mem block
		dec ax
		jnz err_mem
		mov [di+2],dx
		mov ah,0ch
		call XMS_farcall	; and lock ext mem block
		dec ax
		jnz err_mem
		mov [di+4],bx		; DX:BX = 32-bit block address
		mov [di+6],dx

		inc mem_blocks		; block counter
all_mem:
		sub Ext_free_mem,60h	; -384 kb (UMB)
		clc
		ret
err_mem:
		stc
		ret
Allocate_memory	endp

; Create and fill up GDT and page dir+tables
Create_Tables	proc near
		call Fill_Tables	; page dir+tables
		mov ah,48h
		mov bx,3000
		int 21h   		; Temporarily allocate 48Kb for GDT
		jc tablerror

		push ax
		push ax
		mov es,ax
		mov fs,real_mode_buffer
		; create GDT
		call fill_GDT

		; move page tables & dir into memory above 1Mb
		mov dx,real_mode_buffer
		mov edi,page_tables_address
		movzx eax,page_tables
		inc ax				; move tables and page dir
		shl ax,12			; AX = byte counter
		call move_memory
		jc tablerror

		; move GDT into memory above 1Mb
		xor edx,edx
		pop dx
		mov edi,gdt_address
		mov ax,gdt_size
		call move_memory
		jc tablerror

		pop es
		mov ah,49h
		int 21h		; Free the temporary GDT buffer

		clc
		ret
tablerror:
		stc
		ret
Create_Tables	endp

Create_Swap_File proc near
		mov ah,5ah
		mov dx,offset tmp_path
		xor cx,cx
		int 21h			; create temporary file
		jc swp_error

		mov Swap_file_handl,ax
		mov bx,ax
		mov si,DISK_pages       ; Write swap file with rubbish
		xor dx,dx		; up to (DISK_pages * 4Kb)
crnext_page:
		mov ah,40h
		mov cx,1000h
		int 21h
		jc swp_error
		cmp ax,1000h
		jne swp_error
		dec si
		jnz crnext_page
		clc
		ret
swp_error:
		stc
		ret
Create_Swap_File endp

delete_swap_file proc near
		mov ah,3eh
		mov bx,Swap_file_handl
		int 21h			; close swap file

		mov ah,41h
		mov dx,offset tmp_path
		int 21h			; delete swap file
		ret
delete_swap_file endp

; All tables are prepared in real mode (it's easier to debug)
Fill_Tables	proc near
		call get_system_memory	; total system RAM in Kb

		shr ax,2	; 4Kb pages counter
		mov es,real_mode_buffer
		xor di,di
		mov RAM_pages,ax
		mov cx,ax
		mov ax,SWAP_FILE_SIZE SHR 2	; swap file size in 4kb
		mov DISK_pages,ax
		add Ext_free_mem,ax	; common size of VM in 4Kb

		add ax,cx
		mov pages_in_system,ax
		shr ax,10
		adc al,0	; AX = page directory entries (page tables)
		mov page_tables,ax
		shl ax,12
		mov si,ax
		mov eax,0a07h	; fill Page Table entries (for RAM pages)
				; with attributes (default setting):
				; Present (bit 0 = 1)
				; Writable (bit 1 = 1)
				; User (bit 2 = 1)
				; Occupied (bit 9 = 1)
				; not Discardable (bit 11 = 1)
				; not On Disk (bit 10 = 0)
next_page:
		mov es:[di],eax
		add di,4
		add eax,1000h
		dec cx
		jnz next_page

		mov cx,DISK_pages
		mov eax,0fffff000h
				; fill Page Table entries (for DISK pages)
				; with attributes:
				; not Present (bit 0)
				; Address within the swap file
				; is initialized with -1.
		rep stosd

;----------- Mark all pages in free blocks as Free and Discardable --------

		mov bx,offset mem_blocks
		mov cl,[bx]
		inc bx
next_mblock:
		mov edi,[bx+4]
		test di,0fffh
		jz pagealigned
		add edi,1000h
pagealigned:
		shr edi,12
		mov ax,[bx]
		shr ax,2	; block size in pages
make_free_nextp:
		and word ptr es:[edi*4],0f5ffh	; free and discardable
		inc edi
		dec ax
		jnz make_free_nextp

		add bx,8
		dec cl
		jnz next_mblock

;------------------ allocate page tables in memory above 1Mb ------------

		mov cx,pages_in_system
		sub cx,100h
		mov di,400h		; above 1Mb (first page is 256)
chk_nextp:
		test word ptr es:[di],200h	; lookig for first free page
		jz freepage
		add di,4
		dec cx
		jnz chk_nextp
freepage:
		mov eax,es:[di]
		and ax,0f000h
		mov page_tables_address,eax

		mov cx,page_tables
		sub Ext_free_mem,cx		; adjust Ext_free_mem
lock_nextpage:
		mov eax,es:[di]
		mov es:[si],eax
		or word ptr es:[di],0a00h	; not discardable,
						; writable page tables
		and byte ptr es:[di],0bh        ; supervisor
		add di,4
		add si,4
		dec cx
		jnz lock_nextpage

;---------------------- allocate one page for page dir table -------------

		mov eax,es:[di]
		and ax,0f000h
		mov page_dir_address,eax	; for load in CR3

		or word ptr es:[di],0a00h	; not discardable,
						; writable page dir
		and byte ptr es:[di],0bh	; supervisor
		add di,4
		dec Ext_free_mem		; adjust Ext_free_mem

;----------------- allocate memory for collecting statistics -------------------
;		   about page usage

		mov eax,es:[di]
		and ax,0f000h
		mov page_use_stat_address,eax
		mov cx,page_tables
		sub Ext_free_mem,cx		; adjust Ext_free_mem
allocate_stat:
		or word ptr es:[di],0a00h	; non discard,write page tables
		and byte ptr es:[di],0bh        ; supervisor
		add di,4
		dec cx
		jnz allocate_stat

;----------------- allocate buffer for page swapping  -------------------

		mov eax,es:[di]
		and ax,0f000h
		mov swap_buff_address,eax
		mov cx,page_tables
		sub Ext_free_mem,cx		; adjust Ext_free_mem
allocate_buff:
		or word ptr es:[di],0a00h	; not discardable,
						; writable
		and byte ptr es:[di],0bh	; supervisor
		add di,4
		dec cx
		jnz allocate_buff

;--------------------- allocate memory for index table -------------------

		mov eax,es:[di]
		and ax,0f000h
		mov index_table_address,eax
		mov cx,DISK_pages
		or cx,cx
		jz no_vm
		shr cx,10
		inc cx
		sub Ext_free_mem,cx		; adjust Ext_free_mem
allocate_discp:
		or word ptr es:[di],0a00h	; not discardable,
						; writable
		and byte ptr es:[di],0bh        ; supervisor
		add di,4
		dec cx
		jnz allocate_discp
no_vm:
		mov first_free_page,di
		ret
Fill_Tables	endp

; Prepare GDT
fill_GDT	proc near
		mov cx,2000h
		xor eax,eax
		xor di,di
		rep stosd		; all 4096 descriptors free

;----------- allocate GDT above 1Mb ---------------------
		mov di,first_free_page
		mov eax,fs:[di]
		and ax,0f000h
		mov gdt_address,eax
		mov edx,eax
		mov cx,gdt_size/1000h
		sub Ext_free_mem,cx		; adjust Ext_free_mem
lock_gdtpages:
		or word ptr fs:[di],0a00h	; not discardable,
						; writable
		and byte ptr fs:[di],0bh	; supervisor
		add di,4
		dec cx
		jnz lock_gdtpages

		; GDT alias descriptor
		mov bx,gdt_sel
		mov ecx,gdt_size
		mov al,ALLOC_SEG
		mov ah,RW_DATA
		call fill_descriptor

;------------- allocate IDT above 1MB ------------------
		mov eax,fs:[di]
		and ax,0f000h
		mov idt_address,eax
		mov edx,eax
		or word ptr fs:[di],0a00h	; not discardable,
						; writable
		and byte ptr fs:[di],0bh        ; supervisor
		add di,4
		dec Ext_free_mem		; adjust Ext_free_mem

		; IDT descriptor
		mov bx,idt_sel
		mov ecx,idt_size
		mov al,ALLOC_SEG
		mov ah,RW_DATA
		call fill_descriptor

		; allocate TSS for server
		mov eax,fs:[di]
		and ax,0f000h
		mov edx,eax		; serv_tss_address
		push eax
		mov cx,tss_size/1000h
		sub Ext_free_mem,cx		; adjust Ext_free_mem
lock_tsspages:
		or word ptr fs:[di],0a00h	; not discardable,
						; writable
		and byte ptr fs:[di],0bh        ; supervisor
		add di,4
		dec cx
		jnz lock_tsspages

		; Cretae server's TSS
		mov bx,serv_tss
		mov ecx,tss_size
		mov al,ALLOC_SEG
		mov ah,TSS
		call fill_descriptor

		; server's TSS alias descriptor
		mov bx,serv_tss_sel
		pop edx				; serv_tss_address
		mov ecx,tss_size
		mov al,ALLOC_SEG
		mov ah,RW_DATA
		call fill_descriptor

		; allocate 4Kb for PL0 stack
		mov eax,fs:[di]
		and ax,0f000h
		mov edx,eax			; serv_tss_address
		or word ptr fs:[di],0a00h	; non-Discard,
						; writable
		and byte ptr fs:[di],0bh        ; supervisor
		add di,4
		dec Ext_free_mem		; adjust Ext_free_mem

		mov first_free_page,di

		; 32-bit server stack descriptor
		mov bx,stack32_sel
		mov ecx,stack32_size
		mov al,ALLOC_SEG
		mov ah,RW_DATA
		call fill_descriptor

		; 32-bit server code descriptor
		mov bx,code32_sel
		mov edx,server_code32
		shl edx,4
		mov ecx,code32_size
		mov al,BIG_SEG or ALLOC_SEG
		mov ah,ER_CODE
		call fill_descriptor

		; 32-bit server data descriptor
		mov bx,data32_sel
		mov edx,server_data32
		shl edx,4
		mov ecx,data32_size
		mov al,ALLOC_SEG
		mov ah,RW_DATA
		call fill_descriptor

		; 16-bit server code descriptor
		mov bx,code16_sel
		mov edx,server_code16
		shl edx,4
		mov ecx,code16_size
		mov al,ALLOC_SEG
		mov ah,ER_CODE
		call fill_descriptor

		; video buffer descriptor (color monitor assumed)
		mov bx,vtext_sel
		mov edx,0b8000h
		mov ecx,8000h-1
		mov al,ALLOC_SEG
		mov ah,RW_DATA OR DPL3
		call fill_descriptor

		; selector for transfer buffer
		mov bx,real_mode_buffer_sel
		movzx edx,real_mode_buffer
		shl edx,4
		mov ecx,0ffffh
		mov al,ALLOC_SEG
		mov ah,RW_DATA
		call fill_descriptor

		; flat selector (for the whole linear address space)
		mov bx,flat_sel
		xor edx,edx
		mov ecx,0fffffh
		mov al,BIG_SEG or ALLOC_SEG or BIG_GRANUL
		mov ah,RW_DATA
		call fill_descriptor

		; make PSP descriptor
		mov bx,PSP_sel
		movzx edx,PSP_seg
		shl edx,4
		mov ecx,256		; limit 100h
		mov al,ALLOC_SEG
		mov ah,RW_DATA OR DPL3
		call fill_descriptor

		; make Call Gates for Allocate_segment &
		; Free_segment

		mov bx,alloc_seg_gate
		mov eax,offset allocate_segment
		mov cx,code32_sel
		mov dl,PRESENT OR DPL3 OR 2
		call make_callgate

		mov bx,free_seg_gate
		mov eax,offset free_segment
		mov cx,code32_sel
		mov dl,PRESENT OR DPL3 OR 1
		call make_callgate

		; The last descriptor, the type field is set to 0DDh
		mov bx,last_desc
		mov word ptr es:[bx+5],00ddh
		ret
fill_GDT	endp

; move memory block from conv to ext memory
; AX - byte counter
; DX - source block segment (below 1Mb) to move
; EDI - destination address (above 1Mb) in
;       the first(!) free extended memory block
move_memory	proc near
		mov si,offset move_convtoext
		test System_flags,XMS_present
		jz move_int15

		; Work via XMM
		mov [si],eax
		mov word ptr [si+4],0
		shl edx,16
		mov [si+6],edx
		mov ax,word ptr mem_blocks+11		; first handle
		mov [si+10],ax
		sub edi,dword ptr mem_blocks+13	; page aligned
		mov [si+12],edi
		mov ah,0bh
		call XMS_farcall
		dec ax
		jnz move_err

		ret

move_int15:
		; Work via Int 15h
		shr ax,1			; word counter
		push ax
		push di
		mov ax,ds
		mov es,ax
		mov di,si
		xor eax,eax
		mov cx,12
		rep stosd
		pop di
		pop cx

		mov ax,-1
		mov [si+16],ax
		shl edx,4
		mov [si+18],edx
		mov byte ptr [si+21],93h

		mov [si+24],ax
		mov [si+26],edi
		mov byte ptr [si+29],93h

		mov ah,87h
		int 15h
		jc move_err

		clc
		ret
move_err:
		stc
		ret
move_memory	endp

Int_handling	proc near

		; Intercept Int 15h
		mov ax,3515h
		int 21h
		mov old_int15,bx
		mov old_int15+2,es

		push ds
		push cs
		pop ds
		mov dx,offset new_int15
		mov ax,2515h
		int 21h

		; Intercept Int 23h
		mov dx,offset new_int23
		mov ax,2523h
		int 21h
		; Intercept Int 24h
		mov dx,offset new_int24
		mov ax,2524h
		int 21h
		pop ds
		ret
Int_handling	endp

Int_restoring	proc near
		push ds
		lds dx,dword ptr old_int15
		mov ax,2515h
		int 21h
		pop ds

		ret
Int_restoring	endp

; Int 15h emulation in V86 mode
new_int15	proc far
		push ax
		mov ax,15h
		int 0		; invoke PM via int 0
new_int15	endp

new_int23	proc far
		push ds
		push server_data32
		pop ds

		mov CtrlC,1
		pop ds
		iret
new_int23	endp

new_int24	proc far
		; terminate the functions always
		mov al,3
		iret
new_int24	endp

get_A20_state	proc near
		test System_flags,XMS_present
		jz no_XMS_serv
		mov ah,7
		call XMS_farcall	; get A20 state
					; AX = 1 if A20 enabled
		ret
no_XMS_serv:
		push ds
		cli
		mov ax,0ffffh
		mov es,ax
		mov di,10h	; ES:DI -> 0ffffh:0010h
		xor ax,ax
		mov ds,ax
		xor si,si	; DS:SI -> 0000:0000

		not word ptr [si]

		cmpsw		; 0000:[0000h] == 0ffffh:[0010h] ?
		je a20_disabled
		inc eax
a20_disabled:
		not word ptr [si]
		sti

		pop ds
		ret
get_A20_state	endp

Enable_a20	proc near
		test System_flags,XMS_present
		jz no_XMS_serv
		mov ah,5
		call XMS_farcall	; Local Enable A20
		ret
no_XMS_serv:
					; Enable A20 via keyboard controller
		cli
		call cmd_wait		; Wait until ready
		jnz kbd_err
		mov al,0d1h
		out 64h,al              ; Send 0D1h command (A20 control)

		call cmd_wait		; Wait until ready
		jnz kbd_err
		mov al,0dfh
		out 60h,al		; Send 0DFh command (Enable)

		call cmd_wait		; Wait until ready
kbd_err:
		sti
		ret
Enable_a20	endp

Disable_a20	proc near
		test System_flags,XMS_present
		jz no_XMS_serv
		mov ah,6
		call XMS_farcall	; Local disable A20
		ret
no_XMS_serv:
		; Disable A20 via keyboard controller
		cli
		call cmd_wait		; Wait until ready
		jnz kbd_err

		mov al,0d1h
		out 64h,al		; Send 0D1h command (A20 Control)
		call cmd_wait
		jnz kbd_err

		mov al,0ddh
		out 60h,al		; Disable A20
		call cmd_wait
kbd_err:
		sti
		ret
Disable_a20	endp

cmd_wait	proc near
		mov cx,-1
w:
		in al,64h		; Get keybord controller status
		test al,2
		loopnz w 		; Wait until ready
		ret
cmd_wait	endp

code16_size	= ($-server_code16)-1
server_code16	ends

zero_seg segment
zero_seg ends
end start
