; The protected mode kernel (Chapter 16 Assembly Language Master Class)
; Written by Y. Kiselev

; The 32-bit part of the kernel

include tss.inc
include memman.inc

items_in_LDT	= 4096			; max number of segments for user task
ldt_size	= items_in_LDT*8	; 32KB

stack32_size	EQU 1000h-1		; size of PL0 stack

server_data32 segment para public use32
user_tss_alias	dd 0
user_tss	dd 0
user_ldt_alias	dd 0
user_ldt	dd 0
udata32_sel	dd 0
ucode32_sel	dd 0
ustack32_sel	dd 0
to_user		df 0
server_data32 ends

assume ds:server_data32
server_code32 segment para public use32
entry32:
		; Fill up segment registers
		mov eax,stack32_sel
		mov ss,eax
		mov esp,stack32_size+1
		mov eax,data32_sel
		mov ds,eax
		mov fs,eax
		mov gs,eax
		call fill_IDT

		mov ax,serv_tss_sel
		mov es,eax

		xor edi,edi
		xor eax,eax
		mov ecx,800h
		rep stosd		; clear I/O Permission BitMap
		dec eax
		mov es:[edi],al		; 0ffh - end

		pushfd
		pop eax
		and eax,not (IOPL3 or NTFlag)	; NT = 0
		or ax,IOPL2                     ; IOPL = 2
		push eax
		popfd

		mov ax,serv_tss_sel
		mov ds,ax
		xor ebx,ebx
		mov [ebx].tss_str.IO_ofs,104

		mov [ebx].tss_str.esp_0,esp
		mov [ebx].tss_str.ss_0,ss

		mov ax,serv_tss
		ltr ax		; Load TR to point at the server's TSS
		sti

		mov eax,data32_sel
		mov ds,eax

		call Init_mem_manager
		call create_user
		jmp leave32

;--------------- Create user task  ------------------------
create_user	proc near

		; Allocate memory block for LDT
		push ldt_size/1000h
		push TRUE		; non Discardable
		push cs
		call allocate_segment	; in EAX - allocated selector
					; (LDT alias descriptor)
		jc cu_error
		mov user_ldt_alias,eax

		mov es,eax
		mov word ptr es:[5],00ddh; mark first as last descriptor

		;make LDT desriptor in GDT

		mov eax,GDT_sel		; in GDT
		call allocate_descr	; EAX=selector
		jc cu_error
		mov user_ldt,eax
					; LDT and TSS descriptors must
					; immediately follow their alias
					; descriptors in GDT!
					; So the kernel can easily find
					; the alias for an LDT or TSS
					; descriptor

		; Copy the preceeding descriptor from the GDT
		; and change the type field for LDT
		mov eax,GDT_sel
		mov es,eax
		mov edi,user_ldt_alias
		mov eax,es:[edi]
		mov edx,es:[edi+4]
		mov edi,user_ldt
		mov dh,LDT		; change segment type (LDT)
		mov es:[edi],eax
		mov es:[edi+4],edx

		; Allocate TSS
		push tss_size/1000h
		push TRUE		; non Discardable
		push cs
		call allocate_segment	; in EAX - allocated selector
					; (LDT alias descriptor)
		jc cu_error
		mov user_tss_alias,eax

		;make TSS desriptor
		mov eax,GDT_sel		; in GDT
		call allocate_descr	; in EAX selector
		jc cu_error
		mov user_tss,eax
		mov word ptr to_user+4,ax

		; Copy the preceeding descriptor and
		; change the type field for TSS
		mov edi,user_tss_alias
		mov eax,es:[edi]
		mov edx,es:[edi+4]
		mov edi,user_tss
		mov dh,TSS		; change segment type (TSS)
		mov es:[edi],eax
		mov es:[edi+4],edx

		; fill up LDT
		mov eax,user_ldt_alias
		mov es,eax

		mov eax,es		; in LDT
		call allocate_descr	; in EAX selector
		jc cu_error
		or al,RPL3 or TI	; Ring 3, in LDT
		mov ucode32_sel,eax

		; make user 32-bit code descriptor
		mov ebx,eax
		mov edx,user_code
		shl edx,4
		mov ecx,user_code_size
		mov al,BIG_SEG OR ALLOC_SEG
		mov ah,ER_CODE or DPL3
		call fill_descriptor32

		mov eax,es		; in LDT
		call allocate_descr	; in EAX selector
		jc cu_error
		or al,RPL3 or TI
		mov udata32_sel,eax

		; make user 32-bit data descriptor
		mov ebx,eax
		mov edx,user_data
		shl edx,4
		mov ecx,user_data_size
		mov al,ALLOC_SEG
		mov ah,RW_DATA or DPL3
		call fill_descriptor32

		mov eax,es		; in LDT
		call allocate_descr	; in EAX selector
		jc cu_error
		or al,RPL3 or TI
		mov ustack32_sel,eax

		; make user 32-bit stack descriptor
		mov ebx,eax
		mov edx,user_stack
		shl edx,4
		mov ecx,user_stack_size
		mov al,ALLOC_SEG
		mov ah,RW_DATA or DPL3
		call fill_descriptor32

		mov es,user_tss_alias
		xor edi,edi
		xor eax,eax
		mov ecx,800h
		rep stosd		; clear I/O Permission BitMap
		dec eax
		mov es:[edi-4],al	; 0ffh - end

		; Fill up user TSS
		push ds
		mov eax,ds
		mov es,eax

		mov ds,user_tss_alias
		xor ebx,ebx

		mov [ebx].tss_str.IO_ofs,104

		mov eax,es:user_ldt
		mov [ebx].tss_str.ldt_,eax
		mov [ebx].tss_str.gs_,ebx
		mov [ebx].tss_str.fs_,ebx
		mov eax,es:udata32_sel
		mov [ebx].tss_str.ds_,eax

		mov [ebx].tss_str.es_,PSP_sel

		mov eax,es:ucode32_sel
		mov [ebx].tss_str.cs_,eax

		mov eax,es:ustack32_sel
		mov [ebx].tss_str.ss_,eax

		mov eax,cr3
		mov [ebx].tss_str.cr3_,eax

		mov [ebx].tss_str.esp_,user_stack_size-4
		mov [ebx].tss_str.eip_,0
		mov [ebx].tss_str.eflags_,200h
		mov [ebx].tss_str.esp_0,esp
		mov [ebx].tss_str.ss_0,ss
		mov [ebx].tss_str.esp_1,esp
		mov [ebx].tss_str.ss_1,ss
		mov [ebx].tss_str.esp_2,esp
		mov [ebx].tss_str.ss_2,ss
		pop ds

		jmp to_user			; Switch to user task
cu_error:
		ret
create_user	endp

;--------------- exit the 32-bit kernel ------------
leave32:
		db 0eah
		dd offset back_to_real_world
		dw code16_sel

;------------- go here if any exception has occured -----------------
;	       put debug message and get out to real mode

something_wrong:
		push cs
		call print_debug_info

		mov ax,data32_sel
		mov ds,ax
		mov eax,2
		mov ebx,eax
		mov cl,9
		call disp_string

		jmp leave32

; Allocate a descriptor in the GDT or LDT
; EAX=selector for the GDT or LDT
allocate_descr	proc near
		push ds
		mov ds,eax

		mov ecx,items_in_LDT-1
		xor ebx,ebx
		cmp ax,GDT_sel
		jne alloc_in_ldt
		mov ecx,items_in_GDT-1
		add bl,8
alloc_in_ldt:
		test byte ptr [ebx+6],ALLOC_SEG
		jz ok_find
skip_gates:
		add ebx,8
		dec ecx
		jnz alloc_in_ldt

		stc
		pop ds
		ret
ok_find:
		cmp byte ptr [ebx+5],0ddh
		jne no_last_descr
		mov word ptr [ebx+8+5],00ddh	; next descriptor will be
						; the last (and free) descriptor now
		jmp short filldsc
no_last_descr:
		test byte ptr [ebx+5],4
		jnz skip_gates
filldsc:
		or byte ptr [ebx+6],ALLOC_SEG
		mov eax,ebx
		clc
		pop ds
		ret
allocate_descr	endp

code32_size	= ($-server_code32)-1
server_code32 ends

