; The protected mode kernel (Chapter 16 Assembly Language Master Class)
; Written by Y. Kiselev

; Software interrupts emulator
; This module translates only a limited set of DOS and BIOS interrupts.
; Its purpose was to let the kernel work (e.g.,it allows to work with files).
; However, almost all interrupts with register-based parameter transfer are
; supported.
; SOFT_INT doesn't check the destination selector:offset combination for
; validity. If, for example, the selector refers to a nonexisting segment,
; the kernel will GPF.

server_data32 segment para public use32

; This table translate interrupt number to the specific emulator offset
int_dispatch	dd offset int10h
		dd -1
		dd -1
		dd offset int13h
		dd -1
		dd offset int15h
		dd -1
		dd -1
		dd -1
		dd offset int19h
		dd -1
		dd 5 dup (0)	; skip 1b,1c,1d,1e,1fh
		dd offset int20h
		dd offset int21h
		dd 3 dup (0)	; skip 22,23,24
		dd 2 dup (offset int2526)
		dd 2 dup (0)    ; skip 27,28
		dd -1		; 29
		dd 5 dup (0)	; skip 2a,2b,2c,2d,2e
		dd offset int2fh
		dd 0
		dd offset int31h
		dd 0
		dd offset int33h
		dd 28 dup (0)	; 34h-4fh

int10_functions	EQU 13
; Table of supported Int 10h functions
int10_ref       dw 1002h		; set all palette registers
		dd offset func1002
		dw 1009h		; read all palette registers
		dd offset func1009
		dw 1012h		; set block of DAC registers
		dd offset func1012
		dw 1017h		; read block of DAC registers
		dd offset func1017
		dw 1100h		; load user font for text
		dd offset func1100
		dw 1110h
		dd offset func1110
		dw 1120h		; fonst for graphics
		dd offset func1120
		dw 1121h
		dd offset func1121
		dw 4f00h		;VESA function
		dd offset func4f00
		dw 4f01h
		dd offset func4f01
		dw 4f04h
		dd offset func4f04
		dw 1c00h		; video state
		dd offset func1c00
		dw 1c01h
		dd offset func1c01

; These are features codes of DOS functions (Int 21h):

; 0 - not implemented
; 1 - registers only used
; 2 - moveout cx bytes
; 3 - movein cx bytes
; 4 - moveout ASCIIZ bytes
; 5 - moveout $ bytes
; 6 - movein al bytes
; 7 - functions 25h,35h
; 8 - function 44h
; 9 - function 47h
; 10 - functions 48h,49h,4ah
; 11 - function 4bh
; 12 - function 4ch

int21table	db 0,1,1,1,1,1,1,1,1	; 0-8
		db 5,6			; 9,0ah
		db 1,1,1,1		; 0bh-0eh
		db 0,0,0,0,0,0,0,0,0,0	; 0fh-18h (no FCB)
		db 1,0,1,1,0,0,1	; 19h-1fh
		db 0,0,0,0,0		; 20h-24h
		db 7			; 25h
		db 0,0,0,0		; 26h-29h
		db 1,1,1,1,1,1,1	; 2ah-30h
		db 0,0			; 31h,32h
		db 1,1			; 33h,34h
		db 7			; 35h
		db 1,1,0		; 36h-38h
		db 4,4,4,4,4		; 39h-3dh
		db 1,3			; 3eh,3fh
		db 2,4,1,4		; 40h-43h
		db 8			; 44h
		db 1,1			; 45h,46h
		db 9			; 47h
		db 10,10,10		; 48h-4ah
		db 11			; 4bh
		db 12			; 4ch
		db 1,4,1,0,1,0,0,1,0	; 4dh-55h
		db 0			; 56h
		db 1,1,1,4,4,1		; 57h-5ch
		db 0,0,0,0,0,0,0,0,0	; 5dh-65h
		db 1,1,1		; 66h-68h
		db 100h-68h dup (0)	; no_int

int44table	db 1,1,3,2,3,2,1,1,1,1	; 0-9
		db 1,1			; 0ah,0bh
		db 0,0			; 0ch,0dh
		db 1,1,1,1		; 0eh-11h
		db 8 dup (0)

; Offsets of emulators for DOS functions with different features
int21ref	dd offset no_int
		dd offset reg
		dd offset moveout_cx_bytes
		dd offset movein_cx_bytes
		dd offset moveout_Z_bytes
		dd offset moveout_$_bytes
		dd offset movein_al_bytes
		dd offset fun_2535
		dd offset fun_44
		dd offset fun_47
		dd offset fun_48494a
		dd offset fun_4b
		dd offset fun_4c

rest_routs	dd offset rest_ebp
		dd offset rest_edi
		dd offset rest_edx
		dd offset rest_ebx
server_data32 ends

; The structure on the stack that describes the interrupt
; (created by INT emulator, destroyed by IRET emulator)
rest_reg_rout	EQU byte ptr [esp+35]	; Index of the routine that restores
					; the pointer register on return
hard_int	EQU byte ptr [esp+34]	; Hardware/software interrupt tag
copy_back	EQU byte ptr [esp+33]	; Data must be copied from V86
int_number	EQU byte ptr [esp+32]	; Interrupt index
byte_counter	EQU word ptr [esp+30]	; Data length in bytes
dest_sel        EQU word ptr [esp+28]	; Destination selector
dest_offset	EQU dword ptr [esp+24]	; Destination offset

server_code32 segment para public use32
emul_interrupt:
softint1:
		sub esp,12      	; allocate frame for returner
		push eax
		push ecx
		push edi
		push esi
		push ds
		push es

		mov ecx,data32_sel
		mov ds,ecx
		mov hard_int,FALSE	; Software interrupt tag
		mov rest_reg_rout,-1
		mov copy_back,FALSE	; no copy back

		mov es,[esp+28+12]	; get Old cs
		xor ecx,ecx
		mov edi,[esp+24+12]	; get Old eip
		mov cl,es:[edi-1]	; get intterrupt number

		mov edi,real_mode_buffer_sel
		mov es,edi

		mov int_number,cl
		cmp cl,4fh
		jae reg
		mov ecx,int_dispatch[ecx*4-40h]
		or ecx,ecx
		js reg
		jz no_int
		jmp ecx			; Go to the specific interrupt emulator
reg:
		pop es
		pop ds
		pop esi
		pop edi
		pop ecx
		pop eax
		jmp ref_to_v86		; Go to V86 switcher

no_int:
; Interrupt no supported
		pop es
		pop ds
		pop esi
		pop edi
		pop ecx
		pop eax
		add esp,12
		iretd

; ---------------------  int 10 begin --------------------------------
int10h:
		cmp ah,10h
		jb reg
		cmp ah,13h		; write string
		je func13
		cmp ah,1bh		; get video setup
		je func1b
		mov edi,offset int10_ref
		mov esi,int10_functions
next_func:
		cmp ax,[edi]
		je int10hit
		add edi,6
		dec esi
		jnz next_func
		jmp reg
int10hit:
		jmp dword ptr [edi+2]
func13:
		movzx ecx,word ptr [esp+16]
		test al,2
		jz no_attr
		shl ecx,1
no_attr:
movein_esbp:
		mov esi,ebp		; move data from
		xor edi,edi		; PM user buffer (ES:EBP) to
		mov ds,[esp]		; V86 buffer
		mov al,cl
		and al,3
		shr ecx,2
		rep movsd
		mov cl,al
		rep movsb
		mov rest_reg_rout,0	; 0 - returner restore EBP from
		mov dest_offset,ebp	; dest_offset
		xor ebp,ebp
		jmp reg
func1b:
		mov byte_counter,64
moveout_esdi:
		mov copy_back,TRUE		; copy table back
		mov eax,[esp]			; ES
		mov ecx,[esp+12]		; EDI
		mov dest_sel,ax
		mov dest_offset,ecx
		xor ecx,ecx
		mov rest_reg_rout,1	; 1 - returner restore EDI from dest_offset
		mov [esp+12],ecx        ; EDI = 0 (0 offset within V86 buffer)
		jmp reg
func1002:                               ; Set all palette registers
		mov ecx,17
		mov ds,[esp]		; DS = (ES)
moveout_dx:
		mov esi,edx		; move palette data from
		xor edi,edi		; PM user buffer (ES:EDX) to V86 buffer
		mov al,cl
		and al,3
		shr ecx,2
		rep movsd
		mov cl,al
		rep movsb
		mov rest_reg_rout,2	; 2 - returner restore EDX from
		mov dest_offset,edx	; dest_offset
		xor edx,edx		; DX = 0 (offset within real mode buffer)
		jmp reg
func1009:
		mov byte_counter,17
		mov eax,[esp]			; ES
movein_dx:      			; EAX - block selector to move !
		mov copy_back,TRUE		; copy table back
		mov dest_sel,ax
		mov dest_offset,edx		; EDX
		mov rest_reg_rout,2	; 2 - returner restore EDX from dest_offset
		xor edx,edx		; DX = 0 (offset within real mode buffer)
		jmp reg
func1012:       			; Set block of DAC registers
		mov ds,[esp]
		mov ecx,[esp+16]
		mov eax,ecx
		add ecx,eax
		add ecx,eax		; bytes = CX*3
		jmp moveout_dx
func1017:
		mov eax,[esp+16]
		shl eax,1
		add eax,[esp+16]
		mov byte_counter,ax

		mov eax,[esp]			; EAX - block selector to move (ES)
		jmp movein_dx
func1100:
func1110:
		mov eax,[esp+16]
		mul bh
		mov ecx,eax
		jmp movein_esbp
func1120:
		mov ecx,256*8
		jmp movein_esbp
func1121:
		mov ecx,[esp+16]
		shl ecx,8
		jmp movein_esbp
func4f00:				; VESA BIOS Extention
func4f01:
		mov byte_counter,256
		jmp moveout_esdi
func4f04:
		mov word ptr [esp+20],1000h	; function not supported
		jmp no_int
func1c00:
func1c01:
		mov word ptr [esp+20],0		; function not supported
		jmp no_int

; ---------------------  int 10 end --------------------------------

; ---------------------  int 13 begin --------------------------------
; emul only 2,3,4
; rest is reg (be aware 0ah,0bh,5 functions)
int13h:
		cmp ah,2
		je read_sector
		cmp ah,3
		je wrver_sector
		cmp ah,4
		je wrver_sector
		jmp reg
read_sector:
		and ax,00ffh
		shl eax,9		; *512 (!)
		mov byte_counter,ax
		mov copy_back,TRUE	; copy table back
		mov eax,[esp]		; ES
		mov dest_offset,ebx	; EBX
		mov dest_sel,ax
		mov rest_reg_rout,3	; 3 - returner restores EBX from dest_offset
		xor ebx,ebx		; BX = 0 (offset within real mode buffer)
		jmp reg
wrver_sector:
		movzx ecx,al
		shl ecx,9			; *512 (?!)
		mov esi,ebx		; move sector data from
		xor edi,edi		; PM user buffer (ES:EBX) to
		mov ds,[esp]		; V86 buffer
		mov al,cl
		and al,3
		shr ecx,2
		rep movsd
		mov cl,al
		rep movsb
		mov rest_reg_rout,3	; 3 - returner restores EBX from
		mov dest_offset,ebx	; dest_offset
		xor ebx,ebx		; BX = 0 (offset within real mode buffer)
		jmp reg
; ---------------------  int 13 end --------------------------------

int15h:
		cmp ah,83h
		jne no_setevent
no_setevent:
		cmp ah,87h
		jb reg
		cmp ah,89h
		ja reg
		mov byte ptr [esp+20],86h	; function no supported
		or byte ptr [esp+32+12],1	; CF = 1
		jmp no_int
int19h:
		;
		jmp no_int
int20h:
		;
		jmp no_int

; ---------------------  int 21h begin --------------------------------
int21h:
		movzx esi,ah
		movzx esi,byte ptr [esi+int21table]
		jmp [esi*4+int21ref]

moveout_cx_bytes:
		mov ds,[esp+4]
		movzx ecx,word ptr [esp+16]	; ECX
		and ecx,0ffffh
		jmp moveout_dx
movein_cx_bytes:
		mov ax,[esp+16]			; ECX
		mov byte_counter,ax

		mov eax,[esp+4]			; DS
		jmp movein_dx
moveout_Z_bytes:
		mov al,0
find_$:
		push es
		mov es,[esp+8]		; DS:EDX points to ASCIIZ string
		xor ecx,ecx
		mov edi,edx
		dec ecx
		repne scasb
		not ecx
		pop es
		mov ds,[esp+4]
		jmp moveout_dx
moveout_$_bytes:
		mov al,'$'
		jmp short find_$
movein_al_bytes:
		or al,al
		jz no_int

		mov es:[0],al
		mov ah,0
		add ax,2
		mov byte_counter,ax

		mov eax,[esp+4]			; DS
		jmp movein_dx
fun_2535:
		jmp reg
fun_44:
		movzx esi,al
		movzx esi,byte ptr [esi+int44table]
		jmp [esi*4+int21ref]
fun_47:
fun_48494a:
fun_4b:
		jmp reg
fun_4c:
		jmp leave32
; ---------------------  int 21h end --------------------------------

int2526:
		mov word ptr [esp+20],0100h	; function not supported
		or byte ptr [esp+32+12],1	; CF = 1
		jmp no_int
int2fh:
int31h:
int33h:
		jmp reg

; ------------------------------- RETURNER ---------------------------
to_returner:
		cmp byte ptr [esp+26],TRUE	; check hard_int flag
		je hard_int_end

		push ecx
		push ds
		cmp byte ptr [esp+33],TRUE	; check copy_back flag
		jne no_move
		; Copy from V86 buffer to the destination Sel:Ofs
		push eax
		push esi
		push edi
		push es
		mov eax,real_mode_buffer_sel
		mov ds,eax
		mov es,[esp+20+24]
		xor esi,esi
		mov edi,[esp+16+24]
		movzx ecx,word ptr [esp+22+24]
		mov al,cl
		and al,3
		shr ecx,2
		rep movsd
		mov cl,al
		rep movsb
		pop es
		pop edi
		pop esi
		pop eax
no_move:
		movzx ecx,byte ptr [esp+35]	;rest_reg_rout
		cmp cl,-1
		je soft_int_end
		push data32_sel
		pop ds
		jmp [ecx*4+rest_routs]
rest_ebp:
		mov ebp,[esp+24]
		jmp soft_int_end
rest_edi:
		mov edi,[esp+24]
		jmp soft_int_end
rest_ebx:
		mov ebx,[esp+24]
		jmp soft_int_end
rest_edx:
		mov edx,[esp+24]
soft_int_end:
		pop ds
		pop ecx
hard_int_end:
		pop es
		pop ds
		pop fs
		pop gs
		add esp,12
		iretd   		; return to PM user
server_code32 ends
