; A demo user program that works via the protected mode kernel
; (Chapter 16 Assembly Language Master Class)
; Written by Y. Kiselev

; This program a 'flip-top' distortion sound effect (designed by P.Kalatchin):
; signal peaks are flipped around the level specified by DEPTH (in %).
; Run: SERVER16.EXE INPUT.WAV OUTPUT.WAV
; .WAV files must be 8-bit mono. The whole input file is sucked up into
; memory, processed and written to the output file.
; It's just for fun (which is demonstration of the kernel).
; Don't expect that the sound will get clearer, brighter: it's a kind of
; strong non-linear distortion.
; The code of the algorithm itself is too cryptic to be commented, so
; it's given 'as is'.


; Allocate a segment
allocate_block	macro count,flag
		push count
		push flag
		db 09ah
		dd 0
		dw alloc_seg_gate
endm

; Deallocate a segment
free_block	macro sel
		push sel
		db 09ah
		dd 0
		dw free_seg_gate
endm

user_stack_size = 500
user_stack segment para public use32
resstk	db user_stack_size dup (0)
user_stack ends

user_data segment para public use32
Depth		EQU 70			; The effect level in %
					; The less the deeper

WF		dw 0			; Input file handle
OutF		dw 0			; Output file handle

FileName 	db 128 dup (0)
OutName  	db 128 dup (0)

; Miscelleneous variables
Lastchanged 	dd 0
file_size	dd 0
extrem		dd 0

extrlevel	db 0
cut_level	db 0

CurVal		dd 0
CurDer		dd 0
PrevVal		dd 0
PrevDer		dd 0

user_data_size	= ($-user_data )-1
user_data ends

user_code segment para public use32
assume ds:user_data

		; After the kernel switches to the user task,
		; the picture is like this:
		; DS points to data segment
		; ES points to PSP
		; FS = GS = 0

		; Find the command tail in the PSP
		; and assign two first string to filenames
		mov ebx,81h
		mov edi,offset FileName

		mov cl,es:[ebx-1]
		or cl,cl
		jz ferror
		mov ch,0
		mov dl,1
get_nchar:
		mov al,es:[ebx]
		cmp al,' '
		je skip_sp
		mov [edi],al
		mov ch,1
		inc edi
next_ch:
		inc ebx
		dec cl
		jnz get_nchar
		jmp short end_cmd_lin
skip_sp:
		or ch,ch
		jnz next_item
		jmp short next_ch
next_item:
		inc dl
		cmp dl,3
		je two_arg_ok
		mov edi,offset OutName
		mov ch,0
		jmp short get_nchar
end_cmd_lin:
		cmp dl,2
		jb ferror
two_arg_ok:

		; Create output file
		mov ah,3ch
		xor ecx,ecx
		mov edx,offset OutName
		int 21h
		jc ferror
		mov OutF,ax		; AX=handle

		; Open input file
		mov ah,3dh
		mov al,0
		mov edx,offset FileName
		int 21h
		jc ferror
		mov WF,ax		; AX=handle

		; Get size of the input file
		mov ah,42h
		mov bx,WF
		xor ecx,ecx
		xor edx,edx
		mov al,2
		int 21h
		jc ferror

		mov word ptr file_size,ax
		mov word ptr file_size+2,dx

		; Seek to the beginning of file
		mov ah,42h
		mov bx,WF
		xor edx,edx
		xor ecx,ecx
		mov al,0
		int 21h
		jc ferror

		; Allocate a buffer for the file
		mov eax,file_size
		shr eax,12		; /4096
		inc eax
		allocate_block	eax,FALSE
		jc ferror

		mov es,eax

		; Read the input file by 64000-byte blocks
		mov esi,file_size
		xor edx,edx
read_next_blk:
		mov ecx,64000
		cmp esi,ecx
		ja re
		mov ecx,esi
re:
		mov bx,WF
		push ds
		mov eax,es
		mov ds,eax
		mov ah,3fh		; Read file
		int 21h
		pop ds
		jc ferror
		cmp ax,cx
		jne ferror
		add edx,ecx
		sub esi,ecx
		jz end_r
		jmp read_next_blk
end_r:

		; Begin sound processing
		mov edi,45
		xor eax,eax
		mov LastChanged,44
		mov al,es:[edi+1]
		xor ebx,ebx
		mov PrevVal,eax
		mov bl,es:[edi]
		sub eax,ebx
		mov PrevDer,eax
repeate1:
		inc edi
		movzx eax,byte ptr es:[edi]
		mov CurVal,eax

		mov ebx,PrevVal
		sub eax,ebx
		mov CurDer,eax

		or eax,eax
		jnz go_
		mov eax,PrevDer
		mov CurDer,eax
go_:
		mov eax,CurDer
		imul PrevDer
		or eax,eax
		jns _loop

		movsx eax,byte ptr PrevVal
		imul CurDer
		or eax,eax
		jns go_1
_loop:
		mov eax,CurDer
		mov PrevDer,eax
		mov eax,CurVal
		mov PrevVal,eax
		jmp end_rep1
go_1:
		dec edi
		mov extrem,edi
		cmp PrevVal,128
		jb else_

		mov eax,PrevVal
		sub eax,128
		mov ebx,DEPTH
		mul ebx
		mov ebx,100
		div ebx
		add eax,128
		mov cut_level,al
next4:
		mov al,cut_level
		cmp es:[edi],al
		jb go_2
		cmp edi,LastChanged
		jb go_2

		movzx eax,cut_level
		xor ebx,ebx
		shl eax,1
		mov bl,es:[edi]
		sub eax,ebx
		mov es:[edi],al
		dec edi

		jmp short next4
go_2:
		mov edi,extrem
		inc edi
next3:
		mov al,cut_level
		cmp es:[edi],al
		jb go_3
		mov eax,file_size
		dec eax
		cmp edi,eax
		ja go_3

		movzx eax,cut_level
		xor ebx,ebx
		shl eax,1
		mov bl,es:[edi]
		sub eax,ebx
		mov es:[edi],al
		inc edi
		jmp short next3
go_3:
		jmp end_rep
else_:
		mov eax,128
		mov ebx,PrevVal
		sub eax,ebx
		mov ebx,DEPTH
		mul ebx
		mov ebx,100
		div ebx
		mov ebx,128
		sub ebx,eax
		mov cut_level,bl
next2:
		mov al,cut_level
		cmp es:[edi],al
		jae go_4
		cmp edi,LastChanged
		jb go_4

		movzx eax,cut_level
		xor ebx,ebx
		shl eax,1
		mov bl,es:[edi]
		sub eax,ebx
		mov es:[edi],al
		dec edi
		jmp next2
go_4:
		mov edi,extrem
		inc edi
next1:
		mov al,cut_level
		cmp es:[edi],al
		jae go_5
		mov eax,file_size
		dec eax
		cmp edi,eax
		ja go_5

		movzx eax,cut_level
		xor ebx,ebx
		shl eax,1
		mov bl,es:[edi]
		sub eax,ebx
		mov es:[edi],al
		inc edi

		jmp short next1
go_5:
end_rep:
		xor eax,eax
		mov PrevDer,eax
		movzx eax,byte ptr es:[edi]
		mov PrevVal,eax
		mov LastChanged,edi
end_rep1:
		mov eax,file_size
		dec eax
		cmp edi,eax
		jb repeate1
		; End sound processing


		; Write Output File
		mov esi,file_size
		xor edx,edx
write_next_blk:
		mov ecx,64000
		cmp esi,ecx
		ja we
		mov ecx,esi
we:
		mov bx,OutF
		push ds
		mov eax,es
		mov ds,eax
		mov ah,40h		; Write file
		int 21h
		pop ds
		jc ferror
		cmp cx,ax
		jne ferror
		add edx,ecx
		sub esi,ecx
		jz end_w
		jmp write_next_blk
end_w:
		free_block es

		; Close output file
		mov ah,3eh
		mov bx,OutF
		int 21h

		; Close input file
		mov ah,3eh
		mov bx,WF
		int 21h
ferror:
		mov ah,4ch		; Good bye, protected mode!
		int 21h

user_code_size = ($-user_code)-1
user_code ends
