; This is a sample program for ExpAsm Chapter 'Pentium'
; This module contains Pentium-optimized low-level routines
; for calculation of Julia and Mandelbrot sets.
; Main module: EXPLORER.CPP

; Written by Kiselev Y. 1994

.model large
public _calc_one_point
public _calc_one_point_jul
.code
.386
.387
comment # This is the old 486 version of the procedure calc_one_point
;	Left in this file for comparison
;
;int "C" calc_one_point(double x,double y,int max_it,
;			double &im_sq,double &re_sq);
;
;
x		EQU qword ptr[bp+6]
y		EQU qword ptr[bp+14]
max_it		EQU word ptr[bp+22]
im_sq		EQU dword ptr[bp+24]
re_sq		EQU dword ptr[bp+28]

_calc_one_point	proc far
		push bp
		mov bp,sp

		xor dx,dx	; DX - IterNo
		mov cx,max_it

		fld1
		fadd st,st
		fst st(1)	; st = 2
		fmul st,st	; st = 4

		fldz
		fldz
		fldz
		fldz
next_iter:
		fld st
		fadd st,st(2)
		fcomp st(5)	; cmp 4
		fstsw ax
		sahf
		jnb short exit_while
		cmp dx,cx
		jg short exit_while

		inc dx

		;	im =  2*re*im + y;
		fld st(5)	; st <- 2
		fmul st,st(3)	;st*st(3) = 2*re
		fmul st,st(4)	;st*st(4) = 2*re*im
		fadd Y		; + y
		ffree st(4)
		fst st(4)
		;	re =  re_sq - im_sq + x;
		fld st(2)	;st <- st(2) re_sq
		fsub st,st(2)	;st - st(1) - im_sq
		fadd X		; + x
		ffree st(4)
		fst st(4)
		;	re_sq := re*re;
		fmul st,st
		ffree st(3)
		fstp st(3)
		;	im_sq := im*im;
		fmul st,st
		ffree st(1)
		fstp st(1)
		jmp short next_iter
exit_while:
		ffree st
		ffree st(1)
		ffree st(2)
		ffree st(3)
		ffree st(4)
		ffree st(5)

		mov ax,dx
		mov sp,bp
		pop bp
		ret
_calc_one_point	endp
#


; This is the new Pentium-optimized version for Mandelbrot set.
;
;int "C" calc_one_point(double x,double y,int max_it,
;			double &im_sq,double &re_sq);
;
;
x		EQU qword ptr[bp+6]
y		EQU qword ptr[bp+14]
max_it		EQU word ptr[bp+22]
im_sq		EQU dword ptr[bp+24]
re_sq		EQU dword ptr[bp+28]

_calc_one_point	proc far
		push bp
		mov bp,sp

		xor dx,dx	; DX - IterNo
		mov cx,max_it

		fld1
		fadd st,st
		fst st(1)	; st = 2
		fmul st,st	; st = 4
		fldz
		fldz
		fldz
		fldz
next_iter:
		fld st(1)	; re
		fmul st,st(6)	; 2*re
		fld st(3)	; re_sq
		fsub st,st(5)	; im_sq
		fxch st(1)
		fmul st,st(2)	; 2*re*im
		fxch st(1)
		fadd x		; re_sq-im_sq+x
		fxch st(1)
		fadd y		; 2.0*re*im + y
		fxch st(1)
		fst st(3)	; new re
		fxch st(1)
		fst st(2)	; new im
		fmul st,st	; im*im
		fxch st(1)
		fmul st,st	; re*re
		fxch st(1)
		fst st(5)	; im_sq
		fxch st(1)
		fst st(4)	; re_sq
		faddp st(1)	; im_sq+re_sq
		inc dx
		fcomp st(5)	; < 4
		fstsw ax
		sahf
		jnb short exit
		cmp dx,cx
		jg short exit
		jmp short next_iter
exit:
		fxch st(2)
		push ds
		lds bx,im_sq
		fstp qword ptr [bx]	; re_sq
		fxch st(2)
		lds bx,re_sq
		fstp qword ptr [bx]	; im_sq
		pop ds
		ffree st
		ffree st(1)
		ffree st(2)
		ffree st(3)

		mov ax,dx
		mov sp,bp
		pop bp
		ret
_calc_one_point	endp

; This is the Pentium-optimized version for Julia set.
;
;extern "C" int calc_one_point_jul(double x,double y,int max_it,
;			double &im_sq,double &re_sq,double c_image,
;			double c_real);
;

x		EQU qword ptr[bp+6]
y		EQU qword ptr[bp+14]
max_it		EQU word ptr[bp+22]
im_sq		EQU dword ptr[bp+24]
re_sq		EQU dword ptr[bp+28]
c_image		EQU qword ptr[bp+32]
c_real		EQU qword ptr[bp+40]

_calc_one_point_jul proc far
		push bp
		mov bp,sp

		xor dx,dx	; DX - IterNo
		mov cx,max_it

		fld1
		fadd st,st
		fst st(1)	; st = const 2
		fmul st,st	; st = const 4
		fld y		; st(3) = im_sq = im*im
		fld x		; st(2) = re_sq = re*re
		fld x		; st(1) = re
		fmul st(1),st
		fld y		; st(0) = im
		fmul st(3),st
next_iter1:
		fld st(1)	; re
		fmul st,st(6)	; 2*re
		fld st(3)	; re_sq
		fsub st,st(5)	; im_sq
		fxch st(1)
		fmul st,st(2)	; 2*re*im
		fxch st(1)
		fadd c_real	; re_sq-im_sq+x
		fxch st(1)
		fadd c_image	; 2.0*re*im + y
		fxch st(1)
		fst st(3)	; new re
		fxch st(1)
		fst st(2)	; new im
		fmul st,st	; im*im
		fxch st(1)
		fmul st,st	; re*re
		fxch st(1)
		fst st(5)	; im_sq
		fxch st(1)
		fst st(4)	; re_sq
		faddp st(1)	; im_sq+re_sq
		inc dx
		fcomp st(5)	; < 4
		fstsw ax
		sahf
		jnb short exit1
		cmp dx,cx
		jg short exit1
		jmp short next_iter1
exit1:
		fxch st(2)
		push ds
		lds bx,im_sq
		fstp qword ptr [bx]	; re_sq
		fxch st(2)
		lds bx,re_sq
		fstp qword ptr [bx]	; im_sq
		pop ds
		ffree st
		ffree st(1)
		ffree st(2)
		ffree st(3)

		mov ax,dx
		mov sp,bp
		pop bp
		ret
_calc_one_point_jul	endp
end