/************************************************************************/
/* This header defines classes for Mandelbrot-like set exploring	*/
/*									*/
/*    E.Podvoysky from ^Z for WROX press book, 1994			*/
/************************************************************************/

#ifndef MANDSET_H
#define MANDSET_H
#include "graph.h"
#include "common.h"
#include "grfile.h"

typedef struct {
	int iterations;
	float mod;
} SET_POINT;
typedef struct { pointtype a,b;}  RECT;


enum COLOR_SCHEME  {DETAILS, LINEAR, LOG, NONE};
enum SET_TYPE  {MANDELBROT, JULIA, DRAGONM, DRAGONJ};

//basic class for all classes, and mandelbrot set itself
class mandelbrot_set {
protected:
	BYTE *showrow;
	SET_POINT *row;

	int order_col1, order_col2, order_scheme; // use colors from _col1 to _col2
	int chaos_col1, chaos_col2, chaos_scheme; // to show respective part of set
	BYTE *order_color_map; // size max_it
	BYTE chaos_color_map[256];
public:
	BOOL int_calc, calculated; // use integer calculations or float
	image_keeper *keeper;
	double x_min,x_max,y_min,y_max,dx;
	int  screen_x,screen_y, screen_left, screen_top;
	int  max_it;


	mandelbrot_set(int max_it_, BOOL int_calc_,
		int screen_x_, int screen_y_,
		float x_min_, float x_max_, float y_min_, float y_max_);
	virtual ~mandelbrot_set();
	virtual SET_TYPE my_type();
	virtual void calc_one_row(double y);
	virtual void calc_one_row_int(double y);

	void set_keeper(image_keeper *keeper_);
	void set_coloring(int chaos_col1_,int chaos_col2_,int chaos_scheme_,
			  int order_col1_,int order_col2_,int order_scheme_);
	void show_row(int iy);
	void calculate_set();
	void set_window(RECT win);
	void set_ranges(float x_min_, float x_max_, float y_min_, float y_max_);
	virtual void set_c(float c_re, float c_im) {;}
	virtual void give_c(float &c_re, float &c_im) {;}
		// formal here, written in Julia set
};

class julia_set: public mandelbrot_set {
public:
      float c_real,c_image;
      long c_real_long,c_image_long;

      julia_set(int max_it_, BOOL int_calc_,
		int screen_x_, int screen_y_,
		float c_re, float c_im,
		float x_min_, float x_max_, float y_min_, float y_max_);
      virtual void calc_one_row(double y);
      virtual void calc_one_row_int(double y);
      virtual SET_TYPE my_type();
      virtual void set_c(float c_re, float c_im);
      virtual void give_c(float &c_re, float &c_im) {c_re = c_real; c_im = c_image;}
};


class dragon_m_set: public mandelbrot_set {
public:
      dragon_m_set(int max_it_, BOOL int_calc_,
		int screen_x_, int screen_y_,
		float x_min_, float x_max_, float y_min_, float y_max_);
      virtual void calc_one_row(double y);
      virtual void calc_one_row_int(double y);
      virtual SET_TYPE my_type();
};

class dragon_j_set: public julia_set {
public:
      dragon_j_set(int max_it_, BOOL int_calc_,
		int screen_x_, int screen_y_,
		float c_re, float c_im,
		float x_min_, float x_max_, float y_min_, float y_max_);
      virtual void calc_one_row(double y);
      virtual void calc_one_row_int(double y);
      virtual SET_TYPE my_type();
};

void init_palette(colortype *pal, int colors, colortype *colordef);

typedef struct {
	SET_TYPE set_type; // 0 for Mandelbrot set, 1 - Julia, 2 - dragon
	double x_min,x_max,y_min,y_max;
	int  screen_x,screen_y;
	float  c_re,c_im;
	char signature[4]; // must be Mand
	} mset_descr;


#endif

