/* This is a demo program for Expasm Chapter 'Pentium'
This program is an image filter, implementing 
convolution, median, and rank filtering methods.
The low level loops are optimized for Pentium.

Written by Y. Kiselev in Autumn 1994
*/

#include <mem.h>
#include <dos.h>
#include <dir.h>
#include <conio.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "common.h"
#include "graph.h"
#include "formats.h"
#include "formconv.h"

extern "C" void box_row(int Xsize, void* SourcePtr, void *DestPtr,
				int fxsize, int fysize, void *FilterPtr,
				float Koeff);

extern "C" void med_row(int Xsize,void* SourcePtr, void *DestPtr,
				int fxsize, int fysize, int filtsize,
				WORD add_ofs);

extern "C" void rank_row(int Xsize,void* SourcePtr, void *DestPtr, int fxsize,
		  int fysize, int filtsize, WORD hist_array_ofs,
		  float div_filtr_sq);

graph_file_reader *grf_r2;
graph_file_saver *grf_s;

char drive[MAXDRIVE];
char dir[MAXDIR];
char file[MAXFILE];
char ext[MAXEXT];

VGApalette palette;
int fxsize=3,fysize=3,width,height,i,half_filty,half_filtx,j,hi_part,lo_part;
WORD add_ofs,y;
float Koef,div_filtr_sq;
BYTE *source_ptr,*source_ptr_,*s_p,*row_ptr;
BYTE **array_ptr,filt_type;
float *filt_ptr;
/*float filt[] = {1,1,1,1,1,1,1,1,1,
		1,1,1,1,1,1,1,1,1,
		1,1,1,1,1,1,1,1,1,
		1,1,1,1,1,1,1,1,1,
		1,1,1,1,1,1,1,1,1,
		1,1,1,1,1,1,1,1,1,
		1,1,1,1,1,1,1,1,1,
		1,1,1,1,1,1,1,1,1,
		1,1,1,1,1,1,1,1,1,};*/

void main(int argc, char *argv[]) {
if (argc < 3) {
	printf("usage: \n");
	printf("          fastfilt source_file target_file [filt_name] [x] [y]\n\n");

	printf("filt_name: box,rank,med (by default box).\n");
	printf("x is filter halfwidth (by default 3). \n");
	printf("y is filter halfheight (if none then halfheight = halfwidth).\n");
	exit(0);
}

grf_r2 = open_image_file(argv[1]);
if (grf_r2==NULL) {printf("%s\n",format_file_error_report()); exit(0);}
if (grf_r2->source_type != GRAY256IMG && grf_r2->source_type != COLOR256IMG)
	{printf("256 colors only.\n"); exit(0);}

width = grf_r2->width;
height = grf_r2->height;

if (argv[4] != NULL && argv[3] != NULL) fxsize = atoi(argv[4])*2+1;
	else {fxsize = 3; argv[5] = NULL;}
if (argv[5] != NULL) fysize = atoi(argv[5])*2+1; else fysize = fxsize;

if (fxsize <= 1 || fxsize > 80 || fysize <= 1 || fysize > 80) {
	printf("Bad filter (halfsize = [1..40]).");
	exit(0);
}

fnsplit(argv[2],drive,dir,file,ext);
strlwr(&ext[1]);
if (memcmp(&ext[1],"tga",3))  {
	printf("Destination file format MUST be TARGA! (TGA extention).\n");
	exit(0);
}
grf_s = new targa_file_saver(argv[2],
		GRAY256IMG, NO_EXTENTION,
		grf_r2->width, grf_r2->height, NULL,
		TRUE, grf_r2->top_to_bottom, grf_r2->left_to_right);
graphinit();
switchmode(SVGA640x480x256);
for(i=0;i<256;i++) {
	palette[i].r = grf_r2->palette[i].r/4;
	palette[i].g = grf_r2->palette[i].g/4;
	palette[i].b = grf_r2->palette[i].b/4;
}
setDACblock(0,256,palette);

if (!memcmp(argv[3],"rank",3)) {
	filt_type = 0;
	y = 256*2;
	div_filtr_sq = 1.0/(fxsize*fysize);
}
else	if (!memcmp(argv[3],"med",3)) { y = fysize*fxsize; filt_type = 1;}
	else {
		y = 0;
		filt_ptr = new float [fysize*fxsize];
		for(i=0;i<fysize;i++)
			for(j=0;j<fxsize;j++) {
				filt_ptr[(i*fxsize)+j] = 1.0;
				Koef++;
			}
		if (Koef == 0) Koef = 1.0; else Koef = 1/Koef;
		filt_type = 2;
	}

add_ofs = fysize*(width+fxsize-1);
source_ptr = new BYTE [y+add_ofs];

row_ptr = new BYTE [width];
array_ptr = new BYTE *[fysize+1];

if (source_ptr == NULL || row_ptr == NULL || array_ptr == NULL) {
	closegraph();
	printf("Not enougth memory.");
	exit(0);
}

half_filtx = (fxsize-1)/2;
half_filty = (fysize-1)/2;

y = 0;
grf_r2->get_next_line();	// get first line
putrow((rowptr)row_ptr,0,y,width);

for(i=0;i<=half_filty;i++) {
	s_p = source_ptr+i*(width+fxsize-1);
	array_ptr[i] = s_p;
	memmove(s_p+half_filtx,grf_r2->source_line,width);
	memset(s_p,grf_r2->source_line[0],half_filtx);
	memset(s_p+width+half_filtx,grf_r2->source_line[width-1],half_filtx);
}

for(;i<fysize;i++,y++) {
	s_p = source_ptr+i*(width+fxsize-1);
	array_ptr[i] = s_p;
	grf_r2->get_next_line();
	memmove(row_ptr,grf_r2->source_line,width);
	putrow((rowptr)row_ptr,0,y,width);
	memmove(s_p+half_filtx,grf_r2->source_line,width);
	memset(s_p,grf_r2->source_line[0],half_filtx);
	memset(s_p+width+half_filtx,grf_r2->source_line[width-1],half_filtx);
}

// main filter
for(i=0;i<height-half_filty;i++,y++) {
	memmove(row_ptr,grf_r2->source_line,width);
	putrow((rowptr)row_ptr,0,y,width);	// source image

	if (filt_type == 0)
		rank_row(width,array_ptr,row_ptr,fxsize,fysize,
			fxsize*fysize,add_ofs,div_filtr_sq);
	else	if (filt_type == 1)
			med_row(width,array_ptr,row_ptr,fxsize,fysize,
				fxsize*fysize,add_ofs);
		else box_row(width,array_ptr,row_ptr,fxsize,fysize,filt_ptr,Koef);

	putrow((rowptr)row_ptr,320,i,width);	// result image
	grf_s->put_stand_line(row_ptr);

	grf_r2->get_next_line();
	s_p = array_ptr[0];
	memmove(s_p+half_filtx,grf_r2->source_line,width);
	memset(s_p,grf_r2->source_line[0],half_filtx);
	memset(s_p+width+half_filtx,grf_r2->source_line[width-1],half_filtx);
	array_ptr[fysize] = array_ptr[0];
 	memmove(array_ptr,&array_ptr[1],sizeof(void*)*fysize);
	if (kbhit()) { getch(); goto leave_prog;}
}

// rest filter
for(j=0;j<(fysize-1)/2;j++,i++) {
	if (filt_type == 0)
		rank_row(width,array_ptr,row_ptr,fxsize,fysize,
			fxsize*fysize,add_ofs,div_filtr_sq);
	else	if (filt_type == 1)
			med_row(width,array_ptr,row_ptr,fxsize,fysize,
				fxsize*fysize,add_ofs);
		else 	box_row(width,array_ptr,row_ptr,fxsize,fysize,filt_ptr,Koef);

	putrow((rowptr)row_ptr,320,i,width);	// show result image
	grf_s->put_stand_line(row_ptr);

	s_p = array_ptr[0];
	memmove(s_p+half_filtx,grf_r2->source_line,width);
	memset(s_p,grf_r2->source_line[0],half_filtx);
	memset(s_p+width+half_filtx,grf_r2->source_line[width-1],half_filtx);
	array_ptr[fysize] = array_ptr[0];
	memmove(array_ptr,&array_ptr[1],sizeof(void*)*fysize);
}
leave_prog:

delete source_ptr;
delete array_ptr;
delete row_ptr;
delete grf_r2;
delete grf_s;
if (filt_type != 2) delete filt_ptr;
closegraph();
}
