/* This is a demo program for Expasm Chapter 'Pentium'.
This a performance test: it multiplies 1 000 000 matrices 4x4
and reports the elapsed time in seconds. The results of tests on
the Pentium and i486 were listed in the chapter.
Written by Y. Kiselev in Autumn 1994
*/

#include <math.h>
#include <dos.h>
#include <mem.h>
#include <conio.h>
#include <stdio.h>
#include <stdlib.h>

#include "common.h"

typedef float matrix[4][4];
// Get the current value of the Pentium's time-stamp counter
extern "C" void get_tsc(void* p);
// Get the difference (in ms) between the current time-stamp counter
// and the parameter p. Clock rate is passed as freq.
// Be aware to use get_tsc and get_diff only on the Pentium in real mode
// (switch off extended memory managers like EMM386 or QEMM386)
extern "C" int get_diff(void* p,long freq);

// 486- optimized version
extern "C" void mult_matrix(matrix &matr1,matrix &matr2,matrix &res_matr);
// Pentium version, slightly optimized
extern "C" void mult_matrixp(matrix &matr1,matrix &matr2,matrix &res_matr);
// Pentium version, well optimized
extern "C" void mult_matrixp1(matrix &matr1,matrix &matr2,matrix &res_matr);

// On pre-Pentium models I used tick counter to check time
long get_tick(void);
long get_diff(long f);

void mult_matr(matrix &matr1,matrix &matr2,matrix &res_matr);
matrix *a,*b,*c,*c1;

void main(void) {
int jj,i,j;
long l;
void* pp;
long first;

randomize();
pp = new BYTE [8];
a = new matrix [1];
b = new matrix [1];
c = new matrix [1];
c1 = new matrix [1];

for(i=0; i<4;i++)
 for(j=0;j<4;j++) (*a)[i][j] = random(500)/(random(50)-0.1)-10;

for(i=0;i<4;i++)
 for(j=0;j<4;j++) (*b)[i][j] = random(10)/(random(50)+0.128);

/* Pentium test
get_tsc(pp);
for(l=0;l<1000000;l++) mult_matrix(*a,*b,*c1);
printf("Old asm: %d ms\n",get_diff(pp,60000));

get_tsc(pp);
for(l=0;l<1000000;l++) mult_matrixp(*a,*b,*c1);
printf("New asm: %d ms\n",get_diff(pp,60000));

get_tsc(pp);
for(l=0;l<1000000;l++) mult_matrixp1(*a,*b,*c1);
printf("RISC asm: %d ms\n",get_diff(pp,60000)); */

/* 486DX test */
first = get_tick();
for(l=0;l<1000000;l++) mult_matrix(*a,*b,*c1);
printf("Old asm: %f sec \n",get_diff(first)/18.0);

first = get_tick();
for(l=0;l<1000000;l++) mult_matrixp(*a,*b,*c1);
printf("New asm: %f sec \n",get_diff(first)/18.0);

first = get_tick();
for(l=0;l<1000000;l++) mult_matrixp1(*a,*b,*c1);
printf("RISC asm: %f sec \n",get_diff(first)/18.0);
}

void mult_matr(matrix &matr1,matrix &matr2,matrix &res_matr) {
int i,j,k;
long double t;
for(i=0;i<4;i++)
 for(j=0;j<4;j++) {
	t=0;
	for(k=0;k<4;k++) t+=matr1[i][k] * matr2[k][j];
	res_matr[i][j]=t;
  }
}

long get_tick(void) {
asm {
	xor ax,ax
	mov es,ax
	mov ax,es:[0x46c]
	mov dx,es:[0x46e]
}
}

long get_diff(long f) {
asm {
	xor ax,ax
	mov es,ax
	mov ax,es:[0x46c]
	mov dx,es:[0x46e]
	sub ax,word ptr f
	sbb dx,word ptr f+2
}
}