name key_line_edit
;====================================================================
;
; Copyright(c) 1994 Quantasm Corporation. All rights reserved.
;
; Quantasm Corporation Window Library
;
;====================================================================

.model small

EXTRN cur_set_pos        : proc
EXTRN cur_set_block      : proc
EXTRN cur_set_line       : proc
EXTRN key_read_char      : proc
EXTRN scr_display_string : proc

.code

public key_line_edit
key_line_edit proc
;--------------------------------------
;
; inputs: DS:DI buffer for output
;         AX    screen loc
;         BX    screen window width
;         CL    max len
;         CH    input len
;         DX    routine to handle illegal char
;               (0 = none)
;         SI    attributes
;               0 = default attr
;               1 = attr in high half
;               2 = use existing attr
;
; outputs: DI:DI data
;          CX    actual len
;          CF set if ESCAPE
;--------------------------------------

le_buf_beg   equ <word ptr [bp-2]>
le_lm        equ <word ptr [bp-4]>
le_width     equ <word ptr [bp-6]>
le_buf_end   equ <word ptr [bp-8]>
le_len       equ <word ptr [bp-10]>
le_max_len   equ <word ptr [bp-12]>
le_attr      equ <word ptr [bp-14]>
le_loc       equ <word ptr [bp-16]>
le_illegal   equ <word ptr [bp-18]>    ; near ptr to handle illegal keys
le_ins_flag  equ <byte ptr [bp-19]>    ; 0 = overwrite;    1 = insert
le_scroll    equ <byte ptr [bp-20]>    ; 0 = no scrolling; 1 = allow scrolling
local_stack  equ 22


           push   bp
           mov    bp, sp
           sub    sp, local_stack

           push   bx
           push   dx
           push   di
           push   si
           push   es

           mov    le_loc, ax
           mov    le_illegal, dx
           mov    ax, si
           cmp    al, 2
           jle    le0
           xor    ax, ax
le0:
           mov    le_attr, ax

           mov    le_ins_flag, 0
           call   cur_set_line
le2:
           mov    ax, ds
           mov    es, ax

; initialize buffer ptrs

           xor    ax, ax
           mov    al, ch
           mov    le_len, ax           ; init len of string
           mov    ax, di               ; init end
           xor    ch, ch
           add    ax, cx
           mov    le_buf_end, ax       ; ptr to buffer end
           mov    le_buf_beg, di       ; ptr to buffer beginning
           mov    le_lm, di            ; ptr in buffer of on-screen left margin
           mov    le_width, bx         ; width on screen
           dec    cx
           mov    le_max_len, cx       ; max string len

           xor    al, al
           cmp    bl, cl
           jge    le3
           inc    al
le3:
           mov    le_scroll, al

le_disp:
           call   display_buffer

le_disp_cur:
           call   display_cursor

le_rd_key:
           call   key_read_char
           cmp    al, 13             ; CR
           je     le_return
           cmp    ax, 011bh          ; escape key
           je     le_escape
           mov    bx, offset le_key_tbl

le_rd_key1:
           cmp    ax, CS:[bx]
           je     le_rd_key3
           cmp    word ptr CS:[bx], -1
           je     le_rd_key2
           add    bx, 4
           jmp    le_rd_key1

le_rd_key2:
           cmp    al, ' '
           jl     le_rd_key4
           cmp    al, 127
           jae    le_rd_key4
le_rd_key2a:
           call   le_char
           jmp    le_disp

le_rd_key3:
           call   CS:[bx+2]
           jmp    le_disp

; routine for illegal char
le_rd_key4:
           mov    bx, le_illegal
           test   bx, bx
           jz     le_rd_key
           call   bx
           jc     le_return
           cmp    ah, 0
           je     le_rd_key2a
           jmp    le_disp

le_return:
           mov    cx, le_len
           mov    di, le_buf_beg
           add    di, cx
           mov    byte ptr [di], 0     ; insure null at end of string
           clc
le_ret:
           pop    es
           pop    si
           pop    di
           pop    dx
           pop    bx
           add    sp, local_stack
           pop    bp
           ret

le_escape:
           stc
           jmp    le_ret

le_key_tbl label word

dw 4700h, le_home
dw 4f00h, le_end
dw 4d00h, le_right
dw 4b00h, le_left
dw 5300h, le_delete
dw 0e08h, le_backspace
dw 5200h, le_insert
dw -1

key_line_edit endp

le_home proc near
;--------------------------------------
; Home key
;--------------------------------------

           mov    di, le_buf_beg
           mov    le_lm, di
           ret

le_home endp

le_end proc near
;--------------------------------------
; END key
;--------------------------------------


           mov    di, le_buf_beg
           add    di, le_len

           mov    ax, le_lm
           add    ax, le_width
           cmp    di, ax
           jl     le_end_ret
           call   adjust_rm
le_end_ret:
           ret

le_end endp

le_right proc near
;--------------------------------------
; Right arrow
;--------------------------------------

           mov    ax, di
           sub    ax, le_buf_beg
           cmp    ax, le_len
           jge    le_right_ret
           inc    di
           call   adjust_rm
le_right_ret:
           ret

le_right endp

le_left proc near
;--------------------------------------
; Left arrow
;--------------------------------------

           cmp    di, le_buf_beg
           je     le_left_ret
           dec    di
           call   adjust_lm
le_left_ret:
           ret

le_left endp

le_delete proc near
;--------------------------------------
; Delete key
;--------------------------------------

           mov    ax, di
           sub    ax, le_buf_beg
           cmp    ax, le_len
           jge    le_del_ret
           call   shift_left
           dec    le_len
le_del_ret:
           ret

le_delete endp

le_backspace proc near
;--------------------------------------
; Backspace key
;--------------------------------------

           cmp    di, le_buf_beg
           je     le_back_ret
           dec    di
           call   shift_left
           dec    le_len
           call   adjust_lm
le_back_ret:
           ret

le_backspace endp

le_insert proc near
;--------------------------------------
; Insert/Overwrite key
;--------------------------------------

           xor     le_ins_flag, 1
           jz      le_ins1
           call    cur_set_block
           jmp     le_ins2
le_ins1:
           call    cur_set_line
le_ins2:
           ret

le_insert endp


le_char proc near
;--------------------------------------
; Character to insert/overwrite
;--------------------------------------

           push   bx

           cmp    le_ins_flag, 1
           jne    le_char1
           mov    bx, le_len
           cmp    bx, le_max_len
           jl     le_char0

; if last char is a space, then delete it and add this one

           add    bx, le_buf_beg
           cmp    byte ptr [bx-1], ' '
           jne    le_char4
           mov    byte ptr [bx-1], 0
           dec    le_len
le_char0:
           call   shift_right          ; make space for char
           jmp    le_char2

; overwrite a char
le_char1:
           mov    bx, di
           sub    bx, le_buf_beg
           cmp    bx, le_len
           jl     le_char3

; overwrite at end of buffer

           mov    bx, le_len
           cmp    bx, le_max_len
           jg     le_char4
le_char2:
           inc    le_len
le_char3:
           mov    [di], al
           inc    di
           call   adjust_rm
le_char_ret:
           pop    bx
           ret

; not enough room
le_char4:

           jmp    le_char_ret

le_char endp

adjust_lm proc near

           cmp    le_scroll, 1
           jne    adj_ret
           cmp    di, le_lm
           ja     adj_ret
           mov    bx, di
           cmp    bx, le_buf_beg
           je     adj_ret
           dec    bx
           mov    le_lm, bx
adj_ret:
           ret

adjust_lm endp

adjust_rm proc near

           push   bx
           cmp    le_scroll, 1
           jne    ar_ret
           mov    bx, di
           sub    bx, le_lm
           sub    bx, le_width
           or     bx, bx
           jl     ar_ret
           inc    bx
           add    le_lm, bx
ar_ret:
           pop    bx
           ret

adjust_rm endp


shift_right proc near
;--------------------------------------
;  Example ... A B D E F 0  to  A B D D E F 0
; inputs: DS:DI    ^
;--------------------------------------

           push   ax
           push   cx
           push   di
           push   si
           push   es
           pushf

           mov    ax, ds
           mov    es, ax
           std

           mov    cx, le_len
           jcxz   sr_ret
           add    cx, le_buf_beg
           cmp    di, cx               ; no shift if at or beyond end
           jae    sr_ret               ; of current text
           mov    ax, di
           mov    di, cx
           sub    cx, ax
           jcxz   sr_ret
           mov    si, di
           dec    si

           rep movsb
sr_ret:
           popf
           pop    es
           pop    si
           pop    di
           pop    cx
           pop    ax
           ret

shift_right endp

shift_left proc near
;--------------------------------------
; Example ... A B C D E F 0  to  A B D E F 0
; inputs: DS:DI   ^
;--------------------------------------

           push   cx
           push   di
           push   si
           push   es
           mov    ax, ds
           mov    es, ax

           mov    si, di
           inc    si
           mov    cx, le_len
           jcxz   sl_ret
           add    cx, le_buf_beg
           sub    cx, di
           or     cx, cx
           jle    sl_ret

           rep movsb
sl_ret:
           pop    es
           pop    si
           pop    di
           pop    cx
           ret

shift_left endp

display_buffer proc near
;--------------------------------------
; display visible portion of buffer
;--------------------------------------

           push   ax
           push   bx
           push   cx
           push   dx
           push   di

           mov    bx, le_loc
           mov    si, le_lm
           mov    cx, le_width
           mov    ax, le_attr
           call   scr_display_string

           pop    di
           pop    dx
           pop    cx
           pop    bx
           pop    ax

           ret

display_buffer endp

display_cursor proc near
;--------------------------------------
; Move cursor to current position
;--------------------------------------

           push   ax
           push   bx

           mov    ax, le_loc
           mov    bx, di
           sub    bx, le_lm
           add    al, bl
           call   cur_set_pos

           pop    bx
           pop    ax

           ret

display_cursor endp
end
