;====================================================================
;
; Copyright(c) 1994 Quantasm Corporation. All rights reserved.
;
; Quantasm Corporation Window Library
;
;====================================================================

.model small

.code

public num_long_div
num_long_div proc
;--------------------------------------
; Unsigned division
;   64-bit operand by a 32-bit operand
;
; inputs: DX:AX  operand 1 (32-bit)
;         DS:SI  ptr to operand 2 (64-bit)
;                (operand 2 changed to 32-bit quotient)
;
; outputs: BX:CX longword  quotient
;          DX:AX longword remainder
;          CF    set if divide by zero
;--------------------------------------

        push    bp

        mov     bp, 64                  ; number of bits to divide
        cmp     ax, 0                   ; check for div by zero
        jne     nld_0
        cmp     dx, 0
        je      nld_err

nld_0:
        mov     cx, ax                  ; low word of divisor
        mov     bx, dx                  ; high word

        xor     ax, ax
        xor     dx, dx

nld_1:
        clc
        rcl     word ptr [si], 1        ; shift result and source
        rcl     word ptr [si+2], 1
        rcl     word ptr [si+4], 1
        rcl     word ptr [si+6], 1
        rcl     ax, 1
        rcl     dx, 1

        cmp     dx, bx                  ; check high words
        jb      nld_3                   ; if below, go to next bit
        ja      nld_2                   ; if above, do subtract
        cmp     ax, cx                  ; check low words when
        jb      nld_3                   ;   high words equal

nld_2:
        sub     ax, cx                  ; sub when divisor is smaller
        sbb     dx, bx
        inc     word ptr [si]           ; inc result

nld_3:
        dec     bp                      ; next bit
        jnz     nld_1

        mov     cx, [si]                ; low word of result
        mov     bx, [si+2]              ; high word of result
        clc                             ; successful

nld_ret:
        pop     bp

        ret

nld_err:
        stc                             ; error if divide by zero
        jmp     nld_ret

num_long_div endp
end
