;====================================================================
;
; Copyright(c) 1994 Quantasm Corporation. All rights reserved.
;
; Quantasm Corporation Window Library
;
;====================================================================
; Screen Mouse Routines
;====================================================================

.model small

.data

public scr_mouse_found
scr_mouse_found   db 0

.code

public scr_mouse_init
scr_mouse_init proc
;--------------------------------------
;  Initialize the mouse, if it exists
;
; inputs: none
;
; outputs: CF set if no mouse
;          BX number of buttons
;          scr_mouse_found set if a mouse is found
;--------------------------------------

           push   es
           xor    ax, ax               ; load segment for IVT
           mov    es, ax
           cmp    word ptr ES:[33h*4], ax       ; check for a vector of 0
           je     mi_err
           cmp    word ptr ES:[33h*4+2], ax
           je     mi_err
           int    33h                  ; if vector is not 0, then attempt
           or     ax, ax               ; to init the mouse driver
           jz     mi_err

           mov    scr_mouse_found, 1
           clc
mi_ret:
           pop    es
           ret
mi_err:
           stc                         ; no mouse driver
           jmp    mi_ret

scr_mouse_init endp

public scr_mouse_hide
scr_mouse_hide proc
;--------------------------------------
;  Removes the mouse pointer
;
; inputs: none
;
; outputs: none
;
; preserves flags
;--------------------------------------

           pushf
           push   ax
           call   scr_mouse_exists
           jnc    mh_1
           mov    ax, 2
           int    33h
mh_1:
           pop    ax
           popf
           ret

scr_mouse_hide endp

public scr_mouse_unhide
scr_mouse_unhide proc
;--------------------------------------
;  Restores the mouse pointer
;
; inputs: none
;
; outputs: none
;
; preserves flags
;--------------------------------------

           pushf
           push   ax
           call   scr_mouse_exists
           jnc    mu_1
           mov    ax, 1
           int    33h
mu_1:
           pop    ax
           popf
           ret

scr_mouse_unhide endp

public scr_mouse_exists
scr_mouse_exists proc
;--------------------------------------
;  Checks to see if a mouse exists, based on the initialization routine
;
; inputs: none
;
; outputs: CF set if a mouse exists
;--------------------------------------

           push   ax
           mov    al, scr_mouse_found
           shr    al, 1
           pop    ax
           ret

scr_mouse_exists endp


public scr_mouse_read_pos_status
scr_mouse_read_pos_status proc
;--------------------------------------
;  Reads the current mouse position and button status
;
; inputs: none
;
; outputs: BX  button status bits
;              0 left button
;              1 right button
;              2 center button
;          CX  x coordinate
;          DX  y coordinate
;--------------------------------------

           mov    ax, 3
           int    33h
           ret

scr_mouse_read_pos_status endp
end
