; Stand-alone example for protected mode 16-bit and/or 32-bit code
;
; protected mode code with DPMI

.model small
.stack
.386

.data

; message strings

msg1     db 'Mixed 16-bit and 32-bit code example.',13,10,0
msg32    db 'In protected mode, 32-bit segment.',13,10,0
err_msg1 db 'DPMI host not detected.',13,10,0
err_msg2 db 'Could not allocate real-mode memory.',13,10,0
err_msg3 db 'Could not switch to protected mode.',13,10,0


code32_selector     dw 0        ; filled with selector for CODE32 segment
main32_address      df main32

; data returned from INT 2Fh function 1687h

DPMI_mem            dw 0
DPMI_entry          dd 0

cs_descriptor label dword
cs_limit     dw 0                      ; descriptor limit
cs_low_addr  db 0,0,0                  ; low order 3 bytes of address
cs_access    dw 0                      ; descriptor access bytes
cs_high_addr db 0                      ; high order byte of address

.code

main proc

        mov     ax, @data
        mov     ds, ax

; (1) re-size memory allocated

        mov     bx, ss                  ; start of stack segment
        mov     ax, es                  ; start of PSP
        sub     bx, ax                  ; para for code & data
        mov     ax, sp                  ; get stack size
        shr     ax, 4                   ; divide by 16 to get para
        add     bx, ax
        inc     bx                      ; BX = total para needed
        mov     ah, 4ah                 ; modify memory allocation
        int     21h

; (2) check to see if DPMI is available

        call    DPMI_check              ; check for a DPMI server
        jc      error1                  ; exit if none

; (3) Print intro message

        lea     si, msg1                ; intro message
        call    print_string

; (4) Allocate memory for DPMI server

        mov     ax, 0                   ; use seg=0 if no memory required
        mov     bx, DPMI_mem            ; get para count required for DPMI
        test    bx, bx
        jz      main_2                  ; skip down if none required
        mov     ah, 48h                 ; allocate memory
        int     21h
        jc      error2                  ; check for allocate error

; (5) switch to protected mode
main_2:
        mov     es, ax
        mov     ax, 1                   ; indicate 32-bit application
        call    dword ptr DPMI_entry    ; switch to protected mode
        jc      error3                  ; if CF set, error in switching

; <<< 16-bit Protected Mode code here >>>

;====================================
; setup to switch to 32-bit mode
;====================================

; (6) allocate a descriptor for use32 segment

        mov     ax, 0                   ; allocate LDT descriptor
        mov     cx, 1                   ; 1 descriptor
        int     31h                     ; call DPMI host
        mov     code32_selector,ax      ; selector to use32 segment
        mov     word ptr main32_address[2], ax ; store selector

; (7) get copy of LDT entry

        mov     bx, cs                  ; get current CS descriptor
        mov     di, offset cs_descriptor; point to result buffer
        movzx   edi, di                 ; (convert to 32 bits)
        push    ds                      ; ES:EDI => buffer
        pop     es                      ;
        mov     ax, 000bh               ; get descriptor request
        int     31h                     ; call DPMI host

; (8) set segment base address

        mov     ax, seg code32          ; linear address of code32
        movzx   eax, ax                 ;
        shl     eax, 4                  ; segment value times 16
        mov     dx, ax                  ; move to cx:dx
        mov     ecx, eax                ;
        shr     ecx, 16                 ;
        mov     bx, code32_selector     ; identify selector
        mov     ax, 7                   ; set segment base address
        int     31h                     ; call DPMI host

; (9) set segment limit

        mov     cx, 0000h               ; set segment limit to 64K
        mov     dx, 0ffffh              ;
        mov     bx, code32_selector     ; identify selector
        mov     ax, 8h                  ; set segment limit request
        int     31h                     ; call DPMI host

; (10) switch to 32-bit mode by changing the D bit

        mov     cx, cs_access           ; start with current cs access
        or      ch, 40h                 ; change to 32 bit default
        mov     bx, code32_selector     ; identify selector
        mov     ax, 0009h               ; set access rights request
        int     31h                     ; call DPMI host

; (11) Execute a 16:16 far call to 32-bit code

        call    dword ptr main32_address

; When we return we are still in 16-bit PM.

main_exit:
        mov    ax, 4c00h                ; DOS exit
        int    21h

; error messages

error1:
        lea    si, err_msg1             ; DPMI host not found
        jmp    err0
error2:
        lea    si, err_msg2             ; could not allocate memory
        jmp    err0
error3:
        lea    si, err_msg3             ; could not switch to PM
err0:
        call   print_string
        jmp    main_exit

main endp


DPMI_check proc near
;--------------------------------------
; Check for a DPMI server
;
; outputs: CF set if no DPMI server installed
;--------------------------------------

        push    ax
        push    bx
        push    cx
        push    dx
        push    di
        push    si
        push    es

        mov     ax, 1687h               ; request DPMI host address
        int     2fh                     ; via multiplex interrupt (2Fh)
        test    ax, ax                  ; AX = 0 if success
        jnz     not_found               ; exit with error if not found

        and     bl, 00000001b           ; check bit 0 to see if 32-bit
        setne   al                      ; code supported
        mov     DPMI_mem, si
        mov     word ptr DPMI_entry[0], di
        mov     word ptr DPMI_entry[2], es
        clc                             ; success
dc_ret:
        pop     es
        pop     si
        pop     di
        pop     dx
        pop     cx
        pop     bx
        pop     ax
        ret

not_found:
        stc
        jmp     dc_ret

DPMI_check  endp

public print_string
print_string proc near
;--------------------------------------
; Print ASCIIZ string
;
; inputs: DS:SI ptr to string
;--------------------------------------

        push    ax
        push    bx
        push    cx
        push    dx

        call    str_len
        mov     cx, ax
        mov     dx, si
        mov     bx, 1                   ; stdout
        mov     ah, 40h                 ; DOS write handle
        int     21h

        pop     dx
        pop     cx
        pop     bx
        pop     ax
        ret

print_string endp

public str_len
str_len proc near
;--------------------------------------
; Get length of string
;
; inputs: DS:SI ptr to string
;
; outputs: AX length
;--------------------------------------

        push    cx
        push    di
        push    es

        push    ds
        pop     es
        mov     di, si

        mov     cx, -1                  ; max str len
        xor     al, al                  ; char to look for
        cld

        repne   scasb                   ; repeat while ES:[DI] <> 0
        mov     ax, -2
        sub     ax, cx                  ; len

        pop     es
        pop     di
        pop     cx
        ret

str_len endp

code32  segment public para use32 'code'
        assume  cs:code32

public main32
main32 proc far
;--------------------------------------
; 32-bit code main entry point
;
; DS = @data from 16-bit code
;--------------------------------------

        pusha

; (12) display message that we're in 32-bit mode

        mov     esi, offset msg32
        call    print_string32

; (13) call any 32-bit code from here

        call    your_32_bit_code

; restore registers

        popa

; (14) return to the 16-bit code segment we came from

        db      66h                     ; operand size prefix (see note below)
        ret

;---------------------------------------------------------------------
; Note: If we used a RET without the operand size prefix the assembler
;       would generate a normal RETF for 32-bit code (16:32). We need
;       this return to match the CALL that got us here (16:16).
;---------------------------------------------------------------------

main32 endp

your_32_bit_code proc

        ret

your_32_bit_code endp

public print_string32
print_string32 proc near
;--------------------------------------
; 32-bit Print ASCIIZ string
;
; inputs: DS:ESI ptr to string
;--------------------------------------

        push    eax
        push    ebx
        push    ecx
        push    edx

        call    str_len32
        mov     ecx, eax
        mov     edx, esi
        mov     ebx, 1                  ; stdout
        mov     ah, 40h                 ; DOS write handle
        int     21h

        pop     edx
        pop     ecx
        pop     ebx
        pop     eax
        ret

print_string32 endp

public str_len32
str_len32 proc near
;--------------------------------------
; 32-bit Get length of string
;
; inputs: DS:ESI ptr to string
;
; outputs: EAX length
;--------------------------------------

        push    ecx
        push    edi
        push    es

        push    ds
        pop     es
        mov     edi, esi

        mov     ecx, -1                 ; max str len
        xor     al, al                  ; char to look for
        cld

        repne   scasb                   ; repeat while ES:[DI] <> 0
        mov     eax, -2
        sub     eax, ecx                ; len

        pop     es
        pop     edi
        pop     ecx
        ret

str_len32 endp

code32  ends

end main
