;************************************************
;* 	Prog5_1.asm				*
;* 	Driver Dump				* 
;* 	Author Yu.I.Petrenko 			*
;************************************************
;This program is developed for revealing the device
;drivers allocation in RAM, displaying  their
;addresses, names (for character drivers), strategy
;and interrupt routines near addresses and attribute
;setting of each driver. The manual execution of this
;program is very useful for understanding driver
;location in RAM and device driver header structure.
;----------------------------------------------------
.286
cd	segment
assume 	CS:cd,DS:cd,ES:cd,ss:nothing
ORG	100h
Start:		JMP	Begin
;----------------------------------------------------
; The data beneath is intended for headings typing and
;for hexadecimal number conversion into ASCII-strings
;----------------------------------------------------
CLRSCR		DB	25 DUP (0Dh,0Ah),'$'	
TITTLE		DB	20 DUP(' ')
		DB	'RESIDENT DRIVERS CHAIN:',0Dh,0AH,'$'
NL_STRG		DB	0Dh,0AH,'$'
ATTR		DB	48 DUP(' '),'Attributes',0Dh,0AH,'$'
TTL1		DB	'Memory',7 DUP(' '),'Driver',4 DUP(' ')
		DB	'Strategy',4 DUP(' '),'Interrupt'
TTL2		DB	4 DUP(' ')
		DB	'FEDCBA9876543210',0Dh,0AH,0Dh,0AH,'$'
ALPHA		DW	?
DR_SEG		DW	?
OFST		DW	?
HEX		DB	'0123456789ABCDEF'
DRVRS_COUNT	DB	16
;----------------------------------------------------
; The program section between BEGIN and A labels
;clears screen, display headings,points out
;the address of driver header of the very first DOS
;driver, e.g. the NUL driver and stores the address
;----------------------------------------------------
Begin:						;Program begins
 		MOV 	AX,CS			;Usual trick
 		MOV 	DS,AX			;for the COM-programs
		MOV 	DX,offset CLRSCR
		CALL 	TypeString		;Clear screen
		MOV 	DX,offset TITTLE	;Display the title
		CALL 	TypeString
		MOV 	DX,offset Attr		;Display the table
		CALL 	TypeString		;header
		MOV 	DX,offset TTL1		;Continue
		CALL 	TypeString		;it
		MOV 	AH,52h			;The Int 21h Fn 52h
						;returns the list of lists
		INT 	21h			;address in ES:BX
		ADD 	BX,22h			;First driver NUL 
						;address in ES:BX
		MOV 	DR_SEG,ES		;Store the address
		MOV 	OFST,BX		;of current driver
;----------------------------------------------------
A:
;----------------------------------------------------
; This program portion up to GoOut label develops
;sequential processing of the driver headers addresses
;in the RAM driver chain. After the address getting
;it calls the DRV_INFO subroutine for header pursing
;----------------------------------------------------
		MOV	BX,OFST		;Restore driver offset
		MOV 	ES,DR_SEG		;and segment
		DEC 	DRVRS_COUNT	;Drivers number decreased
		CMP 	OFST,0FFFFh		;Last driver?
		JZ 	GoOut			;Yes, end
		CALL 	DriverInfo		;No, display
		MOV 	BX,OFST		;Offset 
		MOV 	ES,DR_SEG		;and segment 
						;of current driver
		MOV 	AX,ES:[BX]		
		MOV 	OFST,AX			;Offset and
		MOV 	AX, WORD PTR ES:[BX+2]	;segment of the next
		MOV 	DR_SEG,AX		
		CMP 	DRVRS_COUNT,1		;Last driver?
		JZ 	GoOut			;Yes,end the program
		MOV 	DX,offset NL_STRG	;No yet
		CALL 	TypeString		;Type the empty string
		JMP 	A			;Next circle
GoOut:
		INT 	20h			;Int 20h Program Terminate
						;E nd
;----------------------------------------------------
; PROCEDURES :
;----------------------------------------------------
TypeSpace:					;Type defined
		MOV 	DL,20h			;in CL register
		CALL 	TypeSymbol		;number of spaces
		DEC	CL
		JNZ	TypeSpace
		RET
;----------------------------------------------------
TypeSymbol:					;This subroutine
		MOV	AH,2			;types the character
		INT	21h			;from the DL register
		RET
;----------------------------------------------------
TypeString:					;This subroutine types
		MOV 	AH,9			;the ASCI-string
		INT	21h			;DS:DX - string address
		RET				;String must be ended with '$'
;----------------------------------------------------
 TypeHexWord:
;----------------------------------------------------
;This subroutine displays the hex code from the BX
;register in form of the ASCII-string. It stores the
;code in ALPHA variable, then reads it 4 times to
;BX register and by left and right shifts extracts
;four digits in BL. After that the subroutine
;reads appropriate character from the HEX table
;and calls TypeSymbol subroutine.
;----------------------------------------------------
		MOV	CL,0
		MOV	ALPHA,BX		;Temporary storage
;----------------------------------------------------
HexDigit:
		MOV	BX,ALPHA		
		SHL	BX,CL			
		SHR	BX,12
		XOR	BH,BH
		MOV	DL,HEX[BX]
		CALL	TypeSymbol		;Types symbol
		ADD	CL,4			;Symbol count
		CMP	CL,16			;Last symbol?
		JZ	Rt
		JMP	HexDigit		;No,continue
Rt:
		RET				;Yes, return
;----------------------------------------------------
AtrDisplay:
;----------------------------------------------------
;This subroutine types the contents of the Attributes
;word as string of characters. If the appropriate attribute
;is set, the corresponding hex digit is displayed, if
;clear - nothing.
;----------------------------------------------------
		MOV 	BX,0
ShiftBit:	SHL 	ALPHA,1			;The leftmost bit 
						;is in the carry flag
		JNC	NotShow		;CF=0, type space
		MOV	DL,TTL2[BX+4]		;CF=1,type symbol
		CALL	TypeSymbol
		JMP	Nxt
NotShow:					;Display nothing
		MOV	DL,20h			;20h is a space code
		CALL	TypeSymbol		;Types space
Nxt:
		INC	BX			;Symbol counting
		CMP	BX,16			;up to 16 bits
		JZ	Rtn			;Last?
		JMP	ShiftBit			;No,continue
Rtn:		RET				;Yes,return
;----------------------------------------------------
DriverInfo:
;----------------------------------------------------
;This subroutine makes a single driver header data
;processing
;-----------------------------------------------------
		MOV	BX,DR_SEG		;Driver address segment in BX
		CALL	TypeHexWord		;Address segment displaying
		MOV	DL,':'			;type the ':'
		CALL	TypeSymbol		
		MOV	BX,OFST		;Driver address offset in BX
		CALL	TypeHexWord		;Type it
		MOV	CL,4			;Type
		CALL	TypeSpace		;4 spaces
;----------------------------------------------------
;Preparing for the driver name displaying
;----------------------------------------------------
		MOV	ES,DR_SEG		;Driver address segment
						;in ES
		MOV	BX,OFST		;Driver address offset
						;in BX
		ADD	BX,0Ah			;Driver name offset
		MOV	DI,0
;----------------------------------------------------
_Name:
;----------------------------------------------------
;The code up to RDS label types driver name from the
;device header. It gets symbols one by one, filters
;the alphabet characters and then types the character
;or the dot, if fails to find it.
;----------------------------------------------------
		PUSH	ES			;Type
		POP	DS			;the driver
		MOV	DL,BYTE PTR[DI+BX]	;name
		CMP	DL,' '			
		JNB	NotAbove		;Dots for the
		MOV	DL,'.'			;lower symbol
		JMP	TpSmbl
NotAbove:
		CMP	DL,'Z'
		JNA	TpSmbl			;Dots for the
		MOV	DL,'.'			;upper one
;----------------------------------------------------
TpSmbl:
		CALL	TypeSymbol		;Type it
		INC	DI
		CMP	DI,8			;Count them
		JNZ	_NAME			;up to 8 bytes
;----------------------------------------------------
TwoSpaces:
;----------------------------------------------------
;Types two spaces between the name and the
;strategy near address
;----------------------------------------------------
		MOV	AX,CS			;DS restore
		MOV	DS,AX			;again
		MOV	CL,2			;type two 
		CALL	TypeSpace		;spaces
;----------------------------------------------------
Adresses:
;----------------------------------------------------
;The strategy and interrupt routine addresses to be typed
;----------------------------------------------------
		MOV	ES,DR_SEG		;Driver address segment
		MOV	BX,OFST		;Driver address offset
		ADD	BX,6			;Strategy routine
		MOV	BX,ES:[BX]		;near pointer
		CALL	TypeHexWord	
		MOV	CL,8			;Again spaces
		CALL	TypeSpace
		MOV	BX,OFST		;Driver address offset
		ADD	BX,8			;interrupt routine
		MOV	BX,ES:[BX]		;pointer
		CALL	TypeHexWord
		MOV	CL,9			;Type
		CALL	TypeSpace		;9 spaces
;----------------------------------------------------
;Well done, but it is nessesary to parse and
;to type Attributes word contents
;----------------------------------------------------
		MOV	BX,OFST		;Driver address offset
		ADD	BX,4			;Attributes word's pointer
		MOV	AX,ES:[BX]		;Attributes word in AX
		MOV	ALPHA,AX
		CALL	AtrDisplay		;Display it!
		RET				;End for one driver, go to next
		INT	20h			;Int 20h Program Terminate
CD ENDS					;End of the program
END START
;====================================================
;****************************************************
