;************************************************
;*	Program 5_2				*
;*      First Character Driver CON1			*
;*	Author: Yu.I.Petrenko			*
;************************************************
.286
code 	segment byte
assume 		CS:code,DS:code,ES:code,SS:nothing
;====================================================
; CON1 - character device driver template
;====================================================
con1header:
;----------------------------------------------------
;The data up to Stratgy label is the device header
;====================================================
Link		DW	-1,-1
Attributes	DW	8000h			;bit 15 is set
Strategy	DW	Stratgy
Interrupt	DW	Intrpt
DrvName		DB	'CON1    '
;====================================================
Stratgy:
;The Strategy Routine only storage the Request Header Address
;====================================================
		MOV 	WORD PTR CS:[ReqHdr],BX
		MOV 	WORD PTR CS:[ReqHdr+2],ES
		RETF				;Now Request Header
						;is stored
;----------------------------------------------------
;The current Request Header Address
;----------------------------------------------------
ReqHdr		DW	0 			;BX = Offset
		DW	0			;ES = Segment
;====================================================
Intrpt:						;Interrupt Routine
;====================================================
		PUSHF				;Register savings
		PUSHA
		PUSH 	DS
		PUSH 	ES
;----------------------------------------------------
;The next code portion determines command code, check it
;and pass it to "execute"
;----------------------------------------------------
		CLD
		LES 	DI,DWORD PTR CS:[ReqHdr];Request header in DS:BX
		MOV 	BL,BYTE PTR ES:[DI+2]	;Command code now is  in AX
		XOR	BH,BH
		SHL	BL,1
		CMP	BL,22			;Is command valid?
		JA	ErrCom			;No, erroneous
		JMP	WORD PTR CS:[BX+ComtaBLe]
						;Yes,valid -> execute it
;----------------------------------------------------
Comtable:
;----------------------------------------------------
;There are labels corresponding to the requested
;driver command. Valid ones are not executed and
;for the erroneous the Error code is generated
;----------------------------------------------------
		DW	Init			;00h - Init
		DW	Outfrom			;01h - Illegal request
		DW	Outfrom			;02h - Illegal request
		DW	ErrCom			;03h - Erroneous Command
		DW	Read			;04h - Read
		DW	NndRead		;05h - Nondestructive Read
		DW	InpStatus		;06h - Input Status
		DW	InpFlush		;07h - Input Flush
		DW	Write			;08h - Write
		DW	WriteVer		;09h - Write with Verify
		DW	Outfrom			;0AH - Illegal request
		DW	Outfrom			;0BH - Illegal request
;----------------------------------------------------
;The program section up to Message label serves for the
;command not executing or Error code generating. The
;only executed command is Init.
;----------------------------------------------------
Read:
;----------------------------------------------------
		JMP 	Done
;----------------------------------------------------
NndRead:
;----------------------------------------------------
		JMP 	Done
;----------------------------------------------------
InpStatus:
;----------------------------------------------------
		JMP 	Done
;----------------------------------------------------
InpFlush:
;----------------------------------------------------
		JMP 	Done
;----------------------------------------------------
Write:
;----------------------------------------------------
		JMP 	Done
;----------------------------------------------------
WriteVer:
;----------------------------------------------------
		JMP 	Done
;====================================================
;Various Exits. They end driver request work.
;====================================================
ErrCom:					;Erroneous command
		MOV 	AX,8103h		;AL=03h - Unknown
						;command Error value
						;AH=81h - error+done bits
		JMP	Outfrom
;----------------------------------------------------
Err_unnwn:					;Unnown error
		MOV	AX,8100h		;AH=81h - error+done bits
		JMP 	Outfrom
Drvbusy:
		MOV 	AX,0300h		;Driver busy
		JMP	Outfrom
Done:
		XOR 	AX,AX
Outfrom:
		OR 	AX,0100h		;Bit done is set
		LES 	BX,DWORD PTR CS:[ReqHdr]
						;Request header in ES:BX
		MOV 	WORD PTR ES:[BX+3],AX	;Status Word = AX
		POP	ES
		POP	DS
		POPA
		POPF
		RETF				;Far return,
						;exit from driver
;====================================================
Init:
;----------------------------------------------------
;This code portion up to Message label does the updating
;of the Request Header
;----------------------------------------------------
		PUSH	CS
		POP	DS
		MOV	DX,offset Message
		MOV	AH,9			;Int 21h Fn 09h
		INT	21h			;Display String
		LES 	DI,DWORD PTR CS:[ReqHdr]
						;Request header in ES:BX
		MOV 	WORD PTR ES:[DI+0Eh],offset Endd
						;End of driver's memory
		MOV 	ES:[DI+10h],CS
		JMP	Done
;----------------------------------------------------
Message		DB	0dh,0Ah
		DB	' Now CON1 is resident',0Dh,0Ah,'$'
;----------------------------------------------------
Endd:
		NOP				;These NOP's marks
		NOP				;the end of the
CODE ENDS					;program (in debugger)
END
;====================================================
;****************************************************

