;************************************************
;*	Program 5_3				*
;*	Lazy Practical Character Driver CON2 	*
;*	Author: Yu.I.Petrenko			*
;************************************************
.286
code 	segment byte
assume 		CS:code,DS:code,ES:code,SS:nothing
;====================================================
; CON2 - lazy,but practical character device driver
;====================================================
Con2header:					;CON2 device driver header
;----------------------------------------------------
;The data up to Stratgy label is the device header
;====================================================
Link		DW	-1,-1
Attributes	DW	8013h			;Bits 15,4,1 and 0 are set
Strategy	DW	Stratgy
Interrupt	DW	Intrpt
DrvName		DB	'CON',5 dup (' ')
;====================================================
Stratgy:
;The Strategy Routine only stores the Request Header Address
;====================================================
		MOV 	WORD PTR CS:[ReqHdr],BX
		MOV 	WORD PTR CS:[ReqHdr+2],ES
		RETF				;Now the address is stored
;----------------------------------------------------
ReqHdr		DD	0			;Current Request 
						;Header Address
;====================================================
Intrpt:						;Interrupt Routine
;====================================================
		PUSHF				;Register savings
		PUSHA
		PUSH	DS
		PUSH 	ES
;----------------------------------------------------
;The next code portion determines command code, check it
;and pass it to execute by twin driver
;----------------------------------------------------
		CLD
		LES 	DI,CS:[ReqHdr]		;Request header in ES:DI
		MOV 	BL,BYTE PTR ES:[DI+2]	;Command code now is in AX
		XOR 	BH,BH
		SHL 	BL,1
		CMP 	BL,22			;Is command valid?
		JA 	ErrCom			;No, erroneous
		JMP	WORD PTR CS:[BX+Comtable]
						;Yes,valid -> execute it
;----------------------------------------------------
ComTable:
;----------------------------------------------------
;There are labels corresponding to the requested driver
;command. Valid ones are passed to twin driver later executed.
;For the erroneous, Error code is genereted.
;----------------------------------------------------
		DW 	Init			;00h - Init
		DW 	OutFrom		;01h - Illegal request
						;Media Check
		DW 	OutFrom		;02h - Illegal request Build BPB
		DW 	ErrCom			;03h - Erroneous Command
						;IOCTL Read
		DW 	Read			;04h - Read
		DW 	NndRead		;05h - Nondestructive Read
		DW 	InpStatus		;06h - Input Status
		DW 	InpFlush		;07h - Input Flush
		DW 	Write			;08h - Write
		DW 	WriteVer		;09h - Write with Verify
		DW 	OutptStatus		;0AH - Output Status
		DW 	OutptFlush		;0BH - Output Flush
;----------------------------------------------------
MakeIt:
;----------------------------------------------------
;This program section (up to Read label) passes the
;Request Header pointer to twin driver. At first, it calls
;the Strategy program, futher - the Interrupt program
;of the hooked driver.
;----------------------------------------------------
		CALL 	Beg
		PUSH 	CS
		POP 	DS
		LES 	BX,ReqHdr
		CALL 	DWORD PTR [ConStrategy]
		PUSH 	CS
		POP 	DS
		LES 	BX,ReqHdr
		CALL	DWORD PTR [ConInterrupt]
		JMP 	Done
;----------------------------------------------------
;Further are command passing
;----------------------------------------------------
Read:
;----------------------------------------------------
		JMP 	MakeIt
;----------------------------------------------------
NndRead:
;----------------------------------------------------
		JMP 	MakeIt
;----------------------------------------------------
InpStatus:
;----------------------------------------------------
		JMP 	MakeIt
;----------------------------------------------------
InpFlush:
;----------------------------------------------------
		JMP 	MakeIt
;----------------------------------------------------
Write:
;----------------------------------------------------
		JMP 	MakeIt
;----------------------------------------------------
WriteVer:
;----------------------------------------------------
		JMP 	MakeIt
;----------------------------------------------------
OutptStatus:
;----------------------------------------------------
		JMP 	MakeIt
;----------------------------------------------------
OutptFlush:
;----------------------------------------------------
		JMP 	MakeIt
;====================================================
; Various Exits
;----------------------------------------------------
;In the section up to the Message label there are
;program completion for commands not to be executed
;====================================================
ErrCom:					;Erroneous command
		MOV 	AX,8103h		;AL=03h - Unknown
						;command Error value
						;AH=81h - error+done bits
		JMP 	OutFrom
;----------------------------------------------------
ErrUnnwn:					;Unnown error
		MOV 	AX,8100h		;AH=81h - error+done bits
		JMP	OutFrom
DrvBusy:
		MOV	AX,0300h		;Driver busy + done bits
		JMP 	OutFrom
Done:
		MOV 	AX,0
OutFrom:
		MOV	AH,1			;Bit done is set
		LES	BX,DWORD PTR CS:[ReqHdr]
						;Request header in DS:BX
		MOV 	WORD PTR ES:[BX+3],AX
						;Status word now = AX
		POP ES
		POP DS
		POPA
		POPF
		RETF				;Exit from  driver
;====================================================
Message		DB	0dh,0Ah
		DB	' Init CON2: Now CON2 is resident',0dh,0AH,'$'
;----------------------------------------------------
DoMessage:;Display Message string
;----------------------------------------------------
		PUSH	CS
		POP	DS
		MOV	DX,offset Message
		MOV	AH,9			;Int 21h FN 09h
		INT	21h			;Display String
		RET
;----------------------------------------------------
Init:
;----------------------------------------------------
;This program section executes the Init command.
;It updates the Request Header
;----------------------------------------------------
		CALL	DoMessage
		LES	DI,DWORD PTR CS:[ReqHdr]
						;Request header in DS:BX
		MOV	WORD PTR ES:[DI+0Eh],offset Endd
						;End of driver's memory
		MOV	ES:[DI+10h],CS
		JMP	Done
;----------------------------------------------------
;This data portion is intended for  current Driver
;addresses. Virtually it contents the CON Driver
;addresses.
;----------------------------------------------------
ConAdr		DD	0
ConStrategy	DD	0
ConInterrupt	DD	0
;----------------------------------------------------
Beg:
;----------------------------------------------------
;At first this subroutine checks if it is the first
;search. If not, it returns. Else this subroutine
;gets the current driver Header address and extracts
;f the driver Name. Later this name  is compared
;with  its own name (CON), byte after byte. If name differs,
;it reads addresses of the next driver, if it exists, and
;saves them. Then the driver search is repeated.
;When the driver with identical name is found,
;the search is finished.
;----------------------------------------------------
		PUSH	CS
		POP	DS
		TEST	WORD PTR [ConAdr+2],0ffffh
;First search? If first, the word contents 0. If not,
;it contents nonzero address segment.
		JZ 	D			;Yes
		RET				;No, return
D:						;Driver search
		LES 	BX,DWORD PTR [Link]
		CMP 	BX,0FFFFh		;Is the driver the last
						;in the System?
		JNZ 	DrSearch
		RET				;Yes, return
DrSearch:					;No, try to meet the twin
		MOV 	CX,8
		MOV 	DI,offset DrvName
		MOV 	SI,0AH
L:						;Comparing of next driver
						;name with the own
		MOV 	AL,ES:[BX+SI]
		CMP 	BYTE PTR [DI],AL	;Identical?
		JZ 	Go			;One character more
						;Coincides yet
		JMP 	Next			;No, that one differs
Go:						;Hope exists yet
		INC 	DI			;Mask shift
		INC 	SI			;The same
		LOOP 	L			;Continue
						;This driver is identical
		MOV 	WORD PTR [ConAdr],BX
		MOV 	WORD PTR [ConAdr+2],ES 	;This is a flag!
		MOV 	AX,WORD PTR ES:[BX+6] 	;Storing
		MOV 	WORD PTR[ConStrategy],AX;the twin's
		MOV 	WORD PTR[ConStrategy+2],ES
						;personal
		MOV 	AX,WORD PTR ES:[BX+8] 	;data
		MOV 	WORD PTR[ConInterrupt+2],ES
						;for the future
		MOV 	WORD PTR[ConInterrupt],AX
						;using
		RET
Next:						;Try if is the driver last
		CMP 	ES:[BX],0ffffh
		JZ 	Last			;Yes, last driver, return
		LES 	BX,DWORD PTR ES:[BX] 	;Not last, try to find the twin
						;Now ES:BX points to next
						;Driver Header
		JMP 	DrSearch
Last:
		RET
;----------------------------------------------------
Endd:
;----------------------------------------------------
		NOP				;These NOP's marks
		NOP				;the end of the
CODE ENDS					;program (in debugger)
END						;
;====================================================
;****************************************************
