
;************************************************
;*						*
;*	Parallel Printer Driver PRN1		*
;*      Author: Yu.I.Petrenko                   *
;*						*
;************************************************
.286
 code segment BYTE
 assume CS:code,DS:code,ES:code
;====================================================
; PRN1 - simple parallel printer driver
;====================================================
 PRN1header: 		; PRN printer driver header
;====================================================
Link		DW	-1,-1
Attributes	DW 	0A040h
			;Bits F,D and 6 are set
Strategy 	DW 	Stratgy
Interrupt 	DW 	Intrpt
DrvName 	DB 	'PRN      '
;====================================================
Stratgy:
;The Strategy Routine only storage the Request Header Address
;====================================================
		MOV 	WORD PTR CS:[ReqHdr],BX
		MOV 	WORD PTR CS:[ReqHdr+2],ES
		RETF				;Now the address is stored
;----------------------------------------------------
ReqHdr		DD	0			;Current Request
						;Header Address
;====================================================
Intrpt:						;Interrupt Routine
;====================================================
		PUSHF				;Register savings
		PUSHA
		PUSH	DS
		PUSH	ES
;----------------------------------------------------
;The next code portion determines command code, check it
;and pass it to execute by twin driver
;----------------------------------------------------
		CLD
		LES 	DI,CS:[ReqHdr]
						;Request header in ES:DI
		MOV 	BL,BYTE PTR ES:[DI+2]	;Command code now is in BL
		XOR 	BH,BH
		SHL 	BL,1
		CMP 	BL,26h			;Is command valid?
		JA 	ErCom 			;No, erroneous
		JMP 	WORD PTR CS:[BX+Comtable]
						;Yes,valid -> execute it
ErCom:		JMP 	ErrCom
;----------------------------------------------------
 ComTaBLe:
;----------------------------------------------------
;There are labels corresponding to the requested driver
;command. Valid ones are passed to twin driver later executed.
;For the erroneous Error code is genereted.
;----------------------------------------------------
		DW 	Init			; 00h - Init
		DW 	Outfrom			
			; 01h - Illegal request Media Check
		DW 	Outfrom
			; 02h - Illegal request Build BPB
		DW 	Outfrom
			; 03h - Illegal request IOCTL Read
		DW 	Outfrom
			; 04h - Illegal request Read
		DW 	Outfrom
			; 05h - Illegal request Nondestructive Read
		DW 	Outfrom
			; 06h - Illegal request Input Status
		DW 	Outfrom
			; 07h - Illegal request Input Flush
		DW	Write
			; 08h - Write - Send 1 character to Printer
		DW 	Outfrom
			; 09h - Illegal request Write with Verify
		DW	OutptStatus
			; 0AH - Output Status ;Printer status
		DW 	OutptFlush
			; 0BH - Output Flush  ;Printer Initialization
		DW	ErrCom,ErrCom,ErrCom,ErrCom,ErrCom,ErrCom,ErrCom
			;7 Erroneous Commands
		DW 	GenericIOCTL
;----------------------------------------------------
 MakeIt:
;----------------------------------------------------
;This program section (up to Read label) passes the
;Request Header pointer to twin driver. At first, it calls
;the Strategy program, futher - the Interrupt program
;of the hooked driver.
;----------------------------------------------------
		CALL	Beg
		PUSH 	CS
		POP 	DS
		LES 	BX,ReqHdr
		CALL 	DWORD PTR[PrnStrategy]
		PUSH 	CS
		POP 	DS
		LES 	BX,ReqHdr
		CALL 	DWORD PTR[PrnInterrupt]
		JMP 	Done
;----------------------------------------------------
 Read:
;----------------------------------------------------
		JMP 	MakeIt
;----------------------------------------------------
 NndRead:
;----------------------------------------------------
		JMP 	MakeIt
;----------------------------------------------------
 InpStatus:
;----------------------------------------------------
		JMP 	MakeIt
;----------------------------------------------------
 InpFlush:
;----------------------------------------------------
		JMP 	MakeIt
;----------------------------------------------------
Write:		JMP	W1
Frst  		DB  	0
W1:		PUSH 	CS
		POP DS
		TEST 	Frst,0ffh
		JNZ 	Second
		MOV 	Frst,1
;====================================================
Inst:            				; Setting Int 17h vector
		MOV 	AH,9			; DOS Int 09h
		MOV 	DX, Offset Hooked	; 
		INT 	21h         		; Display String
		MOV 	AX,3517h		; DOS Int 35h
		INT 	21h 			; Get Int 17h vector
		MOV 	Word PTR [Int17h_Old+0],BX
		MOV 	Word PTR [Int17h_Old+2],ES
		MOV 	DI, Offset Hooked	;Check the string contents
		CMP 	Word PTR ES:[DI],'nI' 	;<<<already set?>>>
		JE 	Cb
		MOV 	DX, Offset Int_17h_Handler
		MOV 	AX,2517h		; DOS Int 25h
		INT 	21h 			; Set Int 17h vector
;----------------------------------------------------
Cb:
		CALL 	Beg
			;<<<Call before the command executing>>>
		JMP 	Second
;------------------------------------------------------
Hooked 		DB 	'Int 17h is hooked',0Ah,0dh,'$'
Int17h_Old 	DD 	?
CtrCodeA	DB 	1Bh,'K'			
			;Control sequence (of bit image mode)
		DW 	000Bh			
			; 11h columns = bit image code length
ImageA 		DB 	1Eh,28h,48h,88h,88h,88h,88h,48h,28h,1Eh,00h
			; Bit image code
BytesA		EQU 	$ - CtrCodeA
;====================================================
Second:		LES 	BX,CS:ReqHdr
		MOV 	CX,ES:[BX+12h]		;Symbol quantity
		TEST 	CX,CX
		JNZ 	LS
		JMP 	Done			; JmpOut
 LS:
		LES 	DI,DWORD PTR ES:[BX+0Eh]
			;ES:DI - buffer
ToPrn:
		MOV 	AL,ES:[DI]
		MOV 	AH,0
		MOV 	DX,0			; DOS Int 17h, function 00h
		INT 	17h			; Print Character
		JMP 	IncDI
;===============================================================
Int_17h_Handler:
		STI
		CLD
		CMP 	AH,0
		JE 	I171
		JMP	To_17h			;Print_Quit
I171:
		CMP 	AL,41h   		; <<<'A' code>>>
		JNE 	To_17h   		; 'A' Image
		MOV 	BX, BytesA		
			;Number of bytes to write<<<>>>
		MOV 	DI, Offset CtrCodeA
			; DS:DI points to buffer with data
_411:
		XOR 	AH,AH
		MOV 	AL,CS:[DI]
		PUSHF
		CALL 	CS:[Int17h_Old]		; Call old Interrupt
		TEST 	AH,8
		JNZ 	_411
		INC 	DI
		DEC 	BX
		JNZ 	_411
		MOV 	AH,02h			
			; Get Printer Status in AH
		To_17h:
		PUSHF
		CALL 	CS:[Int17h_Old] 	; Call old Interrupt
		IRET
;---------------------------------------------------------------
IncDI:  
		INC DI
		LOOP ToPrn
		JMP Outfrom
;		JMP MakeIt
;----------------------------------------------------
 WriteVer:
;----------------------------------------------------
		JMP MakeIt
;----------------------------------------------------
 OutptStatus:
;----------------------------------------------------
		JMP MakeIt
;----------------------------------------------------
OutptFlush:
;----------------------------------------------------
		JMP MakeIt
;----------------------------------------------------
GenericIOCTL:
;----------------------------------------------------
		JMP MakeIt
;====================================================
; Various Exits
;----------------------------------------------------
;In program section to the Message label there are
;program completion for commands not to be executed
;====================================================
ErrCom:
			;Erroneous command
		MOV 	AX,8103h
			;AL=03h - Unknown command Error value
			;AH=81h - error+done bits
		JMP	Outfrom
;----------------------------------------------------
Err_unnwn:
			;Unnown error
		MOV	AX,8100h
			;AH=81h - error+done bits
		JMP 	Outfrom
DrvBusy:
		MOV 	AX,0300h
			;Driver busy + done bits
		JMP	Outfrom
Done:
		MOV 	AX,0
Outfrom:					;Out from driver
		MOV 	AH,1			;Bit done is set
		LES 	BX,DWORD PTR CS:[ReqHdr]
			;Request header in DS:BX
		MOV 	WORD PTR ES:[BX+3],AX
			;Status word now = AX
		POP ES
		POP DS
		POPA
		POPF
		RETF				; Exit from driver
;====================================================
Message		DB	0Dh,0Ah
DB		' Init Prn2: Now Prn2 is resident',0dh,0AH,'$'
;----------------------------------------------------
Domessage:					;Display Message string
		PUSH 	CS
		POP 	DS
		MOV 	DX, Offset Message
		MOV 	AH,9                        ;Int 21h Fn 09h
		INT 	21h                         ;Display String
		RET
;----------------------------------------------------
Init:
;----------------------------------------------------
;This program section executes the Init command.
;It updates the Request Header
;----------------------------------------------------
		CALL 	DoMessage
		LES 	DI,DWORD PTR CS:[ReqHdr]
			;Request header in DS:BX
		MOV 	WORD PTR ES:[DI+0Eh], Offset Endd
			;End of driver's memory
		MOV 	ES:[DI+10h],CS
		JMP 	Done
;----------------------------------------------------
;This data portion is intended for a current Driver
;addresses. Virtually it contents the PRN Driver
;addresses.
;----------------------------------------------------
PrnAdr 		DD 	0
PrnStrategy 	DD 	0
PrnInterrupt 	DD 	0
;----------------------------------------------------
Beg:
;----------------------------------------------------
;At first this subroutine checks if it is the first
;search. If not, it returns. Else this subroutine
;gets the current driver Header address and extract
;from it the driver Name. This name later is compared
;with own name (PRN), byte after byte. If name differs,
;it reads addresses of next driver, if it exists, and
;saves them. Then the driver search is repeated.
;When the driver with identical name is finded,
;the search is ended.
;----------------------------------------------------
			;Is the driver but only one in the System?
		PUSH 	CS
		POP 	DS
		TEST 	WORD PTR [PrnAdr+2],0ffffh
;First search? If first, the word contents 0. If not,
;it contents nonzero address segment.
		JZ 	D
			;Yes, only one driver exists
		RET
D:
			;No, some else ones exists
		LES 	BX,DWORD PTR [Link]
		CMP	BX,0FFFFh
			; Is the driver the last in the System?
		JNZ	Dr_search
		RET				;Yes, return
Dr_search:
			;No, try to meet the twin
		MOV	CX,8
		MOV	DI, Offset DrvName
		MOV	SI,0AH
L:
			;Comparing of next driver name with the own
		MOV 	AL,ES:[BX+SI]
		CMP 	BYTE PTR [DI],AL	;Identical?
		JZ 	Go
			;One character more couinsides yet
		JMP 	Next
			;No, that one differs
Go:
			;Hope exists yet
		INC 	DI			;Mask shift
		INC 	SI			;The same
		LOOP 	L			;Continue
			;This driver is identical
		MOV 	WORD PTR [PrnAdr],BX
		MOV 	WORD PTR [PrnAdr+2],ES 	;This is a flag!
		MOV 	AX,WORD PTR ES:[BX+6] 	;Storing
		MOV 	WORD PTR[PrnStrategy],AX;the twin's
		MOV 	WORD PTR[PrnStrategy+2],ES 
						;personal
		MOV 	AX,WORD PTR ES:[BX+8] 	;data
		MOV 	WORD PTR[PrnInterrupt+2],ES 
						;for the future
		MOV 	WORD PTR[PrnInterrupt],AX 
						;using
		RET
Next:						
			;Try if is the driver last
		CMP 	ES:[BX],0ffffh
		JZ 	Last
			;Yes, last driver, return
		LES 	BX,DWORD PTR ES:[BX]
			;Not last, try to find the twin
		JMP 	Dr_search
Last:
		RET
;----------------------------------------------------
Endd:
;----------------------------------------------------
		NOP	;This may be used as the end mark
		NOP	;of the program for debugging
CODE ENDS
END
;****************************************************

