
;****************************************
;*					*
;*	    Simple Tutorial             *
;*       Disk Reading Program           *
;*	Author: Yu.I.Petrenko		*
;*					*
;****************************************
.286
cd segment
assume cs:cd,ds:cd,es:cd,ss:cd
   ORG 100h
start:  JMP BEG
;program data
Disk_Numb       DB   0
Disk_Head_Nmb   DB   0
Track_Nmb       DB   3
Head_Nmb        DB   0
Sect_Numb       DB   1
Sect_Size_Code  DB   2
Sect_Last       DB   18
GPL             DB   2Ah  ;gap = 2Ah for exchange
Data_Nmb       DB   256 - 1
DCR             EQU  3F2h  ;control register address
DELAY_          DW  ?
Buffer     DB  512 DUP (73h) ;marker code
Buff_Size  DW  511
Status     DB  'Status '
;---------------------------------------------------------
; This program reads the sector 90 contents into the
; Buffer; insert the disk A before to test it!
;---------------------------------------------------------
BEG:          ;
                MOV AX,CS
                MOV DS,AX
 STI
 CALL Find_Track
;---------------------------------------------------------
;
;DMA Program
MOV AL,46h;
OUT 0Ch,AL ; Write = Clear the Trigger
OUT 0Bh,AL ; Byte to mode register:
          ; 01 0 0 01 10
          ;         cannel 2
          ;      write into memory
          ;     no autoinitialization
          ;    address increasing
          ;   single transfer mode
MOV AX, offset Buffer
MOV BX, DS
CALL Address_  ; Addres and page calculating
               ; AX - offset within page
               ; DL - page number

OUT 4,AL       ; base address writing
MOV AL,AH      ; for 2-nd cannel
OUT 4,AL

MOV AL,DL      ; base page writing
OUT 81h,AL     ; for 2-nd cannel

MOV AX,Buff_Size    ; write byte quantity
OUT 5,AL            ; to current word counter
MOV AL,AH
OUT 5,AL

MOV AL,2      ; 00000 0 10
              ;          canal 2
              ;       mask bit 0 = transfer permitted
              ;  reserved, must be zeros
OUT 0Ah,AL    ; reset of 2-nd cannel mask = DMA start
;---------------------------------------------------------
 CALL Disk_Read

 CALL Disk_Int

 CALL Disk_Status

 CALL Disk_Stop

 JMP EXIT_
;---------------------------------------------------------
 Disk_Read:
 MOV AH,66h   ;command code = 45h
 CALL To_Disk

 MOV AH, Disk_Head_Nmb; 00000 0 00
                      ;         disk number 0
                      ;        head number 0
 CALL To_Disk

 MOV AH, Track_Nmb
 CALL To_Disk

 MOV AH,Head_Nmb
 CALL To_Disk

 MOV AH,Sect_Numb
 CALL To_Disk

 MOV AH,Sect_Size_Code
 CALL To_Disk

 MOV AH,Sect_Last
 CALL To_Disk

 MOV AH,GPL
 CALL To_Disk

 MOV AH,Data_Nmb  ;
 CALL To_Disk

 RET
;---------------------------------------------------------
 Disk_Status:
 MOV CX,7
 LEA BX,Status
Stat:
 CALL From_Disk
 MOV [BX],AL
 INC BX
 LOOP Stat
 RET
;---------------------------------------------------------
 Disk_Stop:
 MOV DX,3F2h;3F2h  ;control register address
 MOV AL,0Ch ;0 0 0 0 1 1 00  = reset of controller command
            ;          disk A = 00
            ;         0 =  controller reset
            ;        permitted work with DMA
            ;       motor A OFF
 OUT DX,AL
 RET
;---------------------------------------------------------
; This subroutine calculates full buffer address from the
; input address BX:AX
Address_:
 MOV DL,BH
 SHR DL,4
 SHL BX,4
 ADD AX,BX
 JNC ret_
 INC DL
 ret_: RET
;---------------------------------------------------------
Find_Track:
 MOV DX,DCR ;3F2h  ;control register address
 MOV AX,28  ;0 0 0 1 1 1 00  = reset of controller command
            ;          disk A = 00
            ;         1 = not controller reset
            ;        1 = permitted work with DMA
            ;       motor A ON
 OUT DX,AL  ;

 MOV Delay_,10
 CALL Delay

 MOV AH,0Fh  ; "Set" command
 CALL To_Disk

 MOV AH, Disk_Head_Nmb; 00000 0 00
                      ;         disk number 0
                      ;        head number 0
 CALL To_Disk

 MOV AH,Track_Nmb ; Track number
 CALL To_Disk

 CALL Disk_Int    ;wait disk interrupt

 MOV Delay_,2     ;wait heads to set (>=25 msec)
 CALL Delay
RET
;---------------------------------------------------------
To_Disk:
 CALL Ready_?
 MOV DX,3F5h ; data register
 MOV AL,AH
 OUT DX,AL
 jmp $+2
 jmp $+2
 RET
;---------------------------------------------------------
From_Disk:
 CALL Ready_?
 MOV DX, 3F5h ; data register
 IN AL,DX
 RET
;---------------------------------------------------------
Disk_Int:  ;wate disk interrupt
 MOV AX,40h
 MOV ES,AX
 MOV BX,3Eh
WAIT_:
 MOV DL,ES:[BX]
 TEST DL,80h
 JZ WAIT_
 AND DL,7Fh
 MOV ES:[BX],DL
 RET
;---------------------------------------------------------
Ready_?:
 MOV DX,3F4h  ; MSR - main status register
MORE:
 IN AL,DX
 TEST AL,80h  ; if bit 7 = 1, then data transmit permitted
 JZ MORE
 RET
;---------------------------------------------------------
Delay:
   PUSH ES
   MOV AX,40h
   MOV ES,AX
   MOV AX,ES:[6Ch] ;counter value
   XOR CX,CX
D1: MOV BX,ES:[6Ch]
    CMP BX,AX
    JZ D1
    MOV AX,BX
    INC CX
    CMP CX,Delay_
    JNZ D1
    POP ES
    RET
;---------------------------------------------------------
 EXIT_:
            mov AH,4Ch
            int 21h
            ;.EXIT

cd ends
end start
