;
; MYDRIVER.ASM -- Sample Sound Blaster driver.
;
; Assemble and link, then:
;   EXE2BIN MYDRIVER,MYDRIVER.DRV
;
CODE    SEGMENT 'CODE'
        ASSUME CS:CODE

        ORG 0

        ; Function select
        cmp     al,0
        je      InitDSP
        cmp     al,1
        je      Version
        cmp     al,2
        je      Input
        cmp     al,3
        je      MixerReset
        cmp     al,4
        je      MixerVolume
        cmp     al,5
        je      SetSource
        cmp     al,6
        je      Function6
        retf    ; Unknown function


BasePort        dw      220h            ; Sound Blaster's Base I/O
                                        ; port

; Function 0. Set Base I/O address, passed in BX and initialize DSP.
; Must be called before any other function

InitDSP:
        mov     BasePort,bx
        mov     dx,BasePort
        add     dx,6            ; Reset port, Base + 6
        mov     al,1
        out     dx,al           ; Send 1 to reset port

        mov     cx,500          ; Wait a little...
WaitLoop:
        loop    WaitLoop

        xor     al,al
        out     dx,al           ; Send 0 to reset port

        xor     cx,cx           ; Try up to 10000h times
        add     dx,8            ; Read-Buffer Status port
                                ; Base + Eh
NoData:
        in      al,dx           ; Read Data-Ready status
        or      al,al           ; Is bit 7 set?
        jns     TryAgain        ; No, data not ready

        sub     dx,4            ; Read Data port, Base + Ah
        in      al,dx           ; Read DSP data
        cmp     al,0AAh         ; Is it AAh?
        je      Success         ; Yes, O.K.
TryAgain:
        loop    NoData          ; Try more...

        mov     ax,1            ; Sound Blaster not found...
        retf

Success:
        mov     ax,0            ; Sound Blaster initialized successfully!
        retf


; Function 1. Get DSP Version. Returns AH = major, AL = minor
; version numbers.

Version:
        mov     ah,0E1h         ; Command : Get DSP Version
        call    WriteDSP        ; Send the command
        call    ReadDSP         ; Read version's major byte
        mov     ah,al           ; Major into AH
        call    ReadDSP         ; Read version's minor byte
        retf

; Function 2. 8-bit direct mode digitized sound input.
; Returns sample in AL.

Input:
        mov     ah,20h          ; Command : direct input
        call    WriteDSP        ; Send the command
        call    ReadDSP         ; Read one sample
        retf


; Function 3. Reset mixer to its default state.
; Sound Blaster Pro only.

MixerReset:
        push    bx
        xor     bx,bx           ; Reset register, index 0.
        call    WriteMixer
        pop     bx
        retf

; Function 4. Set volume of the source specified in AH.
; Sound Blaster Pro only.

; AH =  04  - Voice
;       0Ah - Microphone
;       22h - Master
;       26h - FM synthesizer
;       28h - CD
;       2Eh - Line-in

; For microphone: DH = volume level (0 - 7).
; For stereo sources:
; DL - volume for left channel, DH - right.
; Available levels are 0 - 15.

MixerVolume:
        push    bx
        mov     bl,dl           ; Form high-order nibble
        shl     bl,1            ; of register value
        shl     bl,1
        shl     bl,1
        shl     bl,1

        and     dh,0Fh          ; Form low-order nibble
        or      bl,dh

        mov     bh,ah           ; Index of register for
                                        ; the specified source
        call    WriteMixer
        pop     bx
        retf

Function6:
        jmp SpeakerOn           ; Too far to jump directly

; Function 5. Set ADC Input Source.
; Sound Blaster Pro only.

; AH =  0 or 2 - Microphone
;       1 - CD
;       3 - Line-in
;

SetSource:
        push    bx
        mov     bh,0Ch          ; Input Options register 0Ch
        call    ReadMixer       ; Read Input Options

        shl     ah,1            ; Make mask for input source
        and     ah,6

        mov     bl,al           ; Input Options read from mixer
        and     bl,not 6        ; Clear input source bits
        or      bl,ah           ; Set new input source bits
        call    WriteMixer
        pop     bx
        retf

; Function 6. Turn DAC Speaker On.
; AH =  0 - off
;       1 - on

SpeakerOn:
        and     ah,ah
        jz      Off
        mov     ah,0D1h         ; Command : DAC speaker on
        call    WriteDSP        ; Send the command
        retf
Off:
        mov     ah,0D3h         ; Command : DAC speaker off
        call    WriteDSP        ; Send the command
        retf


INCLUDE READWRIT.INC
INCLUDE MIXER.INC


CODE            ENDS
END
