/*
   SBWAV.C
   This program plays and records WAV files, using driver
   CT-VOICE.DRV by Creative Labs. When loading a file, the
   program substitutes file header with VOC sound block header.
   Backward conversion is done when saving recorded sample.

   Compiled and tested in large model with Borland C++ 3.1
*/
#include <stdio.h>
#include <conio.h>
#include <ctype.h>

#include "sbenv.h"
#include "wav.h"

WORD double_buffer_size = 2;  /* 128Kb */

BYTE play_wav = 0;
BYTE record_wav = 0;

int wav_handle;

int _argc;
char **argum;

WORD wav_status;
void far *driver;
char *wav_file = NULL;

/* Load CT-VOICE.DRV, open WAV file */
void prepare_wav(void) {
  int i;
  char driver_name[80];
  WORD drv_error;
  BYTE version_major, version_minor;

  /* Find WAV file name in command line */
  for (i = 1; i < _argc; i++) {
    if (strstr(argum[i], ".WAV") != NULL) {
      wav_file = argum[i];
      break;
    }
  }

  /* No file! */
  if (wav_file == NULL) return;

  /* Shall we record? */
  for (i = 1; i < _argc; i++) {
    if (strcmp(argum[i], "/R") == 0) {
      record_wav = 1;
      break;
    }
  }

  /* Query Sound Blaster settings */
  get_sb_env();

  strcpy(driver_name, sb_dir);
  strcat(driver_name, "\\DRV\\CT-VOICE.DRV");
  driver = load_driver(driver_name);
  if (driver == NULL) {
    printf("Can't load driver %s\n", driver_name);
    exit(8);
  }

  asm {
    xor bx,bx		/* Get driver version */
    call driver
    mov version_major,ah
    mov version_minor,al
  }

  printf("Driver version %d.%d.\n", version_major, version_minor);

  asm {
    mov bx,1		/* Set base address */
    mov ax,base_port
    call driver

    mov bx,2		/* Set IRQ line */
    mov ax,IRQ_line
    call driver

    mov bx,19		/* Set DMA channel */
    mov ax,DMA_channel
    call driver

    mov bx, 3		/* Initialize */
    call driver
    mov drv_error,ax
  }

  if (drv_error != 0) {
    printf(
      "Error %d initializing Sound Blaster!\n", drv_error);
    exit(drv_error);
  }

  asm {
    push es
    push di
    mov bx,5		/* Set up status word */
    mov di, seg wav_status
    mov es, di
    lea di, wav_status
    call driver
    pop di
    pop es
  }

  if (record_wav)
    wav_handle = open(wav_file, O_WRONLY | O_CREAT | O_BINARY);
  else
    wav_handle = open(wav_file, O_RDONLY | O_BINARY);

  if (wav_handle == -1) {
    printf("Error opening %s !\n", wav_file );
    record_wav = 0;
    return;
  }

  play_wav = !record_wav;
}  /* prepare_wav */

/* Load and convert WAV file */
void load_wav(void){
  int res;
  BYTE huge *cur_pos;
  BYTE huge *data_pos;

  read( wav_handle, &wav_head, sizeof( WAV_HEADER ) );

  if ((strncmp(wav_head.WAVEfmt_txt, "WAVEfmt", 7) )) {
    printf("%s is not a WAV file!", wav_file);
    exit(1);
  }

  if (wav_head.data_size + 20 > double_buffer_size*0x10000){
    printf("%s is too big to fit in memory !", wav_file);
    exit(2);
  }

  cur_pos = double_buffer;

  if ( (wav_head.channels > 1) || (wav_head.sample_rate > 22050) ){
    /* Create additional prefix  -
       "override" block before sound block */
    *cur_pos = 8; /* block tag */
    cur_pos += 3;
    *cur_pos++ = 4; /* block size */
    /* time constant*/
    *((WORD*) cur_pos) = (0x10000UL - 256000000UL /
       (wav_head.channels * wav_head.sample_rate) );
    cur_pos += 2;
    *cur_pos++ = 0;
    *cur_pos++ = wav_head.channels - 1;
  }
  /* Create sound block prefix */
  cur_pos++;
  *((DWORD*) cur_pos) = wav_head.data_size + 2; /* Sound data length*/
  *--cur_pos = 1; /* Sound block tag */
  cur_pos += 4;
  /* Time constant*/
  *cur_pos++ = 256 - 1000000/wav_head.sample_rate;
  *cur_pos++ = 0; /* no compression */
  data_pos = cur_pos;
  /* Read digitized sound data */
  while ( (cur_pos - data_pos) < wav_head.data_size )
	  cur_pos += read(wav_handle, cur_pos, 32000);
  *cur_pos++ = 0; /* Terminator, too */
  close( wav_handle );
} /* load_wav */

void save_wav(void){
  BYTE huge *cur_pos;
  BYTE huge *data_pos;
  WORD channels, rate;
  DWORD data_length;

  cur_pos = double_buffer;
  if (*cur_pos == 8) {
    /* Parse override block's info */
    channels = cur_pos[7] + 1;
    cur_pos += 4;
    rate = *((WORD*) cur_pos);
    rate = 256000000 / (0x10000 - rate) /channels;
    cur_pos += 8;
    data_pos = cur_pos + 6;
    data_length = *(DWORD *)cur_pos & 0x00FFFFFFu - 2;
  }
  else if (*cur_pos == 9) {
    channels = cur_pos[9];
    rate = *(DWORD *)&cur_pos[4];
    data_pos = cur_pos + 16;
    data_length = *( (DWORD*) &cur_pos[1]) & 0x00FFFFFFu - 12;
  }
  else {  /* There's only sound block */
    channels = 1;
    rate = 1000000 /(256 - cur_pos[4]);
    data_pos = cur_pos + 6;
    data_length = *(DWORD *)cur_pos & 0x00FFFFFFu - 2;
  }
  /* Calculate all excessive data in WAV header */
  calc_wav_header( &wav_head, channels, 8, rate, data_length );
  /* Write WAV header */
  write(wav_handle, &wav_head, sizeof(WAV_HEADER) );

  cur_pos = data_pos;
  /* Write sound data */
  while ( (cur_pos - data_pos) < wav_head.data_size )
	  cur_pos += write(wav_handle, cur_pos, 32000);
  close( wav_handle );
} /* save_wav */

void main(int argc, char **argv) {
  int i;
  char key;
  WORD rec_source;
  WORD sampling_rate;

  printf("SBWAV Plays/records WAV files.\n");
  if (argc == 1) {
    printf("Syntax : SBWAV myvoice.wav [/r]\n"
      "/r is optional command to record a WAV,"
      " default is to play.\n");
    exit(1);
  }

  for(i = 0; i < argc; i++)
    strupr(argv[i]);

  _argc = argc;
  argum = argv;

  alloc_common_buffer(double_buffer_size);
  if (common_buffer == NULL) {
    printf ("No memory for buffer.\n");
    exit (1);
  }

  prepare_wav();
  if (play_wav)
    load_wav();

  if (!(play_wav | record_wav ))
    exit(0);  // Nothing to do

  if (record_wav) {
    printf(
      "To select recording source on Sound Blaster Pro, press:\n"
      "[M] - microphone (default), [C] - CD, [L] - line-in\n");

    switch (toupper(getch())) {
      case '\0' : getch();

      default :
      case 'M' : rec_source = 0; break;

      case 'C' : rec_source = 1; break;

      case 'L' : rec_source = 3; break;

      case '\x1b' :
	asm {
	  mov bx,9	/* Terminate driver */
	  call driver
        };
	 exit(0);
    }

    asm {
      // Set recording source
      mov bx,17
      mov ax, rec_source
      call driver
      // Set recording mode
      mov bx,16
      mov ax,0  // mono
      call driver
    }

    printf("Enter sampling rate in Hz: ");
    scanf("%d", &sampling_rate);
  }

  printf("Hit any key to start...\r");
  if (getch() == 0) getch();

  if (record_wav) {
    asm {
      push di
      push es

      mov bx,4		/* DAC speaker off */
      mov al, 0
      call driver

      mov bx,7 	/* Record */
      les di, double_buffer
      mov ax,sampling_rate
      /* double_buffer_size is in 64Kb blocks */
      mov dx, double_buffer_size
      xor cx,cx
      call driver
      pop es
      pop di
    }
  }

  if (play_wav) {
    asm {
      push di
      push es

      mov bx,4  /* DAC speaker on */
      mov al, 1
      call driver

      mov bx,6   /* Play */
      les di, double_buffer
      call driver
      pop es
      pop di
    }
  }

  printf("Hit any key to stop...   ");
  if (getch() == 0) getch();

  if (record_wav | play_wav)
    asm {
      mov bx,8		/* Stop VOC I/O */
      call driver

      mov bx,4		/* DAC speaker off */
      mov al, 0
      call driver

      mov bx,9		/* Terminate driver */
      call driver
    }

  if (record_wav)
    save_wav();
}
