/*
   TESTDRV.C
   This program demonstrates cooperation with 
   demo driver MYDRIVER.ASM.
   It performs digitized sound input and displays volume level.
*/

/* Large memory model */
#pragma -ml
#include <stdio.h>
#include <conio.h>
#include <ctype.h>

#include "sbenv.h"

void *driver;

/* Show volume level of sampled sound */
void indicate(void) {
  char s[18];
  int level, i;
  BYTE sample;
  BYTE max_level = 0;
  s[16] = '\xc3';
  s[17] = '\0';

  /* Sample 2000 bytes to find out peak level */
  for (i = 0; i < 2000; i++) {
    asm {
      mov al,2          /* Sample one byte */
      call driver
      mov sample,al
    }
    level = ((sample > 128) ? sample - 128 : 127 - sample) / 8;
    if (level > max_level) max_level = max_level;
  }

  for(i = 0; i <= level; i++) s[i] = '';
  for(i = level + 1; i < 16; i++) s[i] = '-';
  printf("\r%s", s);
}

/* To make SBDEMO.BIN, MYDRIVER.ASM be compiled with MASM,
linked with LINK, and converted with EXE2BIN */
char driver_name[] = "MYDRIVER.DRV";

void main(void) {
  WORD drv_error;
  BYTE version_major, version_minor;
  signed char source = -1; /* ADC source */

  get_sb_env(); /* Query Sound Blaster settings */
  driver = load_driver(driver_name); /* Load driver */
  if (driver == NULL) {
    printf("Can't load driver %s\n", driver_name);
    exit(8);
  }

  asm {
    xor al,al           /* Function 0 :
                            Set Base Address */
    mov bx,base_port
    call driver
    mov drv_error,ax
  }

  if (drv_error != 0) {
    printf("Error initializing Sound Blaster\n");
    exit(8);
  }

  asm {
    mov al,1            /* Get DSP version */
    call driver
    mov version_major,ah
    mov version_minor,al
  }

  printf("DSP version %d.%d\n", version_major, version_minor);

  asm {
    mov al,6            /* Turn speaker... */
    mov ah,0            /*      ... off */
    call driver
  }

  printf("This is sample volume indicator.\n");
  printf("SB Pro users: to test input from a source press :\n");
  printf("[M] - mic, [C] - CD, [L] - Line-in\n");

  while(1) {
    indicate();

    if (kbhit()) {
      switch (toupper(getch())) {
        case '\0': getch(); break;

        case '\x1b': exit(0);

        case 'M' :
          source = 2;
          printf("\r\t\t\tMicrophone");
          break;

        case 'C' :
          source = 1;
          printf("\r\t\t\tCD player ");
          break;

        case 'L' :
          source = 3;
          printf("\r\t\t\tLine-in   ");
          break;
      }

      if (source != -1) {
        asm {
          mov al,5  /* Set ADC source */
          mov ah,source
          call driver
        }
        source = -1;
      }
    }
  }
}
/* The end. */
