/*
   VOCMID.C
   Demonstrates usage of CTVDSK.DRV, a driver by
   Creative Labs for digital sound I/O, and SBMIDI.EXE, a
   resident driver by Creative Labs for MIDI music.

   The program is a simple solution for synchronizing
   digital voice with MIDI, because it plays MID and VOC
   files together.

   Compiled and tested under large model with Borland C++ 3.1.
*/
#include <stdio.h>
#include <conio.h>
#include <ctype.h>

#include "sbenv.h"

/* How many physical memory pages to use for double buffer */
WORD double_buffer_size = 1; /* 64Kb */

BYTE play_voc = 0;
BYTE record_voc = 0;
BYTE play_mid = 0;
int voc_handle;
int argcnt;
char **argum;
WORD voc_status, mid_status;
void *driver;

/* Open VOC file for read or write. Load and initialize driver. */
void prepare_voc(void) {
  int i;
  char *voc_file = NULL;
  char driver_name[80];
  WORD drv_error;
  BYTE version_major, version_minor;

  /* Find VOC file name */
  for (i = 1; i < argcnt; i++) {
    if (strstr(argum[i], ".VOC") != NULL) {
      voc_file = argum[i];
      break;
    }
  }

  /* No VOCs! */
  if (voc_file == NULL) return;

  /* Shall we record? */
  for (i = 1; i < argcnt; i++) {
    if (strcmp(argum[i], "/R") == 0) {
      record_voc = 1;
      break;
    }
  }

  /* Query Sound Blaster settings */
  get_sb_env();

  strcpy(driver_name, sb_dir);
  strcat(driver_name, "\\DRV\\CTVDSK.DRV");
  driver = load_driver(driver_name);

  if (driver == NULL) {
    printf("Can't load driver %s\n", driver_name);
    exit(8);
  }

  asm {
    xor bx,bx 		/* Ask version number */
    call driver
    mov version_major,ah
    mov version_minor,al
  }

  printf("Driver version %d.%d.\n", version_major, version_minor);

  asm {
    mov bx,1		/* Set base address */
    mov ax,base_port
    call driver

    mov bx,2		/* Set IRQ line */
    mov ax,IRQ_line
    call driver

    mov bx,19		/* Set DMA channel */
    mov ax,DMA_channel
    call driver

    mov bx,15               /* Set up double buffer*/
    mov ax, WORD PTR double_buffer
    mov dx, WORD PTR double_buffer+2
    /* Half-size of double buffer in 2Kb units */
    mov cx, double_buffer_size
    shl cx,1
    shl cx,1
    shl cx,1
    shl cx,1
    call driver

    mov bx, 3		/* Initialize */
    mov ax, 8
    call driver
    mov drv_error,ax
  }

  if (drv_error != 0) {
    printf("Error %d initializing Sound Blaster!\n", drv_error);
    exit(drv_error);
  }

  asm {
    push es
    push di
    mov bx,5		/* Set up status word */
    mov di, seg voc_status
    mov es, di
    lea di, voc_status
    call driver
    pop di
    pop es
  }

  if (record_voc)
    voc_handle = open(voc_file, O_TRUNC |
      O_CREAT | O_BINARY, S_IWRITE);
  else
    voc_handle = open(voc_file, O_RDONLY | O_BINARY);

  if (voc_handle == -1) {
    printf("Error opening %s !\n", voc_file );
    record_voc = 0;
    return;
  }
  play_voc = !record_voc;
}  /* prepare_voc */

/* Load MIDI file and prepare driver. */
void prepare_mid(void) {
  int i;
  char *mid_file = NULL;
  int mid_handle;
  void *int80h;

  /* Find MID file name in command line */
  for (i = 1; i < argcnt; i++) {
    if (strstr(argum[i], ".MID") != NULL) {
      mid_file = argum[i];
      break;
    }
  }

  /* No MIDs! */
  if (mid_file == NULL) return;


  /* Is SBMIDI installed! */
  int80h = MK_FP( peek(0, 0x80*4 + 2), peek(0, 0x80*4 ) );
  if (int80h == NULL) {
    printf("SBMIDI not installed.\n");
    return;
  }

  mid_handle = open(mid_file, O_RDONLY | O_BINARY);

  if (mid_handle == -1) {
    printf("Error opening %s !\n", mid_file );
    return;
  }

  if (filelength(mid_handle) > midi_size) {
    printf("%s is too long!\n", mid_file );
    close(mid_handle);
    return;
  }

  read(mid_handle, midi_buffer, filelength(mid_handle));
  close(mid_handle);

  asm{
    mov bx,3		/* Set up status word */
    mov dx, seg mid_status
    lea ax, mid_status
    int 80h		/* Call MIDI driver */

    mov bx,4		/* Set up MIDI buffer */
    mov dx, WORD PTR midi_buffer + 2
    mov ax, WORD PTR midi_buffer
    int 80h
  }

  play_mid = 1;
} /* prepare_mid */


void main(int argc, char **argv) {
  int i;
  char key;
  WORD rec_source;
  WORD sampling_rate;

  printf("VOC'n'MID\n"
    "Plays/records VOC file while MIDI file is being played.\n");
  if (argc == 1) {
    printf("Syntax : VOCMID mysong.mid myvoice.voc [/r]\n"
    "/r is optional command to record a VOC, default is to play.\n");
    exit(1);
  }

  for(i = 0; i < argc; i++)
    strupr(argv[i]);

  argcnt = argc;
  argum = argv;

  /* Allocate all buffers */
  alloc_common_buffer(double_buffer_size);

  if (common_buffer == NULL) {
    printf("No memory for buffer.\n");
    exit(1);
  }

  prepare_mid();
  prepare_voc();

  if (!(play_mid | play_voc | record_voc ))
    exit(0); /* Nothing to do! */

  if (record_voc) {
    printf(
      "To select recording source on Sound Blaster Pro, press:\n"
      "[M] - microphone (default), [C] - CD, [L] - line-in\n");
    switch (toupper(getch())) {
      case '\0' : getch();

      default :
      case 'M' : rec_source = 0; break;

      case 'C' : rec_source = 1; break;

      case 'L' : rec_source = 3; break;

      case '\x1b' :
        asm {
	  mov bx,9	/* Terminate */
	  call driver
	};
	exit(0);
    }

    asm {
      mov bx,17	/* Set recording source */
      mov ax, rec_source
      call driver
    }

    printf("Enter sampling rate in Hz: ");
    scanf("%d", &sampling_rate);
  }

  printf("Hit any key to start...\r");
  if (getch() == 0) getch();

  if (record_voc) {
    asm {
      mov bx,4	/* DAC speaker off*/
      mov al, 0
      call driver

      mov bx,7 /* Record... */
      mov ax,voc_handle /* ... to file... */
      mov dx,sampling_rate /* ... at rate */
      call driver
    }
  }

  if (play_voc) {
    asm {
      mov bx,4	/* DAC speaker on */
      mov al, 1
      call driver
      mov bx,6
      mov ax,voc_handle
      call driver
    }
  }

  if (play_mid) {
    asm {
      mov bx,5	/* Play song */
      int 80h
    }
  }

  printf("Hit any key to stop...   ");
  if (getch() == 0) getch();

  if (record_voc | play_voc)
    asm {
      mov bx,8		/* Stop VOC I/O */
      call driver

      mov bx,4		/* DAC speaker off */
      mov al, 0
      call driver

      mov bx,9		/* Terminate driver */
      call driver
    }

  if (play_mid)
    asm {
      mov bx,4	/* Stop MIDI */
      int 80h
    }
} /* The end. */
