; clear.asm: Sets all bits in ax and displays it.
;            Clears low-order 8 bits of ax and displays ax again.

		.model small
		.stack 100h

		.data

s1      db  'ax before: ',0
s2      db  13, 10, 'ax after:  ',0

		.code

start:
		mov ax, @data
		mov ds, ax

		mov ax, offset s1   ; display before message
		call put_str

		mov ax, 0ffffh      ; set all bits in ax  1111 1111 1111 1111
		call put_bin        ; display ax in binary

		call put_tab        ; display a tab

		call put_hex        ; display ax in hex i.e. ffffh

		mov ax, offset s2   ; display after message
		call put_str

		mov ax, 0ffffh      ; set bits in ax again

		and ax, 0ff00h      ; clear 8 low-order bits

		call put_bin        ; display ax in binary: 1111111100000000
		call put_tab        ; display a tab
		call put_hex        ; display ax in hex: ff00h


		mov ax, 4c00h       ; terminate program
		int 21h

;               User Defined subprograms.


put_hex:                        ; display ax in hexadecimal
		push ax
		push bx
		push cx
		push dx

		mov  bx, 16
		mov  cx, 4

gethx:                          ; repeatedly divide by 16
		mov  dx, 0
		div  bx         ; ax/bx: rem in dx,res in ax
		push dx         ; store digits on stack
		loop gethx
				; now pop digits from stack and display them
		mov cx, 4       ; there 4 digits to be displayed

disphx: 
		pop ax       
		cmp ax, 10      ; if ax > 10, then display a letter a - f
		jge displetter
		add ax, '0'     ; otherwise convert to ASCII and display digit
		call putc
		jmp next        ; process next digit

displetter:                     ; digit values 10-15 map onto a-f
				; (digit value + 'a'-10) gives ASCII code
		add ax, 'a' - 10
		call putc       ; display the letter

next:           loop disphx

		mov al, 'h'     ; display h for hex at end
		call putc
	   
		pop dx          ; restore registers
		pop cx
		pop bx
		pop ax
		ret

put_bin:                         ; Display ax in binary
		push ax          ; save registers
		push bx
		push cx
		push dx

		mov dx, ax      ; al will be used to display bits
		mov cx, 16      ; counter for loop
		mov bx, 8000h   ; bl = 1000 0000 0000

check:
		test dx, bx     ; test if bit set in bx is also set in dx
		jnz display_1   ; if its not set, display a 1
		mov al, '0'     ; otherwise display a 0
		jmp rest

display_1:
		mov al, '1'

rest:           call putc
		shr bx, 1       ; bx = bx / 2 adjust mask 1 bit right
		loop check      ; repeat for all 16 bits

		pop dx          ; restore registers
		pop cx
		pop bx
		pop ax

		ret

put_tab:                        ; display a tab
		push ax
		mov al, 9       ; ASCII code for tab is 9
		call putc
		pop ax
		ret

put_str:                               ; display string terminated by 0 
				       ; whose address is in ax

		push bx                ; save registers
		push cx
		push dx

		mov bx, ax              ; store address in bx
		mov al, byte ptr [bx]   ; In C: al = str[0]

put_loop:       cmp al, 0               ; al == 0 ?
		je  put_fin             ; while al != 0
		call putc               ; display character
		inc bx                  ; bx = bx + 1
		mov al, byte ptr [bx]   ; In C: al = str[i]
		jmp put_loop            ; repeat loop test
put_fin:        
		pop dx                  ; restore registers
		pop cx
		pop bx
		ret




putc:                                   ; display character in al
		push    ax
		push    bx
		push    cx
		push    dx

		mov dl, al
		mov ah, 2h
		int 21h

		pop dx  
		pop cx  
		pop bx  
		pop ax  

		ret
		
		end start


